#!/bin/sh
#
# bcadm: Initialization script for Broadcom Raid Management
#
# chkconfig: 2345 96 4
# description: Loads the bccfg module for the management \
#              utility
#
#

# Source function library.
. /etc/rc.d/init.d/functions

start() {

    # if udev isn't in charge, clean up any old bccfg devices
    if grep udev /proc/sys/kernel/hotplug > /dev/null 2>&1; then
	UDEV=yes
    else
	UDEV=no
	rm -rf /dev/bccfg?
    fi
    
    # Load the bccfg module
    if ! grep bccfg /proc/modules > /dev/null 2>&1 ; then
	[ -x /sbin/modprobe ] || {
	    echo "need /sbin/modprobe"
	    exit 1
	}

	/sbin/modprobe bccfg || exit 1

	# give udev a chance to catch up
	test $UDEV == yes && sleep 1
    fi

    # skip making /dev/bccfg? if udev is taking care of it...
    if [ $UDEV != yes -o ! -b /dev/bccfg0 ]; then

        # Create the device nodes
	major=$(grep bccfg /proc/devices | cut -d' ' -f1)
	if [ -z "$major" ] ; then 
	    echo "bccfg: Couldn't find major device number"
	    exit 1
	fi

	for i in 0 1 2 3 4 5 6 7 8; do
	    test -b /dev/bccfg$i || mknod /dev/bccfg$i b $major $i
	done
    fi
}

stop() {
   grep bccfg /proc/modules > /dev/null 2>&1 && /sbin/rmmod bccfg
   rm -rf /dev/bccfg?
}

# See how we were called.
case "$1" in
    start)
	start
	;;
    stop)
	stop
	;;
    status)
        ;;
    restart)
	stop
	start
	;;
    condrestart)
	;;
    reload)
        ;;
    *)
	echo $"Usage: $0 {start|stop|status|restart|condrestart|reload}"
	;;
esac
exit $RETVAL
