
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

#ifndef _BC_OSHEADERS_H_
#define _BC_OSHEADERS_H_

#include <linux/version.h>

#include <linux/module.h>
/*#include <linux/config.h>*/
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/completion.h>
/*#include <asm/semaphore.h>*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_tcq.h>
#else
#include <linux/blk.h>
#include <linux/tqueue.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"
#endif

#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/highmem.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/random.h>



/* Misc fixups for the 2.6 kernel */

/* Fake out support for interrupt return type */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,22)
typedef void irqreturn_t;
#define IRQ_HANDLED
#endif

/* 2.6 kernel dosen't include a typedef */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
typedef struct scsi_host_template Scsi_Host_Template;
#endif

/* 2.6 kernel dosen't use io_request_lock */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define GET_IO_REQUEST_LOCK                spin_lock_irq(&io_request_lock);
#define PUT_IO_REQUEST_LOCK                spin_unlock_irq(&io_request_lock);
#define GET_IO_REQUEST_LOCK_IRQSAVE(i)     spin_lock_irqsave(&io_request_lock, i);
#define PUT_IO_REQUEST_LOCK_IRQRESTORE(i)  spin_unlock_irqrestore(&io_request_lock, i);
#else
#define GET_IO_REQUEST_LOCK
#define PUT_IO_REQUEST_LOCK
#define GET_IO_REQUEST_LOCK_IRQSAVE(i)
#define PUT_IO_REQUEST_LOCK_IRQRESTORE(i)
#endif


#include "bc_types.h"
#include "bc_srb.h"
#include "bc_scsi.h"
#include "bc_msg.h"
#include "bc_adapter.h"

typedef enum
{
	BC_PANIC = 0,
	BC_ALERT,
	BC_CRITICAL,
	BC_ERROR,
	BC_WARN,
	BC_NOTE, 
	BC_INFO, 
	BC_INFO2,
	BC_DEBUG,	// 8
	BC_DEBUG2,	// 9
	BC_DEBUG3,	// 10
	BC_DEBUG_ISR,	// 11
	BC_DEBUG_HWISR,	// 12
	BC_TAIL
} bc_print_lvl_t;

#define BC_DEFAULT_ERR_LEVEL BC_INFO


void	bc_printk( int flag, const char *fmt, ...)
       __attribute__ ((format (printf, 2, 3)));

/* Fast version of bc_printk for use in data path. */
#define BC_PRINTK(flag, fmt, ...)  if (flag <= bc_msg_level) bc_printk(flag, fmt, __VA_ARGS__);


#ifdef BC_SW_DEBUG
#define BC_ASSERT(cond) BUG_ON(!(cond))
#else
#define BC_ASSERT(cond)
#endif

void bc_start_all_threads(void);
void bc_stop_all_threads(void);
void bc_event(bc_uint32_t type, bc_uint8_t bus, int update_mode);
int bc_event_init(void);
void bc_event_shutdown(void);

extern bc_softstate_t	   bc_state;
extern bc_adapter_t       *bc_dev[];
extern int 		   bc_msg_level;

#endif // _BC_OSHEADERS_H_









