
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 *
 * Module name:
 *
 *    Adapter.h
 *
 * Abstract:
 *
 *	all the stuff we need to know and to track for each physical adapter and the virtual adapter
 *
 * Author:
 *
 *
 */
#ifndef _BC_ADAPTER_
#define _BC_ADAPTER_

struct bc_adapter_s;
struct bc_ide_port_s;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
typedef irqreturn_t (* bc_isr_func_t)     ( int irq, void *arg);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
typedef irqreturn_t (* bc_isr_func_t)     ( int irq, void *arg, struct pt_regs *regs);
#else
typedef void (* bc_isr_func_t)     ( int irq, void *arg, struct pt_regs *regs);
#endif
#endif

typedef int  (* bc_adapter_func_t) ( struct bc_adapter_s *adapter);
typedef int  (* bc_port_func_t)    ( struct bc_ide_port_s  *port);
typedef void  (* bc_iodone_t)    ( struct bc_ide_port_s  *port, void *arg);

/*
 * Hardware info for the adapter
 */

#define MAX_PORTS_PER_HA	8
#define MAX_HBA			4
#define MAX_TOTAL_PORTS		(MAX_HBA * MAX_PORTS_PER_HA)
#define MAX_ARRAY		17
#define SECTOR_SIZE		512


/*
 * Definitions for accessing IDE controller registers
 */
#define IDE_NR_REGS            9

#define IDE_DATA	0
#define IDE_ERROR	1	/* read mode  */
#define IDE_FEATURE	1	/* write mode */
#define IDE_SECTOR_CNT	2
#define IDE_LBA_LOW	3
#define IDE_LBA_MID	4
#define IDE_LBA_HIGH	5
#define IDE_DEVICE	6
#define IDE_COMMAND	7	/* write mode */
#define IDE_CONTROL	8	/* write mode */
#define IDE_STATUS	7	/* read mode  */
#define IDE_ALT_STATUS	8	/* read mode  */

#define IDE_DATA_REG(x)		((x)->ide_regs[IDE_DATA])
#define IDE_ERROR_REG(x)	((x)->ide_regs[IDE_ERROR])
#define IDE_FEATURE_REG(x)	((x)->ide_regs[IDE_FEATURE])
#define IDE_SECTOR_CNT_REG(x)	((x)->ide_regs[IDE_SECTOR_CNT])
#define IDE_LBA_LOW_REG(x)	((x)->ide_regs[IDE_LBA_LOW])
#define IDE_LBA_MID_REG(x)	((x)->ide_regs[IDE_LBA_MID])
#define IDE_LBA_HIGH_REG(x)	((x)->ide_regs[IDE_LBA_HIGH])
#define IDE_DEVICE_REG(x)	((x)->ide_regs[IDE_DEVICE])
#define IDE_STATUS_REG(x)	((x)->ide_regs[IDE_STATUS])
#define IDE_COMMAND_REG(x)	((x)->ide_regs[IDE_COMMAND])
#define IDE_ALT_STATUS_REG(x)	((x)->ide_regs[IDE_ALT_STATUS])
#define IDE_CONTROL_REG(x)	((x)->ide_regs[IDE_CONTROL])

#define CMD_IDENTIFY	0xEC
#define CMD_MEDIA_STAT	0xDA
#define CMD_READ_SECTORS	0x20
#define CMD_READ_SECTORS_EXT	0x24
#define CMD_WRITE_SECTORS	0x30
#define CMD_WRITE_SECTORS_EXT	0x34
#define CMD_FLUSH_CACHE	0xE7




/* 
 * register bits
 */

/* Device Control Register */

#define HIGH_ORDER_BYTE		0x80
#define SOFTWARE_RESET		0x04
#define IE_DISABLE		0x02

/* Device Status Register */
#define STATUS_BUSY		0x80
#define STATUS_READY		0x40
#define STATUS_FAULT		0x20
#define STATUS_DRQ		0x08	/* Data ReQuest */
#define STATUS_ERROR		0x01

#define STATUS_BITS (STATUS_BUSY | STATUS_READY | STATUS_FAULT | STATUS_DRQ | STATUS_ERROR)


/* Device Error Register */
#define READ_UNCORRECTABLE	0x40
#define WRITE_WRITE_PROTECT	0x40

#define MEDIA_CHANGE		0x20
#define LBA_ERROR		0x10
#define MEDIA_CHANGE_REQ	0x08
#define COMMAND_ABORT		0x04
#define NO_MEDIA		0x02

#define RW_ERROR ( MEDIA_CHANGE | LBA_ERROR | MEDIA_CHANGE_REQ | COMMAND_ABORT | NO_MEDIA )


/* Device Device Register */
#define LBA_MASK	0xf
#define DEV_SELECT	0x10
#define SET_LBA_MODE	0x40




typedef struct bc_ide_request_s {
	int			flags;
	int			status;		/* unused ?  GROT */
	int			cmd;
	int32_t			lba;
	int32_t			hob_lba;	/* 48 bit addressing */
	void			*vaddr;		/* start of transfer buffer */
	void			*cur_vaddr;	/* current offset */
	int			byte_count;;
	int			byte_resid;
	int			sector_count;
	int			sector_resid;	/* sectors remaining */
	struct semaphore	wait_sema;
	bc_iodone_t		iodone;
	struct timer_list	timer;
	void 			*scsi_context;
	struct bc_ide_port_s	*port;
} bc_ide_request_t;
/*
 * flags
 */
#define 	HOB_VALID	0x1	/* 48 bit adderessing valid */
#define 	DRIVE_PRESENT	0x2
#define 	DRIVE_LOCKED	0x4
/*
 * cmd
 */
#define BC_READ		0x1
#define BC_WRITE	0x2
#define BC_IDENTIFY	0x3

typedef struct bc_ide_port_s {
	uint32_t 		*ide_regs[IDE_NR_REGS];
/* Ugly.  Granite specific.  needs it's own struct */
	uint32_t 		*edma_cmd_reg;
	uint32_t 		*edma_status_reg;
/* */
	int			state;
	bc_port_func_t 		intr_func;
	char 			*ident;			/* identify packet */
	bc_ide_request_t	*req;
	spinlock_t		lock;
	unsigned long		flags;			/* irq flags */
	int			port_number;
	int			device_type;
	uint32_t		media_cap_low;
	uint32_t		media_cap_high;
	struct bc_adapter_s	*adapter;
	struct timer_list	timer;
	sense_data_t		sense_data;
} bc_ide_port_t;

#define CMD_PENDING		0x01
#define CMD_PENDING_COMPLETE	0x02
#define CMD_COMPLETE		0x04
#define CMD_ERROR		0x08
#define CMD_TIMEOUT		0x10

	

#define PCI_CFG_SIZE	64
typedef struct bc_hw_info {
	int			vendor_id;
	int			device_id;
	int			subvendor_id;
	int			subsystem_id;
	int			pci_bus;
	int			pci_slot;
	int			pci_function;
	int			irq;
	int			vector;		// Interrupt vector, may not be IRQ on smp machines
	void			*vaddr;		// device memory map address
// Need to deal with 64bit paddrs.
// These types come from linux/ioport.h
	unsigned long 		phys_addr;	// 32 bit bus address
	int			mem_len;	// len of memory map

//  We have to map in the sram for Elrond based products
	void			*sram_vaddr;	// SRAM location
	unsigned long 		sram_paddr;	// 64-bit bus address
	int			sram_len;
	int			io_port;
	int			io_port_len;
	int			adapter_number;
	unsigned char		pci_config_space[PCI_CFG_SIZE];
} bc_hw_info_t;


typedef	struct bc_mem_desc_s {
	void		*vaddr;
	dma_addr_t 	dma_address;
	unsigned int	size;
} bc_mem_desc_t;

typedef struct bc_adapter_s {
	struct bc_version		*version;
	char				*name;
	void				*dip;
	uint16_t                     	intr_mask;
	int				instance;	
	int				sys_hba_num;	
	bc_hw_info_t			hardware;
	struct pci_dev			*pdev;
	int				enable_interrupt_moderation;

	void 				*mem;		// memory allocated for the adapter
	bc_ide_port_t			port[MAX_PORTS_PER_HA];
	struct timer_list 		timer;
	struct semaphore		timer_sema;

	bc_mem_desc_t			private_mem;
	
} bc_adapter_t;

/*
 * things need to implement srb_done/srb_queue tasklets
 */

typedef struct bc_srb_queue_s {
	spinlock_t		lock;
	bc_srb_t		*head;
	bc_srb_t		*tail;
	struct tasklet_struct	tasklet;
} bc_srb_queue_t;

typedef struct bc_mop_queue_s {
	spinlock_t		lock;
	bc_mem_op_t		*head;
	bc_mem_op_t		*tail;
} bc_mop_queue_t;


typedef struct bc_stats_s {
	uint		target_total[MAX_ARRAY];
	atomic_t	target_pending[MAX_ARRAY];
	uint		srb_total;			/* SRBs sent to the OSIC */
	atomic_t	srb_pending;			/* SRBs sent to the OSIC */
	uint		scb_total;			/* SCBs from scsi layer  */
	atomic_t	scb_pending;			/* SCBs from scsi layer  */
	uint		max_srbs_sent;			/* ...into the core at one time in bc_msg_srb_q_tasklet */
	uint		max_intr_waiting;		/* max interrupts that came in while sending a batch of SRBs to the core */
	uint		total_intr_waiting;		/* total */
	uint		max_intr_delay;			/* max #cycles an interrupt waited for us to send SRBs to the core */
} bc_stats_t;


/* Global driver info */ 
typedef struct bc_softstate  {
	int 			state;
	struct Scsi_Host	*host_ptr;
	int			num_hba;
	/*
	 * private resources used by the OSIC
	 */
	int			memsize_per_controller;
	int			memsize_per_srb;
	int			timer_interval;		/* in clock ticks */
        uint32_t		virtual_memory_size;
	void                    *virtual_memory;
        uint32_t                cache_memory_size;
	void			*cache_memory;
	struct semaphore	init_sema;
	struct timer_list 	timer;
	spinlock_t		osic_lock;			/* lock for OSIC */
	int			osic_locked;
	char			*osic_lock_holder;
	struct timer_list 	bc_timeout;
	struct semaphore	bc_timeout_sema;
	struct timer_list 	msg_timeout;
	struct semaphore	msg_timeout_sema;
	struct timer_list 	debug_timer;
	struct bc_srb_queue_s	srb_q;
	struct bc_srb_queue_s	srb_done;
	struct bc_mop_queue_s	mop_done;
	struct tasklet_struct	intr_tasklet;
	int			intr_pending;
	spinlock_t		intr_lock;
	bc_stats_t		stats;

} bc_softstate_t;

/*
 * State flags
 */

#define USE_OSIC	0x2
#define ENABLE_TIMER	0x4
#define PROCESS_INTR	0x8		/* ready to process interrupts */




typedef struct bc_version {
	bc_adapter_func_t	init_func;
	bc_adapter_func_t	start_func;
	bc_adapter_func_t	shutdown_func;
	bc_isr_func_t		isr_func;
	char			*device_name;
	char			*vendor;
	char			*model;
	int			num_ports;
	int			window_size;
	int			which_bar;
} bc_version_t;



#define bc_interrupt_adapter(adapter) \
	adapter->adapter_funcs.interrupt_adapter(adapter)

#define bc_notify_adapter(adapter, adapter_event) \
	adapter->adapter_funcs.notify_adapter(adapter, adapter_event)

#define bc_enable_interrupt(adapter, adapter_event, at_device_irq) \
	adapter->adapter_funcs.enable_interrupt(adapter, adapter_event, at_device_irq)

#define bc_disable_interrupt(adapter, adapter_event, at_device_irq) \
	adapter->adapter_funcs.disable_interrupt(adapter, adapter_event, at_device_irq)

#define bc_reset_device(adapter) \
	adapter->adapter_funcs.reset_device(adapter)




#define read8	readb
#define read16	readw
#define read32	readl

#define write8	writeb
#define write16	writew
#define write32	writel


typedef union bc_sig_test_s {
	struct {
		uint32_t	cookie1;
		uint32_t	cookie2;
		uint32_t	lba;
		uint32_t	test_buff[100];
		uint32_t	cookie3;
		uint32_t	cookie4;
	} sig;
	unsigned char data[512];
} bc_sig_test_t;


#endif // _BC_ADAPTER_

