
/*
 * Copyright 2003-2005 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 * Module name:
 *   bc_init.c
 *				
 *	
 */

#include "version.h"

#define BC_DRIVER_VERSION           RC_VERSION_STR
#define BC_DRIVER_BUILD_DATE			__DATE__
#define BC_DRIVER_BUILD_TIME			__TIME__
#define BC_DRIVER_NAME				"bcraid"
#define BC_MAX_CMD_Q_DEPTH			1024
#define BC_DEFAULT_CMD_Q_DEPTH			512
#define BC_MAX_TAG_Q_DEPTH			255
#define BC_DEFAULT_TAG_Q_DEPTH			16

#include "bc.h"
#include "version.h"
#include "build_number.h"
#include "linux/hdreg.h"
#include "linux/reboot.h"
#include "bc_wildcat.h"
#include <asm/uaccess.h>


MODULE_AUTHOR("Broadcom, Inc");
MODULE_DESCRIPTION("Supports Broadcom RAIDCore(tm) BC4000/HT1000 series raid controllers");
MODULE_LICENSE("Dual BSD/GPL");

static int debug = 0;
#ifdef module_param_named
module_param_named(debug, debug, int, 0444);
#else
MODULE_PARM (debug, "i");
#endif
MODULE_PARM_DESC (debug, "debug print level");

static int cmd_q_depth = BC_DEFAULT_CMD_Q_DEPTH;
#ifdef module_param_named
module_param_named(cmd_q_depth, cmd_q_depth, int, 0444);
#else
MODULE_PARM (cmd_q_depth, "i");
#endif
MODULE_PARM_DESC (cmd_q_depth, "total command queue depth");

static int tag_q_depth = BC_DEFAULT_TAG_Q_DEPTH;
#ifdef module_param_named
module_param_named(tag_q_depth, tag_q_depth, int, 0444);
#else
MODULE_PARM (tag_q_depth, "i");
#endif
MODULE_PARM_DESC (tag_q_depth, "individual tagged command queue depth");

static int max_xfer = 4096;
#ifdef module_param_named
module_param_named(max_xfer, max_xfer, int, 0444);
#else
MODULE_PARM (max_xfer, "i");
#endif
MODULE_PARM_DESC (max_xfer, "max sectors per transfer");


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static int highmem_io = 1;
#ifdef module_param_named
module_param_named(highmem_io, highmem_io, int, 0);
#else
MODULE_PARM (highmem_io, "i");
#endif
MODULE_PARM_DESC (highmem_io, "enable highmem_io");
#endif

int	bc_config_debug = 0;



/*
 * Globals
 */

bc_softstate_t	   bc_state;

int bc_msg_level = BC_DEFAULT_ERR_LEVEL; // default MessageLevel

char                bc_name[] = { BC_DRIVER_NAME };
bc_adapter_t       *bc_dev[MAX_HBA];
int                 bc_single_cmd_done = 0;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static unsigned adapter_count = 0;     /* Number of adapters on the PCI bus,
                                          Used to determine when the last
                                          adapter has been initialized. */
#endif

/*
 * function prototypes
 */

void bc_shutdown_adapter(bc_adapter_t *adapter);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int bc_detect (Scsi_Host_Template *template);
int bc_release (struct Scsi_Host *host_ptr);
#endif

int	bc_do_scsi_cmd (struct scsi_cmnd * scmd, int wait);
void	bc_detach_adapter (bc_adapter_t * adapter);
void	bc_detach_interrupt (bc_adapter_t * adapter);
void	*bc_alloc_memory (size_t size, unsigned int flags);

static int bc_eh_abort_cmd (struct scsi_cmnd * scmd);
static int bc_eh_dev_reset (struct scsi_cmnd * scmd);
static int bc_eh_bus_reset (struct scsi_cmnd * scmd);
static int bc_eh_hba_reset (struct scsi_cmnd * scmd);

void bc_dump_scp( struct scsi_cmnd * scp );


int bc_frodo_init(bc_adapter_t *adapter);
int bc_frodo_start(bc_adapter_t *adapter);
int bc_frodo_shutdown(bc_adapter_t *adapter);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
irqreturn_t bc_frodo_isr(int irq, void *arg);
#else
irqreturn_t bc_frodo_isr(int irq, void *arg, struct pt_regs *regs);
#endif
void bc_frodo_intr_disable(bc_adapter_t *adapter);

int bc_elrond_init(bc_adapter_t *adapter);
int bc_elrond_start(bc_adapter_t *adapter);
int bc_elrond_shutdown(bc_adapter_t *adapter);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
irqreturn_t bc_elrond_isr(int irq, void *arg);
#else
irqreturn_t bc_elrond_isr(int irq, void *arg, struct pt_regs *regs);
#endif
void bc_elrond_intr_disable(bc_adapter_t *adapter);


int bc_ioctl (struct scsi_device * scsi_dev_ptr, int cmd, void *arg);

int bc_command (struct scsi_cmnd * scp);

void bc_timeout(int to);
void bc_timeout_done(unsigned long data);

const char *bc_info (struct Scsi_Host *host_ptr);


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int bc_slave_cfg(struct scsi_device *sdev);
int bc_bios_params (struct scsi_device *sdev, struct block_device *bdev, sector_t capacity, int geom[]);
#else
int bc_bios_params (Scsi_Disk *disk, kdev_t dev, int *geom);
#endif

void bc_select_queue_depths (struct Scsi_Host *host_ptr, struct scsi_device * scsi_device_ptr);


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) /* changed between 2.4.26 and 2.6.5 */
int bc_proc_info (char *buf, char **start, off_t offset, int buf_size, int host_num, int write);
#else
int bc_proc_info (struct Scsi_Host *shost, char *buf, char **start, off_t offset, int buf_size, int write);
#endif
static int bc_proc_read(char *page, char **start, off_t off, int count, int *peof, void *data);
static int bc_proc_write(struct file *file, const char *buf, unsigned long count, void *data);

int bc_queue_cmd (struct scsi_cmnd * scp, void (*CompletionRoutine) (struct scsi_cmnd *));
int bc_msg_send_srb(struct scsi_cmnd * scp);
int bc_msg_init(bc_softstate_t *state);
void bc_msg_shutdown( bc_softstate_t *statep);
int bc_reboot_notify ( struct notifier_block *nb, ulong event, void *buf);
int bc_msg_stats(char *buf, int buf_size);
int bc_mop_stats(char *buf, int buf_size);


struct bc_pci_bar {
	struct {
	    bc_uint32_t low;
	    bc_uint32_t high;
	} addr;
	bc_uint32_t len;
	bc_uint32_t flags;
};

bc_version_t       bc_versions[] = {
    {bc_frodo_init, bc_frodo_start, bc_frodo_shutdown, bc_frodo_isr,
	"bcraid", "Broadcom", "BC4000", 8, 0, 5}, /* wildcat */
    {bc_elrond_init, bc_elrond_start, bc_elrond_shutdown, bc_elrond_isr,
	"bcraid", "Broadcom", "BC5050", 8, 0, 2}, /* elrond */
    {bc_frodo_init, bc_frodo_start, bc_frodo_shutdown, bc_frodo_isr,
	"bcraid", "Broadcom", "HT1000", 4, 0, 5}, /* HT1000 */		
};


static struct pci_device_id bcraid_id_tbl[] = {
	/* Original ID conflicts with lib_sata in the kernel release. */
	{0x1166, 0x0241, 0x182f, 0x0005, 0, 0, (unsigned long)&bc_versions[0] },
	/* New ID selectable in the bios for running software raid on kernels
	 * versions 2.6.9 and later.
	 */
	{0x182f, 0x0005, 0x182f, 0x0005, 0, 0, (unsigned long)&bc_versions[0] },	/* Wildcat */
#ifdef notdef
	{0x182f, 0x0009, 0x182f, 0x0009, 0, 0, (unsigned long)&bc_versions[1] },	/* Elrond */
#endif
	{0x182f, 0x000b, 0x182f, 0x000b, 0, 0, (unsigned long)&bc_versions[2] },	/* HT1000 */
	{0,}
};



typedef enum {
	state_idle = 0,
	state_busy,
	state_error,
	state_wait_for_completion,
	state_complete,
	state_invalid
} cmd_state;



typedef struct bc_proc_entry
{
    char *name;
    struct proc_dir_entry *proc_dir;
    int mode;
    int which;
} bc_proc_entry_t;

static bc_proc_entry_t bc_proc_dir_entries[] = {
#define BC_PROC_STATS 0
    { "stats", NULL, 0, BC_PROC_STATS},

#define BC_PROC_DEBUG 1
    { "debug", NULL, S_IRUGO|S_IWUSR, BC_PROC_DEBUG},

#define BC_PROC_VERSION 2
    { "version", NULL, 0, BC_PROC_VERSION},

};

typedef struct bc_bios_disk_parameters
{
	int heads;
	int sectors;
	int cylinders;
} bc_bios_disk_parameters_t;


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static struct notifier_block bc_notifier = {
    bc_reboot_notify,
    NULL,
    0
};
#endif


static Scsi_Host_Template driver_template = {
    module:			THIS_MODULE,
    name:			"bcraid",		/* Driver name			*/
    proc_name:                  "bcraid",               /* Name of the proc dir         */
#if 1 // LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) /* need to migrate to sysfs stuff */
    proc_info:			bc_proc_info,		/* ProcFS Info Func		*/
#endif
    proc_dir:			NULL,			/* /proc/scsi/bcraid entry	*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    detect:			bc_detect,		/* Detect Host Adapter		*/
    release:			bc_release,		/* Release Host Adapter		*/
#endif
    info:			bc_info,		/* Driver Info Function		*/
    ioctl:			bc_ioctl,		/* ioctl Interface		*/
//    command:			bc_command,		/* unqueued command		*/
    queuecommand:		bc_queue_cmd,		/* Queue Command Function	*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    abort:			NULL,			/* Abort Command Function	*/
    reset:			NULL,			/* Reset Command Function	*/
#endif
    bios_param:			bc_bios_params,		/* BIOS Disk Parameters		*/
    can_queue:			1,			/* Default initial value	*/
    this_id:			-1,			/* Default initial value	*/
    sg_tablesize:		1,			/* Default initial value	*/
    max_sectors:		128,			/* max xfer size of 64k		*/
    cmd_per_lun:		0,			/* Default initial value	*/
    present:			0,			/* Default initial value	*/
    unchecked_isa_dma:		0,			/* Default Initial Value	*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    use_new_eh_code:		1,			/* Default initial value	*/
#endif
    eh_abort_handler:		bc_eh_abort_cmd,	/* Abort Command 		*/
/*    eh_strategy_handler:	NULL,			/* Strategy Error Handler 	*/
    eh_device_reset_handler:	bc_eh_dev_reset,	/* Device Reset Handle		*/
    eh_bus_reset_handler:	bc_eh_bus_reset,	/* Bus Reset Handler		*/
    eh_host_reset_handler:	bc_eh_hba_reset,	/* Host reset Handler		*/
    use_clustering:		ENABLE_CLUSTERING,	/* Disable Clustering		*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    highmem_io:			0,			/* Enable HIGHMEM I/O 		*/
    select_queue_depths:        bc_select_queue_depths,
#else
    slave_configure:		bc_slave_cfg,
#endif

};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include "scsi_module.c"
#endif


/*
 *    bc_init_module()
 *
 *        One time initialization stuff, called from the module init function
 *        in 2.6 kernels, or from the scsi detect function in earlier kernels
 */
static void
bc_init_module(void)
{
	extern char	      *bc_ident;

	bc_printk (BC_NOTE, "Broadcom %s raid driver version %s build_number %s built %s\n",
		   BC_DRIVER_NAME, BC_DRIVER_VERSION, RC_BUILD_NUMBER, BC_DRIVER_BUILD_DATE);
	bc_printk(BC_NOTE, "%s %s\n", BC_DRIVER_NAME, bc_ident);

	/*
	 * enforce reasonable limits on module parameters
	 */
	if (cmd_q_depth < 0)
		cmd_q_depth = 1;

	if (cmd_q_depth > BC_MAX_CMD_Q_DEPTH)
		cmd_q_depth = BC_MAX_CMD_Q_DEPTH;

	if (tag_q_depth < 0)
		tag_q_depth = 1;

	if (tag_q_depth > cmd_q_depth)
		tag_q_depth = cmd_q_depth;

	if (max_xfer < 8)
		max_xfer = 8;

	bc_printk( BC_NOTE,"bcraid: cmd_q_depth %d tag_q_depth %d max_xfer %d\n",
			cmd_q_depth,
			tag_q_depth,
			max_xfer);

	bc_msg_level += debug;
	if (bc_msg_level < 0)
		bc_msg_level = 0;

	/* Initialize the global state variable */
	memset(&bc_state, 0, sizeof(bc_softstate_t));
	spin_lock_init(&bc_state.osic_lock);
	init_MUTEX_LOCKED(&bc_state.bc_timeout_sema);

}


/*
 * Takes into account 64-bit bars.
 * If linux changes how pci_resource_ works, this code will have to change
 */
void
bc_get_bar(struct pci_dev *dev, int which_bar, struct bc_pci_bar *result)
{
	int i, index;
	struct bc_pci_bar bar;

	memset(&bar, 0, sizeof(bar));

	for (i = 0, index = 0;  i <= which_bar; i++, index++) {
		bar.addr.low  = pci_resource_start(dev,index);
		bar.len       = pci_resource_len(dev,index);
		bar.flags     = pci_resource_flags(dev,index);

		if (bar.flags & 0x4) {
			index++;
			bar.addr.high =  pci_resource_start(dev,index);
		} else
			bar.addr.high = 0;
		bc_printk(BC_DEBUG, "bar %d addr 0x%x.%x len %d flags 0x%x\n",
			i, bar.addr.high, bar.addr.low, bar.len, bar.flags);
	}

	*result = bar;
}

/*
 *    bc_init_adapter()
 *
 *        Initialize a single adapter, returns 0 on success or a negative
 *        error code on failure.
 */
static int
bc_init_adapter(struct pci_dev *dev, const struct pci_device_id *id)
{
	bc_adapter_t          *adapter;
	bc_hw_info_t	      *hw;
	int                   i;
	struct bc_pci_bar     bar;

	bc_printk (BC_DEBUG, "bc_init_adapter: Matched %x/%x/%x/%x\n",
		   id->vendor, id->device, id->subvendor, id->subdevice);

	if (bc_state.num_hba == MAX_HBA) {
		bc_printk(BC_ERROR, "bc_init_adapter: Execceded maximum adapter count of %d",
			  MAX_HBA);
		return -ENODEV;
	}

	if (pci_enable_device (dev))
		return -ENODEV;

	adapter = kmalloc(sizeof (bc_adapter_t), GFP_KERNEL);
	if (adapter == NULL) {
		bc_printk(BC_ERROR, "bc_init_adapter: can't alloc memory\n");
		return -ENODEV;
	}
	memset(adapter, 0, sizeof (*adapter));

	adapter->pdev = dev;
	adapter->instance = bc_state.num_hba;
	adapter->version = (bc_version_t *)id->driver_data;
	adapter->name = adapter->version->device_name;
	adapter->hardware.adapter_number = bc_state.num_hba;
		     
	hw = &adapter->hardware;

	hw->pci_bus    = dev->bus->number;
	hw->pci_slot   = PCI_SLOT (dev->devfn);
	
	bc_get_bar(dev, adapter->version->which_bar, &bar);

	hw->phys_addr  = bar.addr.low;
	hw->mem_len    = bar.len;

	if (hw->phys_addr == 0) {  	/* No pci memory */
		bc_printk(BC_ERROR, 
			  "bcraid: %s %s with pci IDs %x/%x/%x/%x has no pci address space assigned\n",
			  adapter->version->vendor,
			  adapter->version->model,
			  id->vendor, id->device, id->subvendor, id->subdevice);

		bc_printk(BC_ERROR, "Check BIOS PCI settings\n");
		bc_shutdown_adapter(adapter);
		return -ENODEV;
	}

	/*
	 * make a copy of pci config space
	 */
	for (i = 0; i < PCI_CFG_SIZE; i++) {
		pci_read_config_byte( dev, i, &hw->pci_config_space[i] );
	}

	/*
	 * set dma_mask to 64 bit capabilities.
	 * if that fails, try 32 bit dma
	 */
	if (pci_set_dma_mask (dev, ~0x0) != 0) { 
		if (pci_set_dma_mask (dev, 0xffffffff) != 0) {
			bc_printk (BC_ERROR, "bcraid: bc_init_adapter failed to set dma mask\n");
			bc_shutdown_adapter(adapter);
			return -ENODEV;
		}
	}


	/*
	 * map in the adapter MMIO space
	 */
	adapter->hardware.vaddr = (void *)ioremap(hw->phys_addr, hw->mem_len);
	if (adapter->hardware.vaddr == NULL) {
		bc_printk(BC_ERROR, "bc_init_adapter: can't map %s adapter %d at 0x%lx\n",
		          adapter->name, adapter->instance, hw->phys_addr);
		bc_shutdown_adapter(adapter);
		return -ENODEV;
	}

	init_MUTEX_LOCKED(&adapter->timer_sema);

	/* Call initialization routine */
	bc_printk (BC_DEBUG, "bc_init_adapter: Initializing hardware...\n");
	if ((*adapter->version->init_func)(adapter) != 0) {
		/* Device initialization failed */
		bc_printk (BC_ERROR, "bcraid:%d Device initialization failed\n",
			   adapter->instance);
		bc_shutdown_adapter(adapter);
		return -ENODEV;
	} 


	/* attach the interrupt */
	if ( request_irq ( dev->irq,
			   adapter->version->isr_func,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
			   IRQF_SHARED,
#else
			   SA_INTERRUPT | SA_SHIRQ,
#endif
			   adapter->version->model,
			   (void *)adapter ) < 0) {
		bc_printk (BC_ERROR, "bcraid:%d request_irq failed\n",
			   adapter->instance);
		bc_shutdown_adapter(adapter);
		return -ENODEV;
	}
	hw->irq = dev->irq;

	if ((*adapter->version->start_func)(adapter) != 0) {
		/* Device initialization failed */
		bc_printk (BC_ERROR, "bcraid:%d Device start failed\n",
			   adapter->instance);
		bc_shutdown_adapter(adapter);
		return -ENODEV;
	}

	pci_set_drvdata(dev, adapter);
	bc_dev[bc_state.num_hba++] = adapter;

	bc_printk(BC_NOTE,"bcraid card %d: %s %s\n",
		  adapter->instance,
		  adapter->version->vendor,
		  adapter->version->model);

	return 0;

}


/*
 *    bc_init_host()
 *
 *        Final stage of initialization is called after the last adapter
 *        has been initialized. and the core started. finializes the host
 *        parameters.
 */
static void
bc_init_host(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	int error;
#endif
	struct Scsi_Host	*host_ptr;

	/* start the raid the core. */
	bc_msg_init(&bc_state);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (register_reboot_notifier(&bc_notifier)) {
		bc_printk(BC_ERROR,"bc_detect: register_reboot_notifer failed\n");
	}

	/*
	 * Since we have only one virtual adapter spanning multiple physical adapters,
	 * we just have one host structure.
	 */
	bc_printk(BC_DEBUG,"bc_init_host: calling scsi_register\n");
	host_ptr = scsi_register (&driver_template, 32);
	if (!host_ptr)
		return;             /* need cleanup... */

	if (highmem_io)
		host_ptr->highmem_io = 1;

#else
	bc_printk(BC_DEBUG,"bc_init_host: calling scsi_host_alloc\n");
	host_ptr = scsi_host_alloc (&driver_template, 32);
	if (!host_ptr)
		return;             /* need cleanup... */
#endif

	if (bc_state.state & USE_OSIC) {
		host_ptr->max_id = 17;
		host_ptr->max_channel = 0;
		host_ptr->can_queue = cmd_q_depth;
		host_ptr->sg_tablesize = 32;	
		host_ptr->max_sectors = max_xfer;
		host_ptr->cmd_per_lun = 1;	// untagged queue depth
	} else {
		host_ptr->max_id = 9;
		host_ptr->max_channel = bc_state.num_hba - 1;
		host_ptr->can_queue = 1;
		host_ptr->sg_tablesize = 1;	
		host_ptr->max_sectors = 128;
		host_ptr->cmd_per_lun = 1;	// untagged queue depth
	}

	host_ptr->max_lun = 1;

	host_ptr->irq = 0;
	host_ptr->base = 0;
	host_ptr->max_cmd_len = 16;

	host_ptr->unique_id = 0;
	host_ptr->this_id = -1;	 /* SCSI Id for the adapter itself */

	driver_template.present = 1;	/* one virtual adapter */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	error = scsi_add_host(host_ptr, NULL); 

	if (error) {
		bc_printk(BC_ERROR, "Failed to add scsi host");

		scsi_host_put(host_ptr);
		return;                 /* need more cleanup... */
	}

	bc_state.host_ptr = host_ptr;

	scsi_scan_host(host_ptr);

#else

	bc_state.host_ptr = host_ptr;

#endif

	bc_printk(BC_DEBUG,"bc_init_host: completed\n");        

}


/*
 *    bcraid_probe_one()
 *
 *        Called once by the 2.6 kernel for each adapter discovered.
 *        Returns 0 if successful or a negative error code if the 
 *        device couldn't be initalized.
 *
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int
bcraid_probe_one(struct pci_dev *dev, const struct pci_device_id *id)
{
	int  ret = -ENODEV;

	struct pci_device_id  *probe_id;
	struct pci_dev *probe_dev;

	/* Count the number adapters on the bus. */
	if (!adapter_count) {
		for (probe_id = &bcraid_id_tbl[0]; probe_id->vendor != 0; probe_id++) {
			probe_dev = NULL;
			while ((probe_dev = pci_get_subsys(probe_id->vendor, probe_id->device,
							    probe_id->subvendor, probe_id->subdevice,
							    probe_dev))) {
				/* Found an adapter */
				adapter_count++;
			}
		}
	}

	ret = bc_init_adapter(dev, id);

	if (ret < 0) {
		return ret;
	}

	/* Finished with all of the adapters, start the core and 
	 * initialize the scsi host
	*/
	if (adapter_count && adapter_count == bc_state.num_hba) {
		bc_init_host();
	}

	return ret;
}
#endif


/*
 *	bc_detect()
 *
 *		Probe for  Adapters
 * 		initialize, register, and report the 
 *		configuration of each adapter found.
 *
 *	Preconditions:
 *	Postconditions:
 *		- Returns the number of adapters successfully initialized and 
 *		registered.
 *		- Initialize all data necessary for this particular SCSI driver.
 *	Notes:
 *		The detect routine must not call any of the mid level functions 
 *		to queue commands because things are not guaranteed to be set 
 *		up yet. The detect routine can send commands to the host adapter 
 *		as long as the program control will not be passed to scsi.c in 
 *		the processing of the command. Note especially that 
 *		scsi_malloc/scsi_free must not be called.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int
bc_detect (Scsi_Host_Template *template)
{
	struct pci_dev        *dev = NULL, *saved_dev = NULL;
	struct pci_device_id  *id;

	bc_init_module();

	PUT_IO_REQUEST_LOCK;

	/*
	 *  pci_find_device traverses the pci_devices linked list for devices
	 *  with matching vendor and Device ids.
	 */
	for (id = &bcraid_id_tbl[0]; id->vendor != 0; id++) {
		dev = NULL;	// start from beginning of list
		while ((dev = pci_find_subsys(id->vendor, id->device,
					      id->subvendor, id->subdevice, dev))) {

			/* Found an adapter to initialize */
			if (bc_init_adapter(dev, id) >= 0)
				saved_dev = dev;
		}
	}

	if (bc_state.num_hba == 0) {
		bc_printk(BC_ERROR, "bcraid: Warning: no controllers found\n");
		GET_IO_REQUEST_LOCK;
		return(0);
	}

	/* Start the core and register the scsi host */
	bc_init_host();

	if (bc_state.host_ptr)
		bc_state.host_ptr->pci_dev = saved_dev; /* bit of a kludge, but all our cards are the same */
	/* else cleanup and fail... */

	bc_printk(BC_DEBUG2,"acquiring io_request_lock lock\n");
	GET_IO_REQUEST_LOCK;

	bc_printk(BC_DEBUG2,"bc_detect done\n");

	return (1);
}
#endif


/*
 *  bc_shutdown_host()
 *
 *      Removes all of the proc filesystem entries and shuts down the core.
 *
 */
void
bc_shutdown_host(struct Scsi_Host *host_ptr)
{
	bc_proc_entry_t *pe;
	int i;

	if ((bc_state.state & USE_OSIC) == 0)
		return;

	bc_printk(BC_DEBUG, "bc_shutdown_host\n" );

	for (i = 0, pe = bc_proc_dir_entries;
             i < sizeof (bc_proc_dir_entries) / sizeof (bc_proc_entry_t); ++i, ++pe) {
		if (!pe->proc_dir)
			bc_printk(BC_ERROR, "bc_release: no entry for /proc/scsi/bcraid/%s\n", pe->name);
		else {
			remove_proc_entry(pe->name, driver_template.proc_dir);
			pe->proc_dir = NULL;
		}
	}

	bc_msg_shutdown(&bc_state);

}


/*
 *  bc_shutdown_adapter()
 *
 *    Shuts down an adapter card and frees all of its' resources.
 *
 */
void
bc_shutdown_adapter(bc_adapter_t *adapter)
{

	if (adapter == (bc_adapter_t *) 0) {
		bc_printk(BC_DEBUG, "bc_shutdown_adapter: NULL adapter\n");
		return;
	}

	bc_printk(BC_DEBUG, "bc_shutdown_adapter: adapter %d addr 0x%p\n",  adapter->instance, adapter);

	/* Call the adapter specific shutdown function. */
	if (adapter->version->shutdown_func)
		(*adapter->version->shutdown_func)(adapter);

	bc_printk(BC_DEBUG, "bc_shutdown_adapter: free_irq\n");
	if (adapter->hardware.irq)
		free_irq(adapter->hardware.irq, adapter);

	bc_printk(BC_DEBUG, "bc_shutdown_adapter: unmap MMIO space 0x%p\n", adapter->hardware.vaddr);
	if (adapter->hardware.vaddr)
		iounmap((void *)adapter->hardware.vaddr);

	bc_printk(BC_DEBUG, "bc_shutdown_adapter: unmap SRAM space 0x%p\n", adapter->hardware.sram_vaddr);
	if (adapter->hardware.sram_vaddr)
		iounmap((void *)adapter->hardware.sram_vaddr);

	bc_printk(BC_DEBUG, "bc_shutdown_adapter: free private_mem 0x%p\n",
		  adapter->private_mem.vaddr);
	if (adapter->private_mem.vaddr)  {
		pci_free_consistent(adapter->pdev,
			            bc_state.memsize_per_controller,
				    adapter->private_mem.vaddr,
				    adapter->private_mem.dma_address);
				
	}

	/* pci_disable_device(adapter->pdev); */

	bc_printk(BC_DEBUG, "bc_shutdown_adapter: free adapter 0x%p\n", adapter);
	kfree(adapter);
}


/*
 *      bcraid_shutdown_one()
 *
 *              Dereferences the adapter structure from the device and calls
 *              the interal shutdown routine. Called by the PCI driver layer
 *              once for each adapter card.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
void
bcraid_shutdown_one(struct device *dev)
{
	bc_adapter_t  *adapter = pci_get_drvdata(to_pci_dev(dev));

	if (bc_state.state & USE_OSIC)
		bc_shutdown_host(bc_state.host_ptr);

	bc_shutdown_adapter(adapter);
}
#endif


/*
 *	bc_release()
 *
 *		Release all resources previously acquired 
 *		and unregister the adapter.
 *		Need to stop all timers
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int
bc_release (struct Scsi_Host *host_ptr)
{
	int i;

	bc_printk(BC_DEBUG, "bc_release\n" );

	unregister_reboot_notifier(&bc_notifier);

	bc_printk(BC_DEBUG, "bc_release: unregister host 0x%p\n", host_ptr);
	if (host_ptr) {
		/* Shutdown the scsi host layer. */
		scsi_unregister (host_ptr);

		bc_shutdown_host(host_ptr);
	}

	for (i =0; i < bc_state.num_hba; i++) {
	  bc_shutdown_adapter(bc_dev[i]);
	}

	bc_printk(BC_INFO2, "bc_release: Done\n");

	return (0);	
}

int 
bc_reboot_notify ( struct notifier_block *nb, ulong event, void *buf)
{
	bc_printk( BC_DEBUG, "bc_reboot_notify\n" );

	bc_release(bc_state.host_ptr);

	return(NOTIFY_DONE);
}
#endif

/*
 * Pause all of the SATA ports and wait for acknowledgement. Returns
 * zero if the ports were successfully paused or a negative value
 * if the the operation could not complete on one or more ports.
 */
static int
bc_frodo_pause_ports(bc_adapter_t *adapter)
{
	rc_sw_mmio_port_t   *ports;
	int                 i, reset;
	unsigned long	    start;
	uint32_t	    status;
        int                 ret=0;

	ports = (rc_sw_mmio_port_t *)adapter->hardware.vaddr;

	/* Pause all the ports */
	for (i = 0; i < adapter->version->num_ports; i++) {
		(ports+i)->PortQDMAControlRegister = RC_SW_QCR_QDMA_PAUSE | RC_SW_QCR_QDMA_ENABLE;
	}

	/* wait for Pause acknowledgement */
	for (i = 0; i < adapter->version->num_ports; i++) {
		start = jiffies;
		reset = 0;
		do {
			status = (ports+i)->PortQDMAStatusRegister;
			if (status & RC_SW_QCR_QDMA_PAUSE_ACK) {
				bc_printk(BC_DEBUG,
					  "bc_frodo_pause_ports: port %d PAUSE: PortStatus 0x%x time: %lu \n",
					  i, status, jiffies - start);
				reset = 1;
				break;
			};
			bc_timeout(2);
		} while  ((jiffies - start) < (HZ>>2));

		if (reset == 0) {
			bc_printk(BC_DEBUG,
				  "bc_frodo_pause_ports: port %d failed to PAUSE after 1/2 seconds status: 0x%x\n",
				  i, status);
			ret = -1;
		}
	}

	return ret;
}


/*
 * disable HW interrupts on all ports on an adapter
 */
static inline void
bc_frodo_disable_irq(bc_adapter_t *adapter)
{
	rc_sw_mmio_port_t   *ports;
	int                 i;

	ports = (rc_sw_mmio_port_t *)adapter->hardware.vaddr;

	for (i = 0; i < adapter->version->num_ports; i++) {
		(ports+i)->PortQDMAInterruptMaskRegister = 0;
	}
}


/*
 * Initialize the adapter by pausing all of the ports and disabling the interrupts.
 */
int
bc_frodo_init(bc_adapter_t *adapter)
{
	bc_printk(BC_DEBUG, "bc_frodo_init: MMIO base address 0x%p\n", adapter->hardware.vaddr);
	bc_frodo_pause_ports(adapter);
	bc_frodo_disable_irq(adapter);
	return 0;
}


int
bc_frodo_start(bc_adapter_t *adapter)
{
	return(0);
}

/*
 * Disable all HW interrupts and turn PHYs off.
 */
int
bc_frodo_shutdown(bc_adapter_t *adapter)
{
	rc_sw_mmio_port_t	*ports[MAX_PORTS_PER_HA];
	unsigned char		*base;
	uint32_t		status;
	int 			i;


	base = adapter->hardware.vaddr; 

	for (i = 0; i < adapter->version->num_ports; i++) {
		ports[i] = (rc_sw_mmio_port_t *)base;
		base += RC_SW_MMIO_PORT_OFFSET;
	}

	/*
	 * Clear errors.
	 * Disable Interrupts
	 * Turn off PHYs
	 */
	for (i = 0; i < adapter->version->num_ports; i++) {
		ports[i]->PortAlternateStatusDeviceControl = IE_DISABLE;
		status = ports[i]->PortStatusCommand;
		ports[i]->PortSATAErrorRegister         = RC_SW_SCR1_CLEAR_ALL;
		ports[i]->SATAInterruptMaskRegister     = 0;
		ports[i]->PortSATAControlRegister       = RC_SW_SCR2_PHY_OFF;
		ports[i]->PortQDMAStatusRegister        = RC_SW_QSR_CLEAR_ALL;
		ports[i]->PortQDMAInterruptMaskRegister = 0;
		bc_printk(BC_DEBUG, "bc_frodo_shutdown: port %d:  interupts, PHY off Status: 0x%x\n", i, status);
	}

        bc_frodo_pause_ports(adapter);

	return(0);
}

int bc_spur_int_cnt = 0;
void bc_msg_isr( bc_adapter_t *adapter);


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
irqreturn_t
bc_frodo_isr(int irq, void *arg)
#else
irqreturn_t
bc_frodo_isr(int irq, void *arg, struct pt_regs *regs)
#endif
{
	bc_adapter_t	*adapter;
	bc_softstate_t	*statep;
	static int count = 0;
	rc_sw_mmio_port_t	*base;
	uint32_t		status;
	int			i;

	statep = &bc_state;
	adapter = (bc_adapter_t *)arg;

	bc_frodo_disable_irq(adapter);

	if (statep->state & USE_OSIC) {
		bc_msg_isr(adapter);
//		bc_printk(BC_DEBUG_ISR,"bc_frodo_isr exit, OSIC\n");
		return IRQ_HANDLED;
	}

	if ((count % 1000) == 0)
		bc_printk(BC_DEBUG,"bc_frodo_isr  NO OSIC %d jiffies: %lu\n", count, jiffies);
	count++;

	/*
	 * clear and disable all interrupts
	 */
	for (i = 0; i < adapter->version->num_ports; i++) {

		base = (rc_sw_mmio_port_t *)adapter->hardware.vaddr + i;

		base->PortAlternateStatusDeviceControl = IE_DISABLE;
		status = base->PortStatusCommand;
		if (status & STATUS_ERROR)
			bc_printk(BC_DEBUG_ISR, "bc_frodo_isr: port %d PortStatus 0x%x \n", i, status);

		/*
		 * This register will report a PHY change on shutdown
		 */
		status = base->PortSATAErrorRegister;
		base->PortSATAErrorRegister         = RC_SW_SCR1_CLEAR_ALL;
		if (status) 
			bc_printk(BC_DEBUG_ISR, "bc_frodo_isr: port %d SATAError 0x%x \n", i, status);
		base->SATAInterruptMaskRegister = 0;

		//ports[i]->PortQDMAControlRegister = RC_SW_QCR_QDMA_DISABLE;
		base->PortQDMAInterruptMaskRegister = 0;
		status = base->PortQDMAStatusRegister;
		//if (status)
		//	bc_printk(BC_DEBUG_ISR, "bc_frodo_isr: port %d QDMA status 0x%x \n", i, status);
	}
	return IRQ_HANDLED;

}


/*
 * Pause all of the SATA ports and wait for acknowledgement. Returns
 * zero if the ports were successfully paused or a negative value
 * if the the operation could not complete on one or more ports.
 */
static int
bc_elrond_pause_ports(bc_adapter_t *adapter)
{
	bc_printk(BC_WARN, "bc_elrond_pause_ports: \n");

	return 0;
}


/*
 * disable HW interrupts on all ports on an elrond based adapter
 */
static inline void
bc_elrond_disable_irq(bc_adapter_t *adapter)
{
	bc_uint8_t 	*base;
	bc_uint32_t	*intr_mask_reg;
	int		i;

	base = (bc_uint8_t *)adapter->hardware.vaddr;

	for (i = 0; i < adapter->version->num_ports; i++) {
		intr_mask_reg = (bc_uint32_t *) (base + 0xbc);	/* JWF Magic Number */
		*intr_mask_reg = 0;

		base += RC_SW_MMIO_PORT_OFFSET;
	}

	bc_printk(BC_DEBUG_ISR, "bc_elrond_disable_irq \n");
}


/*
 * Initialize the adapter by pausing all of the ports and disabling the interrupts.
 */
int
bc_elrond_init(bc_adapter_t *adapter)
{
	struct pci_dev       *dev;
	struct bc_pci_bar     bar;
	bc_hw_info_t	      *hw;

	/*
	 * for elrond based products, we also have to map in the SRAM (bar 0)
	 * and the IO space (bar 1)
	 * Note that BAR 0 is a 64-bit bar, but we're only using the low 32 bits.
	 * If this can ever be more that a 32-bit address, the issue will have to be fixed.
	 */

	dev = adapter->pdev;
	hw  = &adapter->hardware;

	bc_get_bar(dev, 0, &bar);
	hw->sram_paddr =  bar.addr.low;
	hw->sram_len  = bar.len;

	bc_get_bar(dev, 1, &bar);
	hw->io_port      = bar.addr.low;
	hw->io_port_len  = bar.len;

	/* 
	 * map in the adapter SRAM space
	 */
	adapter->hardware.sram_vaddr = (void *)ioremap(hw->sram_paddr, hw->sram_len);
	if (adapter->hardware.sram_vaddr == NULL) {
		bc_printk(BC_ERROR, "bc_init_adapter: can't map %s adapter %d  sram at 0x%lx\n",
			  adapter->name, adapter->instance, hw->sram_paddr);
		bc_shutdown_adapter(adapter);
		return -ENODEV;
	}

	bc_printk(BC_DEBUG, "bc_elrond_init: MMIO base address 0x%p\n", adapter->hardware.vaddr);
	bc_printk(BC_DEBUG, "bc_elrond_init: SRAM base address 0x%p\n", adapter->hardware.sram_vaddr);
	bc_printk(BC_DEBUG, "bc_elrond_init: IO Port address 0x%x\n",   adapter->hardware.io_port);
	bc_elrond_pause_ports(adapter);
	bc_elrond_disable_irq(adapter);
	return 0;
}


int
bc_elrond_start(bc_adapter_t *adapter)
{
	return(0);
}

/*
 * Disable all HW interrupts and turn PHYs off.
 */
int
bc_elrond_shutdown(bc_adapter_t *adapter)
{

	bc_printk(BC_WARN, "bc_elrond_shutdown\n");
        bc_elrond_pause_ports(adapter);
	bc_elrond_disable_irq(adapter);

	return(0);
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
irqreturn_t bc_elrond_isr(int irq, void *arg)
#else
irqreturn_t bc_elrond_isr(int irq, void *arg, struct pt_regs *regs)
#endif
{
	bc_adapter_t	*adapter;
	bc_softstate_t	*statep;
	static int count = 0;

	statep = &bc_state;
	adapter = (bc_adapter_t *)arg;

	bc_elrond_disable_irq(adapter);

	if (statep->state & USE_OSIC) {
		bc_msg_isr(adapter);
//		bc_printk(BC_DEBUG_ISR,"bc_elrond_isr exit, OSIC\n");
		return IRQ_HANDLED;
	}

	if ((count % 1000) == 0)
		bc_printk(BC_DEBUG,"bc_elrond_isr  NO OSIC %d jiffies: %lu\n", count, jiffies);
	count++;

	return IRQ_HANDLED;

}


/*
 *	bc_queue_cmd()
 *
 *		Queues a SCSI command
 */
int
bc_queue_cmd (struct scsi_cmnd * scp, void (*CompletionRoutine) (struct scsi_cmnd *))
{
	int result;
	scp->scsi_done = CompletionRoutine;
	result = 0;


//#define FAIL_ALL_IO 0

#ifdef FAIL_ALL_IO
	scp->result = DID_NO_CONNECT << 16;
	scp->scsi_done(scp);
	return(-1);
#endif

//	bc_printk(BC_DEBUG2,"bc_queue_cmd B/T/L %d/%d/%d\n",
//		scp->device->channel,
//		scp->device->id,
//		scp->device->lun);


	if ((bc_config_debug == 1) && (scp->device->id == 16)) 
		bc_dump_scp(scp);

	result = bc_msg_send_srb(scp);
	return(result);

}

/*
 *	bc_done()
 *
 *		Callback function for a non-queued command.
 *
 *	Postconditions
 *		Sets g_single_command done to TRUE
 */
void
bc_done (struct scsi_cmnd * scp)
{
	bc_single_cmd_done = 1;
}

/*
 *	bc_command()
 *	
 *		Accepts a single command for execution by the associated Host Adapter.
 *
 *	Postconditions
 *		Returns an int where:
 *		Byte 0 = SCSI status code
 *		Byte 1 = SCSI 1 byte message
 *		Byte 2 = host error return
 *		Byte 3 = mid level error return
 */
int
bc_command (struct scsi_cmnd * scp)
{
	bc_printk(BC_ERROR,"bc_command: driver misconfigured: is queueing off?\n");
	bc_printk(BC_DEBUG,"bc_command B/T/L %d/%d/%d\n",
		  scp->device->channel, scp->device->id, scp->device->lun);
	scp->result = DID_NO_CONNECT << 16;

	return(-1);
}

/*
 *	bc_abort_command()
 *
 *		Abort command if possible.
 */
int
bc_eh_abort_cmd (struct scsi_cmnd * scp)
{
	bc_srb_t *srb;

	bc_printk (BC_ERROR, "bc_eh_abort_cmd: scp: 0x%p bus %d target %d\n",
		   scp, scp->device->channel, scp->device->id);
//	bc_config_debug = 1;

	srb = (bc_srb_t *)scp->SCp.ptr;
	if (srb != NULL) {
	    bc_printk(BC_DEBUG, "\tsrb: 0x%p seq_num %d function %x status %x flags %x b/t/l %d/%d/%d\n",
		       srb, srb->seq_num, srb->function, srb->status, srb->flags, srb->bus, srb->target, srb->lun);
	    srb->scsi_context = NULL;
	    scp->SCp.ptr = NULL;
	} else
	    bc_printk(BC_DEBUG, "\tsrb: NULL\n");
    
	bc_dump_scp( scp );

	return (SUCCESS);
}

int
bc_eh_dev_reset (struct scsi_cmnd *scp)
{

	bc_printk (BC_ERROR, "bc_eh_dev_reset: scp 0x%p bus %d target %d\n",
		   scp, scp->device->channel, scp->device->id);
	bc_config_debug = 1;
	bc_dump_scp( scp );

	return (SUCCESS);
}

int
bc_eh_bus_reset (struct scsi_cmnd *scp)
{
	bc_printk (BC_ERROR, "bc_eh_bus_reset: scp 0x%p bus %d target %d\n",
		   scp, scp->device->channel, scp->device->id);
	bc_config_debug = 1;
	bc_dump_scp( scp );

	return (SUCCESS);
}

int
bc_eh_hba_reset (struct scsi_cmnd *scp)
{
	bc_printk (BC_ERROR, "bc_eh_hba_reset: scp 0x%p bus %d target %d\n",
		   scp, scp->device->channel, scp->device->id);
	bc_dump_scp( scp );

	return (SUCCESS);
}


/*
 *	bc_driver_info()
 *
 *		Returns the host adapter name
 *		Also creates any additional /proc/scsi/bcraid/... entries that we need.
 *		This is done here instead of in bc_detect() or bc*init() because those are
 *		all called before scsi_register_host() creates /proc/scsi/bcraid.
 */
const char *
bc_info (struct Scsi_Host *host_ptr)
{
	static char	*name = "Broadcom RAIDCore\n";
	bc_proc_entry_t *pe;
	struct proc_dir_entry *pde;
	int i;
 

	if (!driver_template.proc_dir) /* scsi_register_host() should have set this up by now */
		bc_printk(BC_ERROR, "bc_info: no entry for /proc/scsi/bcraid yet\n");
	else
		for (i = 0, pe = bc_proc_dir_entries;
		     i < sizeof (bc_proc_dir_entries) / sizeof (bc_proc_entry_t); ++i, ++pe) {
			if (!(pde = create_proc_entry(pe->name, pe->mode, driver_template.proc_dir)))
				bc_printk(BC_ERROR, "bc_info: can't create entry for /proc/scsi/bcraid/%s\n", pe->name);
			else {
				pe->proc_dir = pde;
				pde->read_proc = bc_proc_read;
				pde->write_proc = bc_proc_write;
				pde->data = (void *)pe;
			}
		}

	return(name);
}


/*
 *	bc_bios_params;
 *
 *		Return the Heads/Sectors/Cylinders BIOS Disk Parameters for Disk.  
 *		The default disk geometry is 64 Heads, 32 Sectors, and the appropriate 
 *		number of Cylinders so as not to exceed drive capacity.  In order for 
 *		disks equal to or larger than 1 GB to be addressable by the BIOS
 *		without exceeding the BIOS limitation of 1024 Cylinders, Extended 
 *		Translation should be enabled.   With Extended Translation enabled, 
 *		drives between 1 GB inclusive and 2 GB exclusive are given a disk 
 *		geometry of 128 Heads and 32 Sectors, and drives above 2 GB inclusive 
 *		are given a disk geometry of 255 Heads and 63 Sectors.  However, if 
 *		the BIOS detects that the Extended Translation setting does not match 
 *		the geometry in the partition table, then the translation inferred 
 *		from the partition table will be used by the BIOS, and a warning may 
 *		be displayed.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
int
bc_bios_params (struct scsi_device *sdev,
		struct block_device *bdev,
		sector_t capacity,
		int geom[])
#else
int
bc_bios_params (Scsi_Disk *disk, kdev_t dev, int *geom)
#endif
{
	bc_bios_disk_parameters_t *param;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned capacity = disk->capacity;
#endif

	param = (bc_bios_disk_parameters_t *) geom;

	bc_printk (BC_DEBUG, "bc_bios_disk_parameters\n");

	// Assuming extended translation is enabled - #REVISIT#
	if (capacity >= 2 * 1024 * 1024)	// 1 GB in 512 byte Sectors
	{
		if (capacity >= 4 * 1024 * 1024)	// 2 GB in 512 byte sectors
		{
			param->heads = 255;
			param->sectors = 63;
		} else {
			param->heads = 128;
			param->sectors = 32;
		}
	} else {
		param->heads = 64;
		param->sectors = 32;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
     sector_div(capacity, param->heads * param->sectors);
     param->cylinders = capacity;
#else
	param->cylinders = capacity / (param->heads * param->sectors);
#endif

	return (0);
}

/*
 *	bc_select_queue_depths()
 *
 *		Selects queue depths for each Target Device based on the host adapter's
 *		total capacity and the queue depth supported by the target Device.
 *		A queue depth of one automatically disables tagged queueing.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
void
bc_select_queue_depths (struct Scsi_Host *host_ptr,
			 struct scsi_device * scsi_device_ptr)
{
	struct scsi_device        *device_ptr;

	bc_printk (BC_NOTE, "bc_select_queue_depths\n");
	bc_printk (BC_NOTE, "Device #   Q Depth   Online\n");
	bc_printk (BC_NOTE, "---------------------------\n");
	for (device_ptr = scsi_device_ptr; device_ptr != NULL; device_ptr = device_ptr->next)
		if (device_ptr->host == host_ptr) {
			device_ptr->queue_depth = tag_q_depth;
			bc_printk (BC_NOTE, "  %2d         %d        %d\n",
				   device_ptr->id, device_ptr->queue_depth, device_ptr->online);
		}
}
#else
static int
bc_slave_cfg(struct scsi_device *sdev)
{
        if (sdev->tagged_supported)
                scsi_adjust_queue_depth(sdev, MSG_ORDERED_TAG, tag_q_depth);
        else
                scsi_adjust_queue_depth(sdev, 0, 1);
        return 0;
}
#endif


int
bc_ioctl (struct scsi_device * scsi_dev_ptr,
	   int cmd,
	   void *arg)
{
	char direction = 'w';

	if (_IOC_DIR(cmd) & _IOC_READ)
		direction = 'r';

	bc_printk (BC_DEBUG, "bc_ioctl: type %i nr %i dir %c\n", _IOC_TYPE(cmd), _IOC_NR(cmd), direction);
	return(-ENOTTY);
}




/*
 *	bc_proc_info()
 *
 *		Implement /proc/scsi/<drivername>/<n>.  Called from scsi layer.
 *		Used to export driver statistics and other infos to the world outside 
 *		the kernel using the proc file system. Also provides an interface to
 *		feed the driver with information.
 *
 *	Postconditions
 *		For reads
 *			- if Offset > 0 return 0
 *			- if Offset == 0 Write data to ProcBuffer and set the Startptr to
 *			beginning of ProcBuffer, return the number of characters written.
 *		For writes
 *			- writes currently not supported, return 0
 * Parameters:
 *	  struct Scsi_Host *shost,	which host we are after (2.6+)
 *    char *buf,	read/write buffer
 *    char **start,	start of valid data in the buffer
 *    off_t offset,	Offset from the beginning of the imaginary file 
 *    int buf_size,	bytes available
 *    int host_num,	SCSI host number (2.4)
 *    int write		direction of dataflow: TRUE for writes, FALSE for reads      
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) /* changed between 2.4.26 and 2.6.5 */
int
bc_proc_info (                         char *buf, char **start, off_t offset, int buf_size, int host_num, int write)
{
#else
int
bc_proc_info (struct Scsi_Host *shost, char *buf, char **start, off_t offset, int buf_size,               int write)
{
	int host_num = shost->host_no;
#endif
	int 	len, cnt;
	char	*cp;
	//bc_printk (BC_DEBUG, "bc_proc_directory_info, buf size %d offset %d\n", buf_size, offset);

	if ((write) || (offset > 0))
		return (0);

	*start = buf;
	cp = buf;
	len = 0;
	cnt = 0;

	len = snprintf (cp, buf_size,
		"Broadcom Controller %s V%s build number %s built on %s  %d\n\n",
		BC_DRIVER_NAME,
		BC_DRIVER_VERSION,
		RC_BUILD_NUMBER,
		BC_DRIVER_BUILD_DATE,
		 host_num);

	cp  += len;
	cnt += len;

	len = bc_msg_stats(cp, buf_size - cnt);
	cp  += len;
	cnt += len;

	len = bc_mop_stats(cp, buf_size - cnt);
	cnt += len;
	return (cnt);
} /* } */


/*
 *	bc_proc_read()
 *
 *		Implement /proc/scsi/<drivername>/... (everything except <host_no>)
 *		Called from the generic /proc filesystem layer.
 *		Used to export driver statistics and other info to the world outside 
 *		the kernel using the proc file system.
 *
 *	Postconditions
 *	    normal:
 *		- if offset > 0 return 0
 *		- if offset == 0 Write data to buf, set *start to
 *		  beginning of buf, set *peof = 1, return the number of characters written.
 *	    offset is "entry number":
 *		write data for the offset'th entry to buf, set *start to number of records returned,
 *		set *peof = 1 if this is the last entry, and return the number of characters written.
 */

static int
bc_proc_read (char *buf,		// read buffer (in kernel space)
	      char **start,		// start of valid data in the buffer or index of element (see proc_file_read())
	      off_t offset,		// Offset from the beginning of the imaginary file 
	      int buf_size,		// bytes available
	      int *peof,		// set to 1 if we are at the end
	      void *data)		// proc_dirent->data
{
	int len, cnt;
	char *cp;

	bc_printk (BC_DEBUG, "bc_proc_read, start %p, offset %lu, buf size %d, data %p\n", 
		   *start, offset, buf_size, data);

	cp = buf;
	len = 0;
	cnt = 0;

	switch (((bc_proc_entry_t *)data)->which) {

	case BC_PROC_STATS:
		*peof = 1;
		if (offset)
			break;
		len = snprintf (cp, buf_size - cnt,
				"Broadcom Controller %s V%s build number %s built on %s proc device %d\n",
				BC_DRIVER_NAME,
				BC_DRIVER_VERSION,
				RC_BUILD_NUMBER,
				BC_DRIVER_BUILD_DATE,
				(int)((bc_proc_entry_t *)data - bc_proc_dir_entries) );
		cp  += len;
		cnt += len;
		*start = buf;
		break;

	case BC_PROC_VERSION:
		*peof = 1;
		if (offset)
			break;
		len = snprintf (cp, buf_size - cnt,
				"V%s %s\n",
				BC_DRIVER_VERSION,
				RC_BUILD_NUMBER);
		cp  += len;
		cnt += len;
		*start = buf;
		break;

	case BC_PROC_DEBUG:
		/* read debug message level, min, and max, scaled to match module parameter */
		*peof = 1;
		if (offset)
			break;

		len = snprintf(cp, buf_size - cnt, "%d %d %d\n",
			       bc_msg_level - BC_DEFAULT_ERR_LEVEL,
			       BC_PANIC - BC_DEFAULT_ERR_LEVEL,
			       BC_TAIL - 1 - BC_DEFAULT_ERR_LEVEL);
		cp  += len;
		cnt += len;
		*start = buf;
		break;

	default:
		bc_printk(BC_ERROR, "bc_proc_read, unexpected data %p\n", data);
		*peof=1;
		break;
	}

	return (cnt);
}


/*
 *	bc_proc_write()
 *
 *		Implement writes to /proc/scsi/<drivername>/... (everything except <host_no>)
 *		Called from the generic /proc filesystem layer.
 *		Used to feed the driver with information.
 *
 *	Postconditions
 *		- if offset > 0 return 0
 *		- if offset == 0 Write data to buf, set *start to
 *		  beginning of buf, set *peof = 1, return the number of characters written.
 */

static int
bc_proc_write(struct file *file,	// our file control structure
              const char *buf,      // the input buffer (user data address)
              unsigned long count,  // how many bytes we are being sent
              void *data)           // which /proc/scsi/bcraid/* file this is
{
	char page[10];
	int ret, tmp;

	/* should check file offset? */
	if (count >= sizeof (page))
		return -EOVERFLOW;
    
	if (!count)
		return 0;

	//if (!(page = (char *) __get_free_page(GFP_KERNEL)))
		//return -ENOMEM;
	if (copy_from_user(page, buf, count)) {
		//free_page((ulong) page);
		return -EFAULT;
	}

	ret = -EINVAL;
	switch (((bc_proc_entry_t *)data)->which) {
	case BC_PROC_STATS:
	case BC_PROC_VERSION:
		ret = count;
		break;

	case BC_PROC_DEBUG:
		/* set debug message level, scaled to match module parameter */
		page[count] = '\0';
		tmp = bc_msg_level - BC_DEFAULT_ERR_LEVEL;
		sscanf(page, "%i", &tmp);
		tmp += BC_DEFAULT_ERR_LEVEL;
		if (tmp >= 0 && tmp < BC_TAIL)
			bc_msg_level = tmp;
		ret = count;
		break;

	default:
		bc_printk(BC_ERROR, "pc_proc_write, unexpected proc device %p\n", data);
		ret = -EINVAL;
		break;
	}

	// free_page((ulong) page);
	return (ret);

}

static char bc_print_buf[1024];
void
bc_printk(
        int flag,
        const char *fmt,
        ...)
{
	static int newline_in_last_fmt = 1;
	va_list args;

	va_start(args, fmt);
	(void) vsnprintf(bc_print_buf, sizeof(bc_print_buf), fmt, args);
	va_end(args);

	if (flag == 0) {
		panic(bc_print_buf);
		return;
	}

	if (flag > bc_msg_level)
		return;

	flag = flag > 7 ? 7 : flag;

	if (newline_in_last_fmt)
		printk("<%d>%s", flag, bc_print_buf);
	else
		printk("%s", bc_print_buf);	

	newline_in_last_fmt = strchr(fmt, '\n') ? 1 : 0;
}

void
bc_timeout_done(unsigned long data)
{
	bc_softstate_t *state;

	state = (bc_softstate_t *)data;
	init_timer(&state->bc_timeout);
	up(&state->bc_timeout_sema);
}

void
bc_timeout( int to)
{
	bc_softstate_t *state;

	state = &bc_state;
	/*
	 * set up timeout
	 */

	init_timer(&state->bc_timeout);
	state->bc_timeout.expires = jiffies  + to;
	state->bc_timeout.data = (unsigned long)state;
	state->bc_timeout.function = bc_timeout_done;
	add_timer(&state->bc_timeout);
	down(&state->bc_timeout_sema);
}


void
bc_dump_scp( struct scsi_cmnd * scp )
{
	unsigned char cmd;
	bc_scb_t *scb;
	int bus, target, lun;
	unsigned int lba, sector_count;
	unsigned long long	lba16;
	int			i;
	dma_addr_t		dma_addr;
	struct scatterlist	*sg;

	if (scp == NULL)
		return;

	scb = (bc_scb_t *) scp->cmnd;
	cmd = scb->scsi6.opcode;

	bus = scp->device->channel;
	target = scp->device->id;
	lun = scp->device->lun;

//	bc_printk(BC_DEBUG, "bc_dump_scp: B/T/L %d/%d/%d\n", bus, target, lun);

	switch (cmd) {
	    case SS_INQUIR:
		bc_printk (BC_DEBUG, "INQUIRY command, BTL: %d/%d/%d\n", bus, target, lun);
		break;


	    case SS_RDCAP:
		bc_printk (BC_DEBUG, "READ CAPACITY command, BTL: %d/%d/%d\n", bus, target, lun);

		break;

	    case SS_MODESEN:
		bc_printk (BC_DEBUG, "MODE SENSE command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

		// These commands are all No-Ops
	    case SS_TEST:
		bc_printk (BC_DEBUG, "TEST UNIT READY command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SS_REQSEN:
		bc_printk (BC_DEBUG, "REQUEST SENSE command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SS_LOCK:
		// bc_printk (BC_DEBUG, "LOCK command, BTL: %d/%d/%d\n", bus, target, lun);

		if (scp->cmnd[4]) {
			bc_printk (BC_DEBUG, "LOCK command, BTL: %d/%d/%d\n", bus, target, lun);
		}
		else {
			bc_printk (BC_DEBUG, "UNLOCK command, BTL: %d/%d/%d\n", bus, target, lun);
		}
		break;


	    case SS_RESERV:
		bc_printk (BC_DEBUG, "RESERVE command\n");
		break;

	    case SS_RELES:
		bc_printk (BC_DEBUG, "RELEASE command\n");
		break;

	    case SS_REZERO:
		bc_printk (BC_DEBUG, "REZERO command\n");
		break;

	    case SS_REASGN:
		bc_printk (BC_DEBUG, "REASSIGN command\n");
		break;

	    case SS_SEEK:
		bc_printk (BC_DEBUG, "SEEK command\n");
		break;

	    case SS_ST_SP:
		bc_printk (BC_DEBUG, "START/STOP command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SS_READ:
		bc_printk (BC_DEBUG, "READ6 command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SM_READ:
		bc_printk (BC_DEBUG, "READ10 command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SM_READ_16:
		bc_printk (BC_DEBUG, "READ16 command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SS_WRITE:
		bc_printk (BC_DEBUG, "WRITE6 command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SM_WRITE:
		bc_printk (BC_DEBUG, "WRITE10 command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SM_WRITE_16:
		bc_printk (BC_DEBUG, "WRITE16 command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    case SM_REPORT_LUNS:
		bc_printk (BC_DEBUG, "REPORT_LUNS command, BTL: %d/%d/%d\n", bus, target, lun);
		break;

	    default:
		bc_printk(BC_DEBUG, "bc_dump_scp: unknown command 0x%x B/T/L %d/%d/%d\n", cmd,  bus, target, lun);
		    break;

	}

	switch (cmd) {
	    case SS_WRITE:
	    case SS_READ:
		lba = ((scb->scsi6.addr[0] & 0x1f) << 16) 
			    | (scb->scsi6.addr[1] << 8)
			    | scb->scsi6.addr[2];
		sector_count = scb->scsi6.len;
		bc_printk( BC_DEBUG, "lba: %d len %d\n", lba, sector_count);
		break;

		

	    case SM_WRITE:
	    case SM_READ:
		lba = (scb->scsi10.addr[0] << 24) 
			    | (scb->scsi10.addr[1] << 16)
			    | (scb->scsi10.addr[2] << 8)
			    | scb-> scsi10.addr[3];
		sector_count = (scb->scsi10.len[0] << 8) | scb->scsi10.len[1];
		bc_printk( BC_DEBUG, "lba: %d len %d\n", lba, sector_count);
		break;

	    case SM_WRITE_16:
	    case SM_READ_16:
		lba16 =       ((uint64_t)scb->scsi16.addr[0] << 56) 
			    | ((uint64_t)scb->scsi16.addr[1] << 48)
			    | ((uint64_t)scb->scsi16.addr[2] << 40)
			    | ((uint64_t)scb->scsi16.addr[3] << 32)
			    | ((uint64_t)scb->scsi16.addr[4] << 24)
			    | ((uint64_t)scb->scsi16.addr[5] << 16)
			    | ((uint64_t)scb->scsi16.addr[6] << 8)
			    |  (uint64_t)scb->scsi16.addr[7];

		sector_count = (scb->scsi16.len[0] << 8) | scb->scsi16.len[1];
		bc_printk( BC_DEBUG, "lba: %lld len %d\n", lba16, sector_count);
		break;
		
	    default:
		break;
	}

	bc_printk(BC_DEBUG, "    scp: 0x%p use_sg %d, buf 0x%p, len %d\n",
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
		scp, scsi_sg_count(scp), scsi_sglist(scp), scsi_bufflen(scp));
#else
		scp, scp->use_sg, scp->request_buffer, scp->request_bufflen);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	sg = (struct scatterlist *)  scsi_sglist(scp);
#else
	sg =  (struct scatterlist *) scp->request_buffer;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	for (i = 0; i < scsi_sg_count(scp); i++)  {
#else
	for (i = 0; i < scp->use_sg; i++)  {
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
		dma_addr = page_to_phys(sg_page(&(sg[i]))) + sg[i].offset;
#else
		dma_addr = page_to_phys(sg[i].page) + sg[i].offset;
#endif
#if defined(CONFIG_HIGHMEM64G) ||  defined(CONFIG_X86_64)
		bc_printk(BC_DEBUG, "    page: 0x%p  offset: 0x%x addr: 0x%16llx len %x\n",
#else
		bc_printk(BC_DEBUG, "    page: 0x%p  offset: 0x%x addr: 0x%8x len %x\n",
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
			 sg_page(&(sg[i])), sg[i].offset, dma_addr, sg[i].length);
#else
			 sg[i].page, sg[i].offset, dma_addr, sg[i].length);
#endif
	}
	bc_printk(BC_DEBUG,"\n");

}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
MODULE_DEVICE_TABLE(pci, bcraid_id_tbl);

static struct pci_driver bcraid_pci_driver = {
	.name		= "bcraid",
	.id_table	= bcraid_id_tbl,
	.probe		= bcraid_probe_one,
	.driver		= {
		.shutdown = bcraid_shutdown_one,
	},
};


static int bcraid_init(void)
{
	bc_init_module();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	return pci_register_driver(&bcraid_pci_driver);
#else
	return pci_module_init(&bcraid_pci_driver);
#endif
}

static void bcraid_exit(void)
{
	int i;

	bc_printk(BC_DEBUG, "bcraid_exit\n");

	if (bc_state.host_ptr) {
		scsi_remove_host(bc_state.host_ptr);
		bc_shutdown_host(bc_state.host_ptr);
		scsi_host_put(bc_state.host_ptr);
		bc_state.host_ptr = NULL;
	}

	for (i =0; i < bc_state.num_hba; i++)
		bc_shutdown_adapter(bc_dev[i]);

	pci_unregister_driver(&bcraid_pci_driver);
}

module_init(bcraid_init);
module_exit(bcraid_exit);

#endif


/*
 * from drivers/scsi/scsi.c
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 4
 * End:
 */
