
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 


#include "linux/vmalloc.h"
#include "linux/wait.h"
#include "linux/sched.h"
#include "linux/string.h"
#include "asm/page.h"
#include "asm/i387.h"

#ifdef CONFIG_HIGHMEM
#include "asm/highmem.h"
#endif

#include "bc.h"

#ifdef CONFIG_X86
#include "asm/i387.h" 

#ifndef CONFIG_X86_64
#include "bc_memcpy.h"
#else
#define bc_memcpy(dst,src,len) memcpy(dst,src,len)
#define bc_memcpy_init() 
#endif

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

#ifndef set_user_nice
#define set_user_nice(tsk, n) do { (tsk)->nice = n; } while(0)
#endif

#ifndef preempt_enable
#define preempt_enable() 	
#define preempt_disable() 
#endif

/*
 * taken from arch/i386/i387.c
 * this code is not exported in 2.4.x
 */

static __inline__ void __save_init_fpu( struct task_struct *tsk )
{
	if ( cpu_has_fxsr ) {
		asm volatile( "fxsave %0 ; fnclex"
			      : "=m" (tsk->thread.i387.fxsave) );
	} else {
		asm volatile( "fnsave %0 ; fwait"
			      : "=m" (tsk->thread.i387.fsave) );
	}
	tsk->flags &= ~PF_USEDFPU;
}

__inline__ void kernel_fpu_begin(void)
{
	struct task_struct *tsk = current;

	preempt_disable();
	if (tsk->flags & PF_USEDFPU) {
		__save_init_fpu(tsk);
		return;
	}
	clts();
}

#ifndef kernel_fpu_end
#define kernel_fpu_end() do { stts(); preempt_enable(); } while(0)
#endif


#endif

#define BC_THREAD_BUF_CNT  16

typedef struct {
	bc_sg_list_t *sg;
	int          size;
} bc_thread_buf_t;


typedef struct bc_thread_s {
	struct task_struct	thread;
	struct semaphore	ctl_sema;
	struct semaphore	sema;
	int			pid;
	int			running;
	int			stop;
	volatile int		num_mop;       /* number on queue */
	bc_mem_op_t		*mop_head;
	bc_mem_op_t		*mop_tail;
	bc_thread_buf_t         buf[BC_THREAD_BUF_CNT];   /* 0 is destination list, 1-15 are sources. */
} bc_thread_t;


bc_thread_t bc_thread[NR_CPUS];

void bc_kthread(void *arg);

void bc_start_all_threads(void);
void bc_stop_all_threads(void);
void bc_wakeup_all_threads(void);

int bc_start_thread(int cpu);
void bc_stop_thread(int cpu);
void bc_wakeup_thread(int cpu);

int bc_mem_copy_list ( bc_sg_list_t *dst, bc_sg_list_t *src, int byte_count);
int bc_mem_clear_list(bc_sg_list_t *dst, int byte_count);
int bc_kthread_mem_copy(bc_thread_t *tp, bc_mem_op_t *mop);
int bc_sync_mem_copy(bc_uint64_t dst, bc_uint32_t dst_id,
					 bc_uint64_t src, bc_uint32_t src_id,
					 bc_uint32_t byte_count);
int bc_sync_mem_clear(bc_uint64_t dst, bc_uint32_t dst_id,
					  bc_uint32_t byte_count);

/* 2.6 added a new macro defined here for older kernel versions. */
#ifndef for_each_online_cpu
#define for_each_online_cpu(cpu)  	for (cpu = 0; cpu < smp_num_cpus; cpu++) 
#endif

/* 2.6 removed cpu logical maps */
#ifndef cpu_logical_map
#define cpu_logical_map(cpu)        cpu
#endif

/* 2.4.22 added a typedef for cpumask_t, need to define it for earlier versions. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,22)
typedef unsigned long cpumask_t;
#endif

/*
 * 2.6.8 changed cpu mask from an unsigned long to a strucure, macros for backwards
 * compatibility
 */
#ifndef cpu_set
#define cpu_set(cpu, dst)       dst |= 1UL << cpu
#endif

#ifndef cpus_clear
#define cpus_clear(dst)         dst = 0UL
#endif

void
bc_start_all_threads(void)
{
	int cpu;

	bc_memcpy_init();
	for_each_online_cpu(cpu) 
		bc_start_thread(cpu);
}

void
bc_stop_all_threads(void)
{
	int cpu;

	for_each_online_cpu(cpu) 
		bc_stop_thread(cpu);
}

void
bc_wakeup_all_threads(void)
{
	int cpu;

	for_each_online_cpu(cpu) 
		bc_wakeup_thread(cpu);
}

int
bc_start_thread(int cpu)
{
	bc_thread_t *tp;

	tp = &bc_thread[cpu];

	memset(tp,0, sizeof(bc_thread_t));

	init_MUTEX_LOCKED(&tp->ctl_sema);
	init_MUTEX_LOCKED(&tp->sema);

        tp->pid = kernel_thread((int (*)(void *))bc_kthread, (void *)tp, 0);
        if(tp->pid < 0) {
                printk(KERN_ERR "bcraid: Unable to create thread for cpu %d.\n", cpu);
                return 0;
	}

	down(&tp->ctl_sema);
	return(1);
}

void
bc_stop_thread(int cpu)
{
	bc_thread_t *tp;

	tp = &bc_thread[cpu];
	tp->stop = 1;
	bc_wakeup_thread(cpu); 
	down(&tp->ctl_sema);
}

void
bc_wakeup_thread(int cpu)
{
	bc_thread_t *tp;

	tp = &bc_thread[cpu];
	up(&tp->sema);
}

void
bc_kthread(void *bc_threadp)
{

	int		cpu, status, i;
	bc_thread_t	*tp;
	bc_mem_op_t 	*mop;
	bc_softstate_t	*state;
	unsigned long	irql;
	cpumask_t       cpumask;

	state = &bc_state;

	tp = (bc_thread_t*)bc_threadp;

	cpu = tp - bc_thread;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        daemonize("bcraid/%d", cpu);
#else
	daemonize();
#endif


        set_user_nice(current, -10);
        sigfillset(&current->blocked);

        /* Migrate to the right CPU */
	cpus_clear(cpumask);
	cpu_set(cpu, cpumask);
        set_cpus_allowed(current, cpumask);
        if (smp_processor_id() != cpu)
                BUG();

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        sprintf(current->comm, "bcraid/%d", cpu);
#endif

	up(&tp->ctl_sema);

	bc_printk(BC_INFO,"bc_kthread: cpu %d thread started\n", cpu);

	while (1) {
	    preempt_disable();
	    local_irq_save(irql);
	    while (tp->mop_head) {
		mop = tp->mop_head;
		tp->mop_head = mop->next;
		mop->next = (bc_mem_op_t *)0;

		tp->num_mop--;

		local_irq_restore(irql);
		preempt_enable();

		switch (mop->opcode) {

		case BC_MEM_LIST_COPY:
			status = bc_kthread_mem_copy(tp, mop);
			break;

		case BC_MEM_LIST_XOR:
		case BC_MEM_LIST_CMP:
		case BC_MEM_CLEAR:
		case BC_MEM_COPY:
			bc_printk(BC_ERROR, "bc_msg_mem_op: Unsupported memory operation- %d\n",
				  mop->opcode);		  
			status = 0;
		    break;

		default:
			bc_printk(BC_ERROR, "bc_msg_mem_op: Invalid memory operation- %d\n",
				  mop->opcode);		      
			status = 0;

		}

		mop->status = status;
		/*
		 * Send a successfull response to the OSIC
		 * We just internally queue the response and
		 * let the tasklet send the response. Otherwise,
		 * we'll block on the OSIC lock.
		 */
		spin_lock_irqsave(&state->mop_done.lock, irql);

		if (state->mop_done.head == (bc_mem_op_t *)0) {
			state->mop_done.head = mop;
			state->mop_done.tail = mop;
		} else {
			state->mop_done.tail->next = mop;
			state->mop_done.tail = mop;
		}
		spin_unlock_irqrestore(&state->mop_done.lock, irql);

		tasklet_schedule(&state->srb_q.tasklet);

		preempt_disable();
		local_irq_save(irql);

	    }

	    local_irq_restore(irql);
	    preempt_enable();

	    BC_PRINTK(BC_DEBUG,"bc_kthread: cpu %d thread sleeping\n", cpu);
	    tp->running = 0;
	    /*
 	     * down() is counted in load average, down_interruptible() is not
	     */
	    if (down_interruptible(&tp->sema))
		bc_printk(BC_WARN, "bc_kthread: cpu %d interrupted\n", cpu);	

	    tp->running = 1;
	    BC_PRINTK(BC_DEBUG,"bc_kthread: cpu %d thread awake\n", cpu);

	    if (tp->stop == 1)
		break;

	}

	for (i = 0; i < BC_THREAD_BUF_CNT; i++) {
		if (tp->buf[i].size) {
			kfree(tp->buf[i].sg);
		}
	}

	bc_printk(BC_INFO,"bc_kthread: cpu %d thread stopped\n", cpu);
	up(&tp->ctl_sema);
}


static __inline__ int
bc_sg_list_size( bc_sg_list_t *src)
{
	int i, size;
	
	size = 0;

	if (!src) {
		bc_printk(BC_ERROR,"bc_sg_list_size: src == 0\n");
		return(0);
	}

	for (i = 0; i < src->sg_num_elem; i++)
		size += src->sg_elem[i].size;
	return(size);
}


#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))

#ifndef page_to_pfn
#define page_to_pfn(page)       ((unsigned long)((page) - mem_map))
#endif

#ifndef pfn_to_page
#define pfn_to_page(pfn)        (mem_map + (pfn))
#endif


/*
 * clear memory described by dst list.
 * if the map function fails, the whole clear fails
 */
int
bc_mem_clear_list(bc_sg_list_t *dst, int byte_count)
{
	int	dst_idx;
	int	dst_cnt;
	int	dst_mapped;	/* amount of sg list element currently mapped */
	int	dst_offset;	/* offset into current element */
	dma_addr_t	dst_dma_addr;
	unsigned char	*dst_vaddr;
	struct page 	*dst_page;
	int	xfer_cnt;
	int	residual;
	int	offset, size, pfn;

	dst_cnt = bc_sg_list_size(dst);

	if (byte_count > dst_cnt) {
		bc_printk(BC_ERROR, "bc_mem_clear_list, count %d dst_count %d\n", 
			byte_count, dst_cnt);
		BUG();
	}

	dst_idx       = 0;
	dst_mapped    = 0;
	dst_offset    = 0;
	dst_dma_addr  = 0;
	dst_vaddr     = (unsigned char *)0; 
	dst_page      = (struct page *)0;

	residual = byte_count;


	while (residual > 0) {

		if (dst_mapped == 0) {
			if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
			    dst_dma_addr = dst->sg_elem[dst_idx].dma_paddr + dst_offset;
			    offset = dst_dma_addr & (PAGE_SIZE-1);

			    size = dst->sg_elem[dst_idx].size - dst_offset;
			    size = min(size, PAGE_SIZE - offset);

			    pfn = dst_dma_addr >> PAGE_SHIFT;
			    dst_page = pfn_to_page(pfn);
			    dst_vaddr = kmap_atomic(dst_page, KM_USER1);
			    if (dst_vaddr == 0)
				    return(0);

			    dst_vaddr += offset;
			    dst_mapped = size;
			    dst_offset = dst_offset + size;
			    BC_PRINTK(BC_DEBUG3,
				      "dst[%d]: dma_addr 0x%llx size 0x%x map_addr 0x%x vaddr %p offset 0x%x mapped 0x%x\n",
				      dst_idx,
				      (bc_uint64_t)dst_dma_addr,
				      size,
				      pfn << PAGE_SHIFT,
				      dst_vaddr,
				      dst_offset,
				      dst_mapped);
			}
			else if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_VADDR) {
			    dst_vaddr = dst->sg_elem[dst_idx].v_addr;
			    dst_mapped = dst->sg_elem[dst_idx].size;
			    bc_printk(BC_DEBUG, "dst[%d]: vaddr %p size: %x\n", dst_idx, dst_vaddr, dst_mapped);
			}
			else
			    BUG();
		}


		xfer_cnt = min(residual, dst_mapped);

		BC_PRINTK(BC_DEBUG3, "bc_mem_clear_list: clearing %x bytes at %p\n",
			  xfer_cnt, dst_vaddr);

		memset(dst_vaddr, 0, xfer_cnt);

		dst_mapped -= xfer_cnt;
		residual   -= xfer_cnt;

		if (dst_mapped == 0) {
			if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
				if (dst_page)
					kunmap_atomic(dst_vaddr, KM_USER1);
				dst_page = (struct page *)0;
				if (dst_offset == dst->sg_elem[dst_idx].size) {
					dst_idx++;
					dst_offset = 0;
				}
			} else  if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_VADDR) {
				dst_idx++;
			} else
				BUG();
		} else
		    dst_vaddr += xfer_cnt;
	
	}

	if (dst_page)
		kunmap_atomic(dst_vaddr, KM_USER1);

	if (residual)
		bc_printk(BC_PANIC, "bc_mem_clear_list: residual is not 0 at end of loop\n");

	return(1);
}


/*
 * copy memory described by src list to dst list.
 * if the map function fails, the whole copy fails
 * and it will have to be rescheduled
 */
int
bc_mem_copy_list ( bc_sg_list_t *dst, bc_sg_list_t *src, int byte_count)
{

	int	src_idx,    dst_idx;
	int	src_cnt,    dst_cnt;
	int	src_mapped, dst_mapped;	/* amount of sg list element currently mapped */
	int	src_offset, dst_offset;	/* offset into current element */
	dma_addr_t	src_dma_addr,  dst_dma_addr;
	unsigned char	*src_vaddr,  *dst_vaddr;
	struct page 	*src_page, *dst_page;
	int	xfer_cnt;
	int	residual;
	int	offset, size, pfn;

	src_cnt = bc_sg_list_size(src);
	dst_cnt = bc_sg_list_size(dst);

	if (byte_count > src_cnt) {
		bc_printk(BC_ERROR, "bc_mem_copy_list, count %d src_count %d\n", 
			byte_count, src_cnt);
		BUG();
	}

	if (byte_count > dst_cnt) {
		bc_printk(BC_ERROR, "bc_mem_copy_list, count %d dst_count %d\n", 
			byte_count, dst_cnt);
		BUG();
	}

	src_idx       = 0;
	dst_idx       = 0;
	src_mapped    = 0;
	dst_mapped    = 0;
	src_offset    = 0;
	dst_offset    = 0;
	src_dma_addr  = 0;
	dst_dma_addr  = 0;
	src_vaddr     = (unsigned char *)0;
	dst_vaddr     = (unsigned char *)0; 
	src_page      = (struct page *)0;
	dst_page      = (struct page *)0;

	residual = byte_count;


	while (residual > 0) {

		if (src_mapped == 0) {
			if ((src->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
			    src_dma_addr = src->sg_elem[src_idx].dma_paddr + src_offset;
			    offset = src_dma_addr & (PAGE_SIZE-1);

			    size = src->sg_elem[src_idx].size - src_offset;
			    size = min(size, PAGE_SIZE - offset);

			    pfn = src_dma_addr >> PAGE_SHIFT;
			    src_page = pfn_to_page(pfn);
			    src_vaddr = kmap_atomic(src_page, KM_USER0);
			    if (src_vaddr == 0) {
				    if (dst_page)
					    kunmap_atomic(dst_vaddr, KM_USER1);
				    return(0);
			    }

			    src_vaddr += offset;
			    src_mapped = size;
			    src_offset = src_offset + size;
			    BC_PRINTK(BC_DEBUG3,
				      "src[%d]: dma_addr 0x%llx size %x map_addr 0x%x vaddr %p offset 0x%x mapped 0x%x\n",
				      src_idx,
				      (bc_uint64_t)src_dma_addr,
				      size,
				      pfn << PAGE_SHIFT,
				      src_vaddr,
				      src_offset,
				      src_mapped);
			}
			else if ((src->sg_mem_type & MEM_TYPE) == BC_MEM_VADDR) {
			    src_vaddr = src->sg_elem[src_idx].v_addr;
			    src_mapped = src->sg_elem[src_idx].size;
			    BC_PRINTK(BC_DEBUG3, "src[%d]: vaddr %p size %x\n", src_idx, src_vaddr, src_mapped);
			}
			else
				BUG();

		}

		if (dst_mapped == 0) {
			if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
			    dst_dma_addr = dst->sg_elem[dst_idx].dma_paddr + dst_offset;
			    offset = dst_dma_addr & (PAGE_SIZE-1);

			    size = dst->sg_elem[dst_idx].size - dst_offset;
			    size = min(size, PAGE_SIZE - offset);

			    pfn = dst_dma_addr >> PAGE_SHIFT;
			    dst_page = pfn_to_page(pfn);
			    dst_vaddr = kmap_atomic(dst_page, KM_USER1);
			    if (dst_vaddr == 0) {
				    if (src_page)
					    kunmap_atomic(src_vaddr, KM_USER0);
				    return(0);
			    }

			    dst_vaddr += offset;
			    dst_mapped = size;
			    dst_offset = dst_offset + size;
			    BC_PRINTK(BC_DEBUG3,
				      "dst[%d]: dma_addr 0x%llx size 0x%x map_addr 0x%x vaddr %p offset 0x%x mapped 0x%x\n",
				      dst_idx,
				      (bc_uint64_t)dst_dma_addr,
				      size,
				      pfn << PAGE_SHIFT,
				      dst_vaddr,
				      dst_offset,
				      dst_mapped);
			}
			else if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_VADDR) {
			    dst_vaddr = dst->sg_elem[dst_idx].v_addr;
			    dst_mapped = dst->sg_elem[dst_idx].size;
			    bc_printk(BC_DEBUG, "dst[%d]: vaddr %p size: %x\n", dst_idx, dst_vaddr, dst_mapped);
			}
			else
			    BUG();
		}


		xfer_cnt = min(src_mapped, dst_mapped);
		xfer_cnt = min(residual, xfer_cnt);

		BC_PRINTK(BC_DEBUG3, "bc_mem_copy_list: moving %x bytes from %p to %p\n",
			  xfer_cnt,src_vaddr, dst_vaddr);

		kernel_fpu_begin();
		bc_memcpy( dst_vaddr, src_vaddr, xfer_cnt);
		kernel_fpu_end();

		src_mapped -= xfer_cnt;
		dst_mapped -= xfer_cnt;
		residual   -= xfer_cnt;

		if (src_mapped == 0) {
			if ((src->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
				if (src_page)
					kunmap_atomic(src_vaddr, KM_USER0);
				src_page = (struct page *)0;
				if (src_offset == src->sg_elem[src_idx].size) {
					src_idx++;
					src_offset = 0;
				}
			} else if ((src->sg_mem_type & MEM_TYPE) == BC_MEM_VADDR) {
				src_idx++;
			} else
				BUG();
		} else
			src_vaddr += xfer_cnt;
	
		if (dst_mapped == 0) {
			if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
				if (dst_page)
					kunmap_atomic(dst_vaddr, KM_USER1);
				dst_page = (struct page *)0;
				if (dst_offset == dst->sg_elem[dst_idx].size) {
					dst_idx++;
					dst_offset = 0;
				}
			} else  if ((dst->sg_mem_type & MEM_TYPE) == BC_MEM_VADDR) {
				dst_idx++;
			} else
				BUG();
		} else
		    dst_vaddr += xfer_cnt;
	
	}

	if (src_page)
		kunmap_atomic(src_vaddr, KM_USER0);
	if (dst_page)
		kunmap_atomic(dst_vaddr, KM_USER1);

	if (residual)
		bc_printk(BC_PANIC, "bc_mem_copy_list: residual is not 0 at end of loop\n");

	return(1);
}

#ifndef virt_addr_valid
#define virt_addr_valid(kaddr)   pfn_valid(__pa(kaddr) >> PAGE_SHIFT)
#endif

/*
 * Sanity check a single address, making sure it is reasonable
 * based on the memory type.  Returns 1 if the address is good
 * or 0 if an error is found.
 */
static bc_uint32_t
bc_check_addr_one(bc_uint64_t addr, bc_uint32_t id, bc_uint32_t	byte_count)
{
	bc_uint32_t status = 0;

	id &= MEM_TYPE;

	if (id != BC_MEM_VADDR && id != BC_MEM_PADDR) {
		bc_printk(BC_ERROR,
			  "bc_check_addr_one: Bad memory type %x\n", id);
		return status;
	}

	if (id == BC_MEM_VADDR) {
		void *vaddr = ((addr_elem_t)addr).virt_addr;
		if ( ((unsigned long)vaddr < VMALLOC_START || 
			  (unsigned long)vaddr >= VMALLOC_END) &&
			 !virt_addr_valid(vaddr) ) {
			bc_printk(BC_PANIC,
					  "bc_check_addr_one: Invalid virtual address - %p\n",
					  vaddr);
			return status;
		}
	}
	else { /* Physical address */
		bc_uint64_t paddr = ((addr_elem_t)addr).phys_addr;
		unsigned long long pfn = paddr >> PAGE_SHIFT;
		if (!pfn_valid(pfn)) {
			bc_printk(BC_PANIC,
					  "bc_check_addr_one: Invalid physical address - 0x%llx\n",
					  paddr);
			return status;
		}
	}

	status = 1;
	return status;

}


/*
 * Sanity checks a address list structure making sure all the
 * addresses are reasonable based on the memory type. Returns
 * 1 if the list is good or 0 if an error is found.
 */
static bc_uint32_t
bc_check_addr_list(bc_addr_list_t *addr_list)
{
	int i;
	bc_uint16_t mem_type;
        bc_addr_list_t *ap;
	bc_uint32_t status = 0;

	mem_type = addr_list->mem_id & MEM_TYPE;

	if (mem_type != BC_MEM_VADDR && mem_type != BC_MEM_PADDR) {
		bc_printk(BC_ERROR,
			  "bc_check_addr_list: Bad memory type %x\n",
			  mem_type);
		return status;
	}

	for (ap = addr_list; ap; ap = ap->next) {
		if (ap->mem_elem_count > RC_SINGLE_IO_ADDRESS_COUNT) {
			bc_printk(BC_ERROR,
				  "bc_check_addr_list: Element count too large - %u\n",
				  addr_list->mem_elem_count);
			return status;
		}

		for (i = 0; i < ap->mem_elem_count; i++) {

			if (mem_type == BC_MEM_VADDR) {
				void *vaddr = ap->sg_list[i].v_addr;
				if ( ((unsigned long)vaddr < VMALLOC_START || 
				      (unsigned long)vaddr >= VMALLOC_END) &&
				     !virt_addr_valid(vaddr) ) {
					bc_printk(BC_PANIC,
						  "bc_check_addr_list: Invalid virtual address - %p\n",
						  vaddr);
					return status;
				}
			}
			else { /* Physical address */
				unsigned long long pfn = addr_list->sg_list[i].dma_paddr >> PAGE_SHIFT;
				if (!pfn_valid(pfn)) {
					bc_printk(BC_PANIC,
						  "bc_check_addr_list: Invalid physical address - 0x%llu\n",
						  ap->sg_list[i].dma_paddr);
					return status;
				}
			}
		}
	}

	status = 1;
	return status;

}


static int __inline__
bc_addr_list_elements( bc_addr_list_t  *ap)
{
	int num_elem;

	num_elem = 0;
	while (ap) {
		num_elem += ap->mem_elem_count;
		ap = ap->next;
	}
	return(num_elem);
}


/*
 * determine the size of the src and dst arrays.
 * allocate memory big enough to hold all the lists
 * convert the lists from a chained list to arrays
 * enqueue the mememory operation to be done by the bcraid mem_op thread
 */
void
bc_msg_mem_op(bc_mem_op_t *mop)
{
	int cpu, i;
	unsigned long irql;
	bc_thread_t *tp;
	bc_uint32_t status = 1;

	if (bc_msg_level >= BC_DEBUG) {

		switch (mop->opcode) {

		case BC_MEM_LIST_COPY:
			status &= bc_check_addr_list(mop->mem.list->dst_addr_list);
			status &= bc_check_addr_list(mop->mem.list->src.addr_list);
			break;

		case BC_MEM_LIST_XOR:
			status &= bc_check_addr_list(mop->mem.list->dst_addr_list);
			for (i = 0; i < mop->mem.list->array_count; i++) 
				status &= bc_check_addr_list(mop->mem.list->src.addr_list_array[i]);
			bc_printk(BC_ERROR, "bc_msg_mem_op: Unsupported memory operation- %d\n",
				  mop->opcode);
			status = 0;	
			break;

		case BC_MEM_LIST_CMP:
			for (i = 0; i < mop->mem.list->array_count; i++) 
				status &= bc_check_addr_list(mop->mem.list->src.addr_list_array[i]);
			bc_printk(BC_ERROR, "bc_msg_mem_op: Unsupported memory operation- %d\n",
				  mop->opcode);
			status = 0;
			break;

		case BC_MEM_COPY:
			status &= bc_check_addr_one(mop->mem.cp.dst, mop->mem.cp.dst_id,
										mop->mem.cp.byte_count);
			status &= bc_check_addr_one(mop->mem.cp.src, mop->mem.cp.src_id,
										mop->mem.cp.byte_count);
#if 0
			if (console_loglevel < 8)
				console_loglevel = 8;
#endif
			bc_printk(BC_DEBUG2, "BC_MEM_COPY: dst %llx id %x src %llx id %x count %x status1 %i\n",
					  mop->mem.cp.dst, mop->mem.cp.dst_id,
					  mop->mem.cp.src, mop->mem.cp.src_id,
					  mop->mem.cp.byte_count, status);
			break;

		case BC_MEM_CLEAR:
			status &= bc_check_addr_one(mop->mem.clr.dst, mop->mem.clr.dst_id,
										mop->mem.clr.byte_count);
#if 0
			if (console_loglevel < 8)
				console_loglevel = 8;
#endif
			bc_printk(BC_DEBUG2, "BC_MEM_CLEAR: dst %llx id %x count %x status1 %i\n",
					  mop->mem.clr.dst, mop->mem.clr.dst_id,
					  mop->mem.clr.byte_count, status);
			break;

		default:
			bc_printk(BC_ERROR, "bc_msg_mem_op: Invalid memory operation- %d\n",
				  mop->opcode);		      
			status = 0;

		}
	}

	mop->status = status;

	if (status) {

		switch (mop->opcode) {

		case BC_MEM_COPY:
			mop->status = bc_sync_mem_copy(mop->mem.cp.dst, mop->mem.cp.dst_id,
										   mop->mem.cp.src, mop->mem.cp.src_id,
										   mop->mem.cp.byte_count);
			BC_PRINTK(BC_DEBUG2, "BC_MEM_COPY: status2 %d\n", mop->status);
			break;

		case BC_MEM_CLEAR:
			mop->status = bc_sync_mem_clear(mop->mem.clr.dst, mop->mem.clr.dst_id,
											mop->mem.clr.byte_count);
			BC_PRINTK(BC_DEBUG2, "BC_MEM_COPY: status2 %d\n", mop->status);
			break;

		default:
			/*
			 * queue the morb to the correct cpu
			 */
			preempt_disable();
			cpu = smp_processor_id();
			tp = &bc_thread[cpu];

			local_irq_save(irql);

			mop->next = 0;
			if (tp->mop_head == (bc_mem_op_t *)0) {
				tp->mop_head = mop;
				tp->mop_tail = mop;
			} else {
				tp->mop_tail->next = mop;
				tp->mop_tail = mop;
			}

			if (tp->num_mop == 0) 
				up(&tp->sema);

			tp->num_mop++;

			local_irq_restore(irql);
			preempt_enable();
			break;
		}
	}
}

// STATS
int
bc_mop_stats(char *buf, int buf_size)
{
	int cnt, len;
	bc_softstate_t *state;
	bc_stats_t *sp;
	char *cp;
	int cpu;

	state = &bc_state;
	sp = &state->stats;
	

	cnt = 0;
	cp = buf;

	len = snprintf(cp, buf_size - cnt, "\nExternal memory opeations\nCPU      ");
	cp  += len;
	cnt += len;

	for_each_online_cpu(cpu)  {
		len = snprintf(cp, buf_size - cnt, "%8d  ", cpu);
		cp  += len;
		cnt += len;
	}

	len = snprintf(cp, buf_size - cnt, "\nPending  ");
	cp  += len;
	cnt += len;

	for_each_online_cpu(cpu)  {
		len = snprintf(cp, buf_size - cnt, "%8d  ", bc_thread[cpu].num_mop);
		cp  += len;
		cnt += len;
	}

	len = snprintf(cp, buf_size - cnt, "\n");
	cp  += len;
	cnt += len;
	*cp = '\0';

	return(cnt);
}

/*
 * Set up for a synchronous copy and call bc_mem_copy_list
 * to do the work.  Return 1 on success, 0 if the copy failed
 * for any reason.
 */
int
bc_sync_mem_copy(bc_uint64_t dst, bc_uint32_t dst_id,
				 bc_uint64_t src, bc_uint32_t src_id,
				 bc_uint32_t byte_count)
{
	bc_sg_list_t dst_sg, src_sg;

	dst_sg.sg_mem_type = dst_id;
	dst_sg.sg_num_elem = 1;
	dst_sg.sg_elem[0].size = byte_count;
	if ((dst_id & MEM_TYPE) == BC_MEM_VADDR)
		dst_sg.sg_elem[0].v_addr = ((addr_elem_t)dst).virt_addr;
	else /* physical */
		dst_sg.sg_elem[0].dma_paddr = ((addr_elem_t)dst).phys_addr;

	src_sg.sg_mem_type = src_id;
	src_sg.sg_num_elem = 1;
	src_sg.sg_elem[0].size = byte_count;
	if ((src_id & MEM_TYPE) == BC_MEM_VADDR)
		src_sg.sg_elem[0].v_addr = ((addr_elem_t)src).virt_addr;
	else /* physical */
		src_sg.sg_elem[0].dma_paddr = ((addr_elem_t)src).phys_addr;

	return bc_mem_copy_list (&dst_sg, &src_sg, byte_count);
}

/*
 * Set up for a synchronous clear and call bc_mem_clear_list
 * to do the work.  Return 1 on success, 0 if the copy failed
 * for any reason.
 */
int
bc_sync_mem_clear(bc_uint64_t dst, bc_uint32_t dst_id,
				  bc_uint32_t byte_count)
{
	bc_sg_list_t dst_sg;

	dst_sg.sg_mem_type = dst_id;
	dst_sg.sg_num_elem = 1;
	dst_sg.sg_elem[0].size = byte_count;
	if ((dst_id & MEM_TYPE) == BC_MEM_VADDR)
		dst_sg.sg_elem[0].v_addr = ((addr_elem_t)dst).virt_addr;
	else /* physical */
		dst_sg.sg_elem[0].dma_paddr = ((addr_elem_t)dst).phys_addr;

	return bc_mem_clear_list (&dst_sg, byte_count);
}

/* Flattens an address list and copies it to a scatter/gather list. */
bc_sg_list_t *
bc_mem_sg_list(bc_addr_list_t *ap,
	       bc_uint32_t starting_elem,
	       bc_uint32_t offset,
	       bc_thread_buf_t *buf)
{

	int i, indx = 0;
	int elems = bc_addr_list_elements(ap);
	int size = sizeof(bc_sg_list_t) + (elems - 1) * sizeof(bc_sg_elem_t);
	bc_sg_list_t *sg;

	/* (re)allocate the buffer if it isn't big enough */
	if (size > buf->size) {
		buf->size += 2048;
		kfree(buf->sg);
		buf->sg = kmalloc(buf->size, GFP_KERNEL);
	}
	
	sg = buf->sg;

	sg->sg_mem_type = ap->mem_id;

	if (starting_elem || offset) {

		bc_sg_elem_t *ep= &ap->sg_list[starting_elem];

		if ((sg->sg_mem_type & MEM_TYPE) == BC_MEM_PADDR) {
			sg->sg_elem[indx].dma_paddr = ep->dma_paddr + offset;
		} 
		else {
			sg->sg_elem[indx].v_addr = ep->v_addr + offset;
		}

		sg->sg_elem[indx].size = ep->size - offset;
		starting_elem++;
		indx++;
	}

	while(ap) {
		for ( i = starting_elem ; i < ap->mem_elem_count; i++) {
			sg->sg_elem[indx].addr = ap->sg_list[i].addr;
			sg->sg_elem[indx].size = ap->sg_list[i].size;
			indx++;
			offset = 0;
		}
		ap = ap->next;
		starting_elem = 0;
		       
	}
	sg->sg_num_elem = indx;

	return sg;
}


/*
 * Converts a memory operation structure to a memory copy request structure. 
 */
int
bc_kthread_mem_copy(bc_thread_t	*tp, bc_mem_op_t *mop)
{
	bc_sg_list_t *dst, *src;
	int byte_count;

	dst = bc_mem_sg_list(mop->mem.list->dst_addr_list, 
			     mop->mem.list->dst_element,
			     mop->mem.list->dst_offset,
			     &tp->buf[0]);

        src = bc_mem_sg_list(mop->mem.list->src.addr_list, 
			     mop->mem.list->src_element,
			     mop->mem.list->src_offset,
			     &tp->buf[1]);

	byte_count = mop->mem.list->sector_count << 9;

        return bc_mem_copy_list (dst, src, byte_count);
}


void 
bc_msg_end(void)
{

}
