
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 *
 *   MODULE: bc_msg.c
 *
 *	This module contains the code to inteface the SCSI upper layer
 *	with the RAIDCore specific hardware module.
 *
 *	Created By: Jon Fraser
 *
 */
#include "bc.h"
#include "linux/sysrq.h"
#include "asm/msr.h"

/* 2.6.x kernel moved the macro for determing if a page is highmem */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/page-flags.h>
#endif


int bc_setup_communications(void *p_start_addr, long size);
void bc_send_msg(struct bc_send_arg_s *p_send_arg);
void bc_receive_msg(void);
void bc_send_test(void);
int bc_msg_send_srb(struct scsi_cmnd * scp);

void bc_msg_send_srb_function (bc_softstate_t *state, int function_code);
void bc_msg_end(void);

void bc_msg_timer(unsigned long data);

void bc_msg_timeout(int to);
void bc_msg_timeout_done(unsigned long data);

void bc_msg_isr( bc_adapter_t *adapter);
void bc_msg_srb_done(struct bc_srb_s *srb);
void bc_msg_srb_complete(struct bc_srb_s *srb);
void bc_msg_build_sg(bc_srb_t *srb);
void bc_msg_release_sg(bc_srb_t *srb);
void bc_msg_map_mem(struct map_memory_s *map);
void bc_msg_unmap_mem(struct unmap_memory_s *unmap);
void bc_msg_shutdown( bc_softstate_t *statep);

void bc_msg_srb_q_tasklet(  unsigned long arg);
void bc_msg_srb_done_tasklet(  unsigned long arg);
void bc_dump_scp( struct scsi_cmnd * scp );
void bc_clear_stack(void);
int bc_msg_stats(char *buf, int buf_size);
int bc_mop_stats(char *buf, int buf_size);
void bc_wakeup_all_threads(void);

int bc_elrond_init(bc_adapter_t *adapter);

static struct bc_interface_s *bc_interface_header = 0;

int bc_srb_seq_num = 0;

void bc_sysrq_intr (	int key,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
			struct pt_regs *pt_regs,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			struct kbd_struct * kbd,
#endif
#endif
			struct tty_struct * tty);

void bc_sysrq_state (	int key,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
			struct pt_regs *pt_regs,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			struct kbd_struct * kbd,
#endif
#endif
			struct tty_struct * tty);


struct sysrq_key_op bc_skey_ops_intr = {
	handler: bc_sysrq_intr,
	help_msg: "bcraid-Force-interrupt",
	action_msg: "forcing bcraid interrupt",
};

struct sysrq_key_op bc_skey_ops_dump = {
	handler: bc_sysrq_state,
	help_msg: "bcraid-Dump-state",
	action_msg: "Dumping bcraid state",
};

#define check_lock(sp) { if (sp->osic_locked) {bc_printk(BC_WARN,"%s: osic already locked by %s\n", \
					__FUNCTION__, sp->osic_lock_holder); \
					panic("osic_lock already held\n"); } }

void
bc_set_sense_data (char	*sense,
		    uint8_t	sense_key,
		    uint8_t	sense_code,
		    uint8_t	add_sense_code,
		    uint8_t	incorrect_len,
		    uint8_t	bit_ptr,
		    uint32_t	field_ptr,
		    uint32_t	residue);


static char bc_stats_buf[1024];


#define TWO_TRIP_WRITE_RATE_DEFAULT	  16
#define TWO_TRIP_WRITE_RATE_MIN		   8
#define TWO_TRIP_WRITE_RATE_MAX		1024

static int	TwoTripWriteRate = TWO_TRIP_WRITE_RATE_DEFAULT;
#ifdef module_param_named
module_param_named(TwoTripWriteRate, TwoTripWriteRate, int, 0);
#else
MODULE_PARM (TwoTripWriteRate, "i");
#endif
MODULE_PARM_DESC (TwoTripWriteRate, "min write rate (kb/s) to bypass cache");



#define ONE_TRIP_WRITE_RATE_DEFAULT	  8
#define ONE_TRIP_WRITE_RATE_MIN		  4
#define ONE_TRIP_WRITE_RATE_MAX		256

static int	OneTripWriteRate = ONE_TRIP_WRITE_RATE_DEFAULT;
#ifdef module_param_named
module_param_named(OneTripWriteRate, OneTripWriteRate, int, 0);
#else
MODULE_PARM (OneTripWriteRate, "i");
#endif
MODULE_PARM_DESC (OneTripWriteRate, "min write rate (kb/s) to bypass cache");;


#define WRITE_BYPASS_THRESHOLD_DEFAULT	 50
#define WRITE_BYPASS_THRESHOLD_MIN	  4
#define WRITE_BYPASS_THRESHOLD_MAX	128

static int	WriteBypassThreshold = WRITE_BYPASS_THRESHOLD_DEFAULT;
#ifdef module_param_named
module_param_named(WriteBypassThreshold, WriteBypassThreshold, int, 0);
#else
MODULE_PARM (WriteBypassThreshold, "i");
#endif
MODULE_PARM_DESC (WriteBypassThreshold, "min write size (kb) to bypass cache");

#define ACTIVE_RAID5_FLUSHES_LIMIT_DEFAULT	64
#define ACTIVE_RAID5_FLUSHES_LIMIT_MIN		 8
#define ACTIVE_RAID5_FLUSHES_LIMIT_MAX		128
static int	ActiveRaid5FlushesLimit = ACTIVE_RAID5_FLUSHES_LIMIT_DEFAULT;

#ifdef module_param_named
module_param_named(ActiveRaid5FlushesLimit, ActiveRaid5FlushesLimit, int, 0);
#else
MODULE_PARM (ActiveRaid5FlushesLimit, "i");
#endif
MODULE_PARM_DESC (ActiveRaid5FlushesLimit, "Active Raid5 Flushes Limit");

static int ForcePhysAddr = 0;
#ifdef module_param_named
module_param_named(ForcePhysAddr, ForcePhysAddr, int, 0);
#else
MODULE_PARM (ForcePhysAddr, "i");
#endif
MODULE_PARM_DESC (ForcePhysAddr, "force all memory references to use physical addresses");

static unsigned int DoSmart = 1;
#ifdef module_param_named
module_param_named(DoSmart, DoSmart, uint, 0444);
#else
MODULE_PARM (DoSmart, "i");
#endif
MODULE_PARM_DESC (DoSmart, "Do SMART polling");

#define SMART_POLL_INTERVAL_DEFAULT 900
static unsigned int SmartPollInterval = SMART_POLL_INTERVAL_DEFAULT;	/* 15 minutes */
#ifdef module_param_named
module_param_named(SmartPollInterval, SmartPollInterval, uint, 0444);
#else
MODULE_PARM (SmartPollInterval, "i");
#endif
MODULE_PARM_DESC (SmartPollInterval, "SMART poll interval in seconds");

/*
 *
 *   ROUTINE: bc_setup_communications
 *
 *		This routine sets up the communications with the code that is
 *		installed in CPU memory.
 *
 *	Returns:
 *		1 = Success
 *		0 = Failure
 *
 */
int
bc_setup_communications(void *p_start_addr, long size)
{
	char *p_next_addr;
	long i;
	int	found;

	p_start_addr = (void *)((unsigned long)p_start_addr & ~0x3);

	found = 0;
	p_next_addr = (char *)p_start_addr;
	for (i=0; i<size; i++) {
		bc_interface_header = (struct bc_interface_s *)p_next_addr;
		if (   (bc_interface_header->cookie_lo == BC_COOKIE_VALUE_LO)
			&& (bc_interface_header->cookie_hi == BC_COOKIE_VALUE_HI)
			&& (bc_interface_header->version == BC_INTERFACE_VERSION)
			&& (bc_interface_header->checksum == BC_INTERFACE_CHECKSUM)) {
			//
			//	We found the interface structure!! Fill in the values
			//	with what we will use!
			//
			bc_printk(BC_DEBUG, "found init struct at %p\n", p_next_addr);
			bc_printk(BC_DEBUG, "send_function: offset: %p\n", 
				bc_interface_header->send_function); 

			if ((void *)bc_interface_header->send_function  < (void *)0x100000)
			    bc_interface_header->send_function = 
				(bc_function_t *)((char *)(bc_interface_header->send_function) 
				+ (long)p_next_addr);
			bc_interface_header->receive_function = &bc_receive_msg;
			found = 1;
			bc_printk(BC_DEBUG, "send_function: final: %p\n", 
				bc_interface_header->send_function); 

			break;
		}
		p_next_addr += 4; // Assume 32-bit alignment
	}
	
	return (found);
}



/*
 *
 *   ROUTINE: bc_send_msg
 *
 *		This routine sends a command to the hardware communications
 *
 *	Returns:
 *		None
 *
 */
void
bc_send_msg(struct bc_send_arg_s *args)
{
	bc_interface_header->send_arg = args;
	(*bc_interface_header->send_function)();
}





/*
 *
 *   ROUTINE: bc_send_test
 *
 *		send a test command to the hardware communications
 *
 *	Returns:
 *		None
 *
 */
void
bc_send_test(void)
{
	struct bc_send_arg_s args;
	args.call_type = BC_CTS_TEST;
	bc_send_msg(&args);
}


void
bc_msg_shutdown( bc_softstate_t *statep)
{

	/*
         * send a message to the OSIC to suhtdown
	 * have to wait for it to stop doing IO.
         */
	bc_printk(BC_INFO2, "bc_msg_shutdown: flushing cache OSIC\n");
	bc_msg_send_srb_function(statep, BC_SRB_FLUSH);

	bc_printk(BC_INFO2, "bc_msg_shutdown: shutting down OSIC\n");
	bc_msg_send_srb_function(statep, BC_SRB_SHUTDOWN);

	bc_printk(BC_INFO2, "bc_msg_shutdown: stop OSIC timer\n");
	statep->state &= ~ENABLE_TIMER;
	del_timer(&statep->timer);
	bc_printk(BC_DEBUG, "bc_msg_shutdown: pausing for 1/4 second\n");
	bc_msg_timeout(HZ>>2);



	bc_printk(BC_INFO2, "bc_msg_shutdown: OSIC disabled \n");
	statep->state &= ~USE_OSIC;

	bc_stop_all_threads();
	bc_event_shutdown();

	bc_printk(BC_INFO2, "bc_msg_shutdown: shutting down srb tasklets\n");
	tasklet_kill(&statep->srb_done.tasklet);
	tasklet_kill(&statep->srb_q.tasklet);

	if (statep->virtual_memory) {
		bc_printk(BC_DEBUG, "bc_msg_shutdown: free virtual memory %p\n", statep->virtual_memory);
		vfree(statep->virtual_memory);
		statep->virtual_memory_size = 0;
		statep->virtual_memory = (void *)0;
	}

	if (statep->cache_memory) {
		bc_printk(BC_DEBUG, "bc_msg_shutdown: free cache memory %p\n", statep->cache_memory);
		vfree(statep->cache_memory);
		statep->cache_memory_size = 0;
		statep->cache_memory = (void *)0;
	}
	      
	unregister_sysrq_key('f', &bc_skey_ops_intr);
	unregister_sysrq_key('d', &bc_skey_ops_dump);

}


#define FMT_BUF_LEN 128
unsigned char bc_msg_fmt[FMT_BUF_LEN];
void
bc_msg_print(bc_print_args_t *args )
{
	const char *fmt;
	int   severity, count, i;
	unsigned long buf[20];
	char *bp;
	static int newline_in_last_fmt = 1;

	struct param {
	    union {
		u64		v64;		// 64 bits
		u32		v32[2];
	    } val;
	    bc_uint64_t		type;
	} *p;



	if (newline_in_last_fmt) {
		snprintf(bc_msg_fmt, sizeof(bc_msg_fmt), "bcraid: %s", args->format_string);
		fmt = bc_msg_fmt;
	} else {
		fmt = args->format_string;
	}

	newline_in_last_fmt = strchr(fmt, '\n') ? 1 : 0;

	severity = args->severity;
	count  = args->count;
	p = (struct param *) args->params;

	if (fmt == NULL) {
		bc_printk(BC_WARN, "bc_msg_print: NULL fmt\n");
		return;
	}

		
	/*
	 * this code assumes the arguments should be tightly packed on the stack.
	 * i.e. 64 bit values don't have to be on 64 bit boundaries
	 */
 
	memset(buf,0, sizeof(buf));
	bp = (char *)buf;

	for (i = 0; i < count; i++) {
	    switch(p[i].type) {
		case BC_CTR_PFTYPE_32_BIT:
#ifdef CONFIG_X86_64
		    *(u64 *)bp = p[i].val.v32[0];
		    bp += sizeof(u64);
#else
		    *(u32 *)bp = p[i].val.v32[0];
		    bp += sizeof(u32);
#endif
		    break;

		case BC_CTR_PFTYPE_64_BIT:
		    *(u64  *)bp = p[i].val.v64;
		    bp += sizeof(u64);
		    break;

		case BC_CTR_PFTYPE_POINTER:
#ifdef CONFIG_X86_64
		    *(unsigned long  *)bp = (unsigned long) p[i].val.v64;
#else
		    *(u32  *)bp =  p[i].val.v32[0];
#endif
		    bp += sizeof(void *);
		    break;

		default:
		    break;
	    }
	}

	bc_printk(severity, fmt,
			 buf[0],
			 buf[1],
			 buf[2],
			 buf[3],
			 buf[4],
			 buf[5],
			 buf[6],
			 buf[7],
			 buf[8],
			 buf[9],
			 buf[10],
			 buf[11],
			 buf[12],
			 buf[13],
			 buf[14],
			 buf[15],
			 buf[16],
			 buf[17],
			 buf[18],
			 buf[19]);


}
/*
 *
 *   ROUTINE: bc_receive_msg(void)
 *
 *		This routine receives a message call from the hardware.
 *
 *	Returns:
 *		None
 *
 */
void
bc_receive_msg(void)
{
	struct bc_receive_arg_s *args;
	bc_softstate_t *state;
	int delay;


	args = bc_interface_header->receive_arg;
	state = &bc_state;


	switch (args->call_type) {
	    case BC_CTR_TEST:
		bc_printk(BC_DEBUG, "bc_receive_msg Send/Receive test passed\n");
		break;

	    case BC_CTR_PRINTF:
		
		bc_msg_print(&args->u.printf);
		break;

	    case BC_CTR_SRB_DONE:
		bc_msg_srb_done(args->u.srb_done.srb);
		break;

	    case BC_CTR_INIT_DONE:
		bc_printk(BC_INFO,"bc_receive_msg: init done callback\n");
		up(&state->init_sema);
		break;

	    case BC_CTR_ASSERTION_FAILURE:
		break;

	    case BC_CTR_EVENT:
		bc_event(args->u.event.bc_notification_type,
				args->u.event.bc_bus_changed,
				BC_SRB_GLOBAL_UPDATE);
		break;

	    case BC_CTR_MAP_MEMORY:
		bc_msg_map_mem(&args->u.map_memory);
		break;

	    case BC_CTR_UNMAP_MEMORY:
		bc_msg_unmap_mem(&args->u.unmap_memory);
		break;

	    case BC_CTR_WAIT_MICROSECONDS:
		delay = args->u.wait_microseconds.microseconds;
//		bc_printk(BC_DEBUG2, "delay %d microseconds\n", delay);

		if (delay <= 1000) {
			udelay(delay);
			break;
		}
		if (state->osic_locked)
			udelay(delay);
		else {
			int ticks;
			int usec_per_tick;

			usec_per_tick =  1000000/HZ;
			delay += usec_per_tick >> 1;	/* round up by 1/2 clock tick */
			ticks = delay / usec_per_tick;

			if (ticks == 0)
				ticks = 1;
			bc_msg_timeout(ticks);
		}
		break;

		case BC_CTR_MEMORY_OP:
		    bc_msg_mem_op(args->u.mem_op);
		break;

	    default:
		bc_printk(BC_WARN,"bc_receive_msg: unknown msg type 0x%x\n", args->call_type);
		break;
	}
}



/*
 * setup the whole message passing system
 */


int
bc_msg_init(bc_softstate_t *state)
{

	void 		*init_addr;
	long	 	init_size;
	bc_send_arg_t	args;
	int 		i, size,  period;
	bc_adapter_t	*adapter;
	
	init_addr = &bc_msg_end;
	init_size = 600 * 1024;


	/*
	 * find the initialization struct and fill in our function pointer
	 */
	if (bc_setup_communications(init_addr, init_size) == 0) {
		bc_printk(BC_ERROR,"bc_msg_init: could not find init structure\n");
		return(1);
	}

	bc_send_test();

	/*
	 * setup lock and counter for processing pending interrupts
	 */
	spin_lock_init(&state->intr_lock);
	state->intr_pending = 0;

	/*
	 * setup tasklet for srb q processing;
	 */

	state->srb_q.head = (bc_srb_t *)0;
	state->srb_q.tail = (bc_srb_t *)0;
	spin_lock_init(&state->srb_q.lock);

	tasklet_init(&state->srb_q.tasklet, bc_msg_srb_q_tasklet, (unsigned long)state);

	/*
	 * setup tasklet for srb done processing;
	 */

	state->srb_done.head = (bc_srb_t *)0;
	state->srb_done.tail = (bc_srb_t *)0;
	spin_lock_init(&state->srb_done.lock);

	tasklet_init(&state->srb_done.tasklet, bc_msg_srb_done_tasklet, (unsigned long)state);

	state->mop_done.head = (bc_mem_op_t *)0;
	state->mop_done.tail = (bc_mem_op_t *)0;
	spin_lock_init(&state->mop_done.lock);

	state->osic_locked = 0;
	state->osic_lock_holder = "INIT";

	init_MUTEX_LOCKED(&state->msg_timeout_sema);

	bc_start_all_threads();
	bc_event_init();

	/*
	 *  send a  get_info msg to get setup info
	 */

	memset(&args, 0, sizeof(args));
	args.call_type = BC_CTS_GET_INFO;
	args.u.get_info.controller_count = state->num_hba;
	args.u.get_info.max_sg_map_elements = BC_SG_MAX_ELEMENTS;
	args.u.get_info.max_print_severity = bc_msg_level;
	get_random_bytes( &args.u.get_info.random_seed, sizeof(args.u.get_info.random_seed)); 

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	args.u.get_info.max_lba = 0xffffffffLL;
#else
	args.u.get_info.max_lba = 0xffffffffffffffffLL;
#endif




	if (TwoTripWriteRate < TWO_TRIP_WRITE_RATE_MIN)
		TwoTripWriteRate = TWO_TRIP_WRITE_RATE_MIN;
	if (TwoTripWriteRate > TWO_TRIP_WRITE_RATE_MAX)
		TwoTripWriteRate = TWO_TRIP_WRITE_RATE_MAX;

	args.u.get_info.parameter_mask |= (1 << bc_TwoTripWriteRate);
	args.u.get_info.tunable_parameters[bc_TwoTripWriteRate] = TwoTripWriteRate << 1;

	if (TwoTripWriteRate != TWO_TRIP_WRITE_RATE_DEFAULT)
		bc_printk(BC_INFO,"bcraid: set parameter TwoTripWriteRate = %d\n", TwoTripWriteRate);


	if (OneTripWriteRate < ONE_TRIP_WRITE_RATE_MIN)
		OneTripWriteRate = ONE_TRIP_WRITE_RATE_MIN;
	if (OneTripWriteRate > ONE_TRIP_WRITE_RATE_MAX)
		OneTripWriteRate = ONE_TRIP_WRITE_RATE_MAX;

	args.u.get_info.parameter_mask |= (1 << bc_OneTripWriteRate);
	args.u.get_info.tunable_parameters[bc_OneTripWriteRate] = OneTripWriteRate << 1;

	if (OneTripWriteRate != ONE_TRIP_WRITE_RATE_DEFAULT)
		bc_printk(BC_INFO,"bcraid: set parameter OneTripWriteRate = %d\n", OneTripWriteRate);


	if (WriteBypassThreshold < WRITE_BYPASS_THRESHOLD_MIN)
		WriteBypassThreshold = WRITE_BYPASS_THRESHOLD_MIN;

	if (WriteBypassThreshold > WRITE_BYPASS_THRESHOLD_MAX)
		WriteBypassThreshold = WRITE_BYPASS_THRESHOLD_MAX;

	args.u.get_info.parameter_mask |= (1 << bc_WriteBypassThreshold);
	args.u.get_info.tunable_parameters[bc_WriteBypassThreshold] = WriteBypassThreshold << 1;

	if (WriteBypassThreshold != WRITE_BYPASS_THRESHOLD_DEFAULT)
		bc_printk(BC_INFO,"bcraid: set parameter WriteBypassThreshold = %d\n", WriteBypassThreshold);


	if (ActiveRaid5FlushesLimit < ACTIVE_RAID5_FLUSHES_LIMIT_MIN)
		ActiveRaid5FlushesLimit = ACTIVE_RAID5_FLUSHES_LIMIT_MIN;

	if  (ActiveRaid5FlushesLimit > ACTIVE_RAID5_FLUSHES_LIMIT_MAX)
		ActiveRaid5FlushesLimit = ACTIVE_RAID5_FLUSHES_LIMIT_MAX;

	args.u.get_info.parameter_mask |= (1 << bc_ActiveRaid5FlushesLimit);
	args.u.get_info.tunable_parameters[bc_ActiveRaid5FlushesLimit] = ActiveRaid5FlushesLimit;		

	if (ActiveRaid5FlushesLimit != ACTIVE_RAID5_FLUSHES_LIMIT_DEFAULT)
		bc_printk(BC_INFO,"bcraid: set parameter ActiveRaid5FlushesLimit = %d\n", ActiveRaid5FlushesLimit);

	if  (DoSmart > 1)
		DoSmart =  1;

	args.u.get_info.parameter_mask |= (1 << bc_DoSmart);
	args.u.get_info.tunable_parameters[bc_DoSmart] = DoSmart;		

	if (DoSmart != 1)
		bc_printk(BC_INFO,"bcraid: set parameter DoSmart = %d, SMART polling disabled\n", DoSmart);

	if (SmartPollInterval < 60)		/* don't allow polling more often than every minute */
		SmartPollInterval = 60;
	args.u.get_info.parameter_mask |= (1 << bc_SmartPollInterval);
	args.u.get_info.tunable_parameters[bc_SmartPollInterval] = SmartPollInterval;		

	if (SmartPollInterval != SMART_POLL_INTERVAL_DEFAULT)
		bc_printk(BC_INFO,"bcraid: set parameter SmartPollInterval = %d seconds\n", SmartPollInterval);

	bc_send_msg(&args);

	/* should sanity check sizes */
	state->memsize_per_controller = args.u.get_info.memory_size_per_controller;
	state->memsize_per_srb        = args.u.get_info.memory_size_per_srb;
	period = 1000 / HZ;		/* milliseconds per interval */
	state->timer_interval         = (args.u.get_info.timer_interval + period - 1) / period;
	

	/*
	 * now go allocate the private memory for each controller
	 */
	size = state->memsize_per_controller;
	for (i = 0; i < state->num_hba; i++) {
		void *addr;
		adapter = bc_dev[i];

		bc_printk(BC_DEBUG, "bc_msg_init: get private mem for controller %d\n", i);
		if (adapter == (bc_adapter_t *)0) {
			bc_printk(BC_ERROR, "bc_msg_init null adapter\n");
		}
		addr = pci_alloc_consistent(adapter->pdev,
			    state->memsize_per_controller,
			    &adapter->private_mem.dma_address);

		if (addr == (void *)0) {
			bc_printk(BC_ERROR,"bc_msg_init: can not alloc %d bytes of per controller memory\n", state->memsize_per_controller);
			/* need to free previously allocated.... */
			return(1);
		}

		adapter->private_mem.vaddr = addr;
		adapter->private_mem.size = state->memsize_per_controller;

		bc_printk(BC_DEBUG, "bc_msg_init: private mem controller %d @ %p len %d\n",
				i,
				adapter->private_mem.vaddr,
				adapter->private_mem.size );
		
	}


	/*
	 * now we can initialize each controller 
	 */
	for (i = 0; i < state->num_hba; i++) {
		adapter = bc_dev[i];

		memset(&args, 0, sizeof(args));

		args.call_type = BC_CTS_INIT_CONTROLLER;

                args.u.init_controller.controller_memory       = adapter->private_mem.vaddr;
                args.u.init_controller.controller_memory_id    = BC_MEM_VADDR;
                args.u.init_controller.controller_handle       = (void *)(adapter);
                args.u.init_controller.pci_config_space        = adapter->hardware.pci_config_space;
                args.u.init_controller.pci_config_space_length = PCI_CFG_SIZE;

                args.u.init_controller.bar_memory[adapter->version->which_bar] = adapter->hardware.vaddr;
                args.u.init_controller.bar_length[adapter->version->which_bar] = adapter->hardware.mem_len;

		if (adapter->version->init_func == bc_elrond_init) {
			args.u.init_controller.bar_memory[0] = adapter->hardware.sram_vaddr;
			args.u.init_controller.bar_length[0] = adapter->hardware.sram_len;

			args.u.init_controller.bar_memory[1] = (void *)(unsigned long)adapter->hardware.io_port;
			args.u.init_controller.bar_length[1] = adapter->hardware.io_port_len;
		}

                // Fill in the rest
		bc_printk(BC_DEBUG, "bc_msg_init: init controller %d\n", i);
                bc_send_msg(&args);
		bc_printk(BC_DEBUG, "bc_msg_init: init controller %d Done\n", i);
		/* return status ??? */
	}


        /* Allocate virtual memory for the core*/
	state->virtual_memory_size = args.u.init_controller.virtual_memory_size_needed;
	if (state->virtual_memory_size) {
		if ((state->virtual_memory = vmalloc(state->virtual_memory_size)) == 0) {
			bc_printk(BC_ERROR, "bc_msg_init: can't allocate virtual memory, size %d bytes\n", 
				state->virtual_memory_size);
			/* need to free previously allocated.... */
			return(1);
		}	
		memset(state->virtual_memory, 0, state->virtual_memory_size);
	}
	bc_printk(BC_DEBUG,"bc_msg_init: alloc %d @ %p for virtual memory\n",
			state->virtual_memory_size,
			state->virtual_memory);

	
        /* Allocate cache memory for the core */
	state->cache_memory_size = args.u.init_controller.cache_memory_size_needed;

	if (args.u.init_controller.cache_memory_size_needed) {
		if ((state->cache_memory = vmalloc(state->cache_memory_size)) == 0) {
			bc_printk(BC_ERROR, "bc_msg_init: can't allocate %u bytes of cache memory\n", state->cache_memory_size);
			/* need to free previously allocated.... */
			return(1);
		}	
		memset(state->cache_memory, 0, state->cache_memory_size);
	}
	bc_printk(BC_DEBUG,"bc_msg_init: alloc %d @ %p for cache memory\n",
			state->cache_memory_size,
			state->cache_memory);


	/*
	 * do the final init for the OSIC
	 */
	state->state |= USE_OSIC;

	args.call_type = BC_CTS_FINAL_INIT;
	args.u.final_init.virtual_memory                   = state->virtual_memory;
        args.u.final_init.cache_memory                     = state->cache_memory;
	args.u.final_init.size_of_virtual_memory_allocated = state->virtual_memory_size;
        args.u.final_init.size_of_cache_memory_allocated   = state->cache_memory_size;
	args.u.final_init.virtual_memory_id                = BC_MEM_VADDR;
        args.u.final_init.cache_memory_id                  = BC_MEM_VADDR;
	args.u.final_init.timer_interval                   = state->timer_interval * period;

	bc_printk(BC_INFO2,"bc_msg_init: doing final init\n");

	init_MUTEX_LOCKED(&state->init_sema);

	spin_lock(&state->osic_lock);
	check_lock(state);
	state->osic_locked = 1;
	state->osic_lock_holder = "bc_msg_init";
	bc_send_msg(&args);
	state->osic_locked = 0;
	spin_unlock(&state->osic_lock);


	state->state |= PROCESS_INTR;

	/*
	 * intialize the periodic timer for the OSIC
	 */

	init_timer(&state->timer);
	state->timer.expires = jiffies  + state->timer_interval ;
	state->timer.data = (unsigned long)state;
	state->timer.function = bc_msg_timer;
	state->state |= ENABLE_TIMER;

	add_timer(&state->timer);

	bc_printk(BC_INFO2,"bc_msg_init: timer started.... wait for callback \n");
	down(&state->init_sema);
	bc_printk(BC_INFO2,"bc_msg_init: we're back\n");

	register_sysrq_key('f', &bc_skey_ops_intr);
	register_sysrq_key('d', &bc_skey_ops_dump);
	
	/*
	 * Inventory what devices are connected.
	 * The flag means only build the device table, but don't send
	 * anything to the upper scsi layer.
	 */
	bc_event(BC_CTR_EVENT_CONFIG_CHANGE_DETECTED,
		    (bc_uint8_t) 0, 
		    BC_SRB_LOCAL_UPDATE_ONLY);

	/*
	 * OK.  we're ready to accept IO for the OSIC
	 */
	return(0);
}
void
bc_msg_timer(unsigned long data)
{
	bc_softstate_t *state;
	bc_send_arg_t	args;


	state = (bc_softstate_t *)data;
	
	if ((state->state & ENABLE_TIMER) == 0)
		return;

	/*
	 * set up timeout
	 */
	init_timer(&state->timer);
	state->timer.expires = jiffies  + state->timer_interval;
	state->timer.data = (unsigned long)state;
	state->timer.function = bc_msg_timer;
	add_timer(&state->timer);

	spin_lock(&state->osic_lock);
	check_lock(state);
	state->osic_locked = 1;
	state->osic_lock_holder = "bc_msg_timer";
	args.call_type = BC_CTS_TIMER_CALL;
	bc_send_msg(&args);
	state->osic_locked = 0;
	spin_unlock(&state->osic_lock);
}


/*
 * just increment intr_pending and wake up the tasklet
 */
void
bc_msg_isr( bc_adapter_t *adapter)
{
	unsigned long 	flags;
	bc_softstate_t *state;

	state = &bc_state;

//	bc_printk(BC_DEBUG_HWISR, "bc_msg_isr\n");

	spin_lock_irqsave(&state->intr_lock,flags);
	state->intr_pending += 1;			/* this really could become an atomic counter now */
	spin_unlock_irqrestore(&state->intr_lock, flags);

	tasklet_schedule(&state->srb_q.tasklet);
}

int
bc_msg_send_srb(struct scsi_cmnd * scp)
{
	bc_softstate_t	*state;
	bc_srb_t	*srb;
	int		size;
	int		bus, target, lun;
	unsigned long	irql;
	int		sg_list_size;

	state = &bc_state;

	
	bus    = scp->device->channel;
	target = scp->device->id;
	lun    = scp->device->lun;

	sg_list_size =  sizeof (bc_sg_list_t) + (BC_SG_MAX_ELEMENTS-1) * sizeof(bc_sg_elem_t);
	size = sizeof(bc_srb_t) + sg_list_size + state->memsize_per_srb;

	/*
	 * Use GFP_ATOMIC instead of GP_KERNEL because we get called from the srb tasklet
	 */
	if ((srb =  kmalloc( size, GFP_ATOMIC)) == 0) {
	    if ((srb = kmalloc( size << 1, GFP_ATOMIC)) == 0) {
		bc_printk(BC_DEBUG,"bc_msg_send_srb: could not alloc %d bytes for srb\n", size);
		bc_msg_stats(bc_stats_buf, sizeof(bc_stats_buf));
		bc_printk(BC_DEBUG, bc_stats_buf);
		return(SCSI_MLQUEUE_HOST_BUSY);
	    }
	}
	
	memset(srb, 0, size);

	srb->function     = BC_SRB_EXECUTE_SCSI;
	srb->status       = BC_SRB_STATUS_SUCCESS;
	srb->bus          = bus;
	srb->target       = target;
	srb->lun          = lun;
	srb->flags        = BC_SRB_FLAGS_UNSPECIFIED_DIRECTION;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	srb->data_len     = scsi_bufflen(scp);
#else
	srb->data_len     = scp->request_buflen;
#endif
	srb->cdb          = scp->cmnd;
	srb->cdb_len      = scp->cmd_len;
	srb->scsi_context = scp;
	srb->sg_list	  = (bc_sg_list_t *)&srb->private32[0];
	srb->dev_private  = (char *)srb->sg_list + sg_list_size;
	srb->seq_num 	  = bc_srb_seq_num;
	bc_srb_seq_num++;

	scp->SCp.ptr = (void *)srb;	/* the scsi_cmnd pointer points at our srb, at least until the command is aborted */

	bc_msg_build_sg(srb);

	bc_printk(BC_DEBUG2, "\nbc_msg_send_srb: seq_num: %d B/T/L %d/%d/%d\n", 
			srb->seq_num, bus, target, lun);
        // bc_dump_scp(scp);

	/*
	 *   Add the srb to the tail of the queue, schedule the q tasklet, and return.
	 *   Sending srb's through the OSIC can take a long time. We don't want other
	 *   CPU's spinning on the OSIC lock and just wasting cycles.  So we queue the srb's
	 *   and just have one tasklet entering the OSIC.
	 */

// STATS 

	state->stats.scb_total++;
	state->stats.target_total[target]++;

	atomic_inc(&state->stats.scb_pending);
	atomic_inc(&state->stats.target_pending[target]);

	spin_lock_irqsave(&state->srb_q.lock, irql);

	if (state->srb_q.head == (bc_srb_t *)0) {
		state->srb_q.head = srb;
		state->srb_q.tail = srb;
	} else {
		state->srb_q.tail->next = srb;
		state->srb_q.tail = srb;
	}
	spin_unlock_irqrestore(&state->srb_q.lock, irql);

	tasklet_schedule(&state->srb_q.tasklet);

	return(0);
}

void
bc_msg_process_srb(bc_srb_t *srb )
{
	bc_softstate_t	*state;
	bc_send_arg_t	args;
	struct scsi_cmnd	*scp;
	int		queued;

	state = &bc_state;

	bc_printk(BC_DEBUG2, "\nbc_msg_process_srb: seq_num: %d B/T/L %d/%d/%d\n", 
			srb->seq_num, srb->bus, srb->target, srb->lun);

	scp = (struct scsi_cmnd *) srb->scsi_context;
	
	if (srb->callback == NULL &&  scp == NULL) {
		/* the command was aborted so forget about it... */

// STATS
		state->stats.srb_total++;
		atomic_dec(&state->stats.target_pending[srb->target]);
		atomic_dec(&state->stats.scb_pending);

		bc_printk(BC_DEBUG2, "bc_msg_process_srb: seq_num %d Aborted\n",
			srb->seq_num);
		srb->seq_num = -1;
		kfree(srb);
		return;
	}
	
	if (scp)
	    scp->result = 0xffffffff;

	spin_lock(&state->osic_lock);
	check_lock(state);
	state->osic_locked = 1;
	state->osic_lock_holder = "bc_msg_process_srb";
// STATS
	atomic_inc(&state->stats.srb_pending);
	state->stats.srb_total++;

	args.call_type = BC_CTS_SEND_SRB;
	args.u.send_srb.srb = srb;
        args.u.send_srb.queued = 0;
	bc_send_msg(&args);
	queued = args.u.send_srb.queued;

	state->osic_locked = 0;
	spin_unlock(&state->osic_lock);

	/*
	 * expected state is queued == 1
	 */

	if (queued) {
		bc_printk(BC_DEBUG2, "bc_msg_process_srb: seq_num %d Queued\n",
			srb->seq_num);
		return;
	}
	/*
	 * be carefull about using srb as it may already have been freed
	 * if the srb wasn't queued, we can still use srb.
	 */

	if (srb->seq_num == -1) {
		bc_printk(BC_WARN, "bc_msg_process_srb: seq_num %d  already freed\n",
			srb->seq_num);
		return;
	}

	/*
	 * unexpected status PENDING
	 */
	if (srb->status == BC_SRB_STATUS_PENDING) {
		bc_printk(BC_WARN, "bc_msg_process_srb: seq_num %d STATUS_PENDING\n",
			srb->seq_num);
		bc_dump_scp((struct scsi_cmnd *) srb->scsi_context);
		return;
	}

	/*
	 * just queue up the srb and let the tasklet handle the completion
	 */
	bc_msg_srb_done(srb);
	return;
}

/*
 * 
 * First check for any interrupts that need to be processed.
 * Next send any memory operattion responses
 * Then queue to the OSIC any pending SRBs
 */

void
bc_msg_srb_q_tasklet(  unsigned long arg)
{
	unsigned long	irql;
	bc_srb_t	*srb;
	bc_softstate_t *state;
	bc_stats_t	*sp;
	bc_send_arg_t	args;
	int		progress;
	bc_mem_op_t	*mop;
// STATS
	unsigned int stat_last_pending;
	unsigned int stat_nsrbs;
	unsigned int stat_intr_low, stat_intr_hi;

	state = (bc_softstate_t *)arg;
	sp = &state->stats;

	bc_printk(BC_DEBUG_ISR,"bc_msg_srb_q_tasklet: running\n");

	/*
	 * If our card shares interrupts with another card,
	 * we may get scheduled because of an interrupt on that card.
	 * If the OSIC isn't ready to process interrupts yet,
	 * just defer running the tasklet
	 */
	if (( state->state & PROCESS_INTR) == 0) {
		tasklet_schedule(&state->srb_q.tasklet);
		return;
	}


	/*
	 * Keep checking for more work as long as we did anything
	 */
	do {
		/*
		 * Process any pending interrupts
		 */
		progress = 0;
		stat_last_pending = 0;
		stat_intr_low = 0;
		spin_lock_irqsave(&state->intr_lock, irql);
		if (state->intr_pending) {
			  state->intr_pending -= 1;
			  stat_last_pending = state->intr_pending;
			  spin_unlock_irqrestore(&state->intr_lock, irql);
			  ++progress;

			  bc_printk(BC_DEBUG_ISR, "bc_msg_srb_q_tasklet, doing intr call\n");
			  spin_lock(&state->osic_lock);
			  check_lock(state);
			  state->osic_locked = 1;
			  state->osic_lock_holder = "bc_msg_srb_q_tasklet:intr";
			  args.call_type = BC_CTS_INTERRUPT_CALL;
			  args.u.interrupt_call.controller_handle = (bc_adapter_t *)0;
			  bc_send_msg(&args);
			  state->osic_locked = 0;
			  spin_unlock(&state->osic_lock);

			  spin_lock_irqsave(&state->intr_lock, irql);
		}
		spin_unlock_irqrestore(&state->intr_lock, irql);


		/*
		 * process all pending MORBs
		 */
		spin_lock_irqsave(&state->mop_done.lock, irql);

		while (state->mop_done.head) {
			/*
			 * take an morb off the list
			 */
			mop = state->mop_done.head;
			state->mop_done.head = mop->next;
			mop->next = (bc_mem_op_t *)0;
			spin_unlock_irqrestore(&state->mop_done.lock, irql);

			spin_lock(&state->osic_lock);
			check_lock(state);
			state->osic_locked = 1;
			state->osic_lock_holder = "bc_msg_srb_q_tasklet: morb";

			args.call_type = BC_CTS_MEMORY_OP_RESP;
			args.u.mem_op_resp.id = mop->id;
			args.u.mem_op_resp.status = 1;
			bc_send_msg(&args);

			state->osic_locked = 0;
			spin_unlock(&state->osic_lock);

			spin_lock_irqsave(&state->mop_done.lock, irql);
			progress++;
		}
		spin_unlock_irqrestore(&state->mop_done.lock, irql);

		/*
		 * process all pending SRBs
		 */
		spin_lock_irqsave(&state->srb_q.lock, irql);

		stat_nsrbs = 0;
		while (state->srb_q.head) {
			/*
			 * take an srb off the list
			 */
			srb = state->srb_q.head;
			state->srb_q.head = srb->next;
			srb->next = (bc_srb_t *)0;
			spin_unlock_irqrestore(&state->srb_q.lock, irql);
			++progress;
			++stat_nsrbs;

			bc_msg_process_srb(srb);
			spin_lock_irqsave(&state->srb_q.lock, irql);
			if (stat_last_pending != state->intr_pending) {
				if (!stat_intr_low)
					rdtscl(stat_intr_low);
			}
		}
// STATS
		/*
		 * compute:
		 * 	max number of srbs sent to the core at once
		 *	max interrupts that came in while in the core
		 *	total interrupts that came in while in the core
		 *	max time an interrupt waited for the core
		 */
		if (sp->max_srbs_sent < stat_nsrbs)
			sp->max_srbs_sent = stat_nsrbs;
		if (stat_last_pending < state->intr_pending) {
		  	stat_last_pending = state->intr_pending - stat_last_pending;
		  	if (sp->max_intr_waiting < stat_last_pending)
				sp->max_intr_waiting = stat_last_pending;
			sp->total_intr_waiting += stat_last_pending;
			if (stat_intr_low) {
				rdtscl(stat_intr_hi);
				stat_intr_hi -= stat_intr_low;
				if (sp->max_intr_delay < stat_intr_hi)
					sp->max_intr_delay = stat_intr_hi;
			}
		}

		if (state->intr_pending)
			++progress;
		spin_unlock_irqrestore(&state->srb_q.lock, irql);

	} while (progress);

	bc_printk(BC_DEBUG_ISR,"bc_msg_srb_q_tasklet: done\n");
}
void
bc_msg_srb_done_tasklet(  unsigned long arg)
{
	unsigned long irql;
	bc_srb_t *srb;
	bc_softstate_t *state;

	state = (bc_softstate_t *)arg;


	spin_lock_irqsave(&state->srb_done.lock, irql);

	while (state->srb_done.head) {
		/*
		 * take an srb off the list
		 */
		srb = state->srb_done.head;
		state->srb_done.head = srb->next;
		srb->next = (bc_srb_t *)0;

		spin_unlock_irqrestore(&state->srb_done.lock, irql);
		bc_msg_srb_complete(srb);
		spin_lock_irqsave(&state->srb_done.lock, irql);
	}

	spin_unlock_irqrestore(&state->srb_done.lock, irql);
}

void 
bc_msg_srb_done(bc_srb_t *srb)
{
	bc_softstate_t *state;
	unsigned long irql;

	state = &bc_state;

	srb->next = (bc_srb_t *)0;

	bc_printk(BC_DEBUG2, "bc_msg_srb_done: seq_num %d  queued\n",
		srb->seq_num);

	/*
	 * OSIC lock is already held entering this function
	 */
// STATS
	atomic_dec(&state->stats.srb_pending);

	/*
	 * add to tail of srb queue
	 */

	spin_lock_irqsave(&state->srb_done.lock, irql);

	if (state->srb_done.head == (bc_srb_t *)0) {
		state->srb_done.head = srb;
		state->srb_done.tail = srb;
	} else {
		state->srb_done.tail->next = srb;
		state->srb_done.tail = srb;
	}
	spin_unlock_irqrestore(&state->srb_done.lock, irql);

	tasklet_schedule(&state->srb_done.tasklet);
}

void 
bc_msg_srb_complete(struct bc_srb_s *srb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned long irql;
#endif
	struct scsi_cmnd	*scp;
	bc_softstate_t *state;
	int warn;

	warn  = (bc_msg_level >= BC_DEBUG) ? 1: 0;

	state = &bc_state;

	if ( srb->function == BC_SRB_SHUTDOWN
	   || srb->function == BC_SRB_FLUSH )
	{
		bc_softstate_t *state;

		state = &bc_state;
		bc_printk(BC_DEBUG,"bc_msg_srb_complete: shutdown complete\n");
		kfree(srb);
		up(&state->init_sema);
		return;
	}


// STATS	
	atomic_dec(&state->stats.target_pending[srb->target]);
	atomic_dec(&state->stats.scb_pending);

	
	if (srb->callback) {
		srb->callback(srb);
		/* kfree will be done much later */
		return;
	}

	if ((scp = (struct scsi_cmnd *)srb->scsi_context) == NULL) {
		/* the request was aborted, so forget about it... */
		bc_printk(BC_DEBUG2, "bc_msg_srb_complete: seq_num %d  Aborted\n",
			srb->seq_num);
		srb->seq_num = -1;
		kfree(srb);
		return;
	}

	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,1)

	/*
         * Normally, report the arrays as removable disks.
	 * That way, the kernel will send us lock/unlock commands so we can prevent the
	 * removal of arrays that are in use.
	 *
	 * However, early versions of Red Hat's Anaconda installer refuse to install to removable disks.
	 * The only 386 kernel we support is the boot kernel, so if we're building 
	 * the boot kernel, report the disks as fixed.
	 */
	if (scp->cmnd[0] == SS_INQUIR) {
		bc_inq_data_t 		*inq_data;
		inq_data = (bc_inq_data_t *) scp->request_buffer;
#ifdef CONFIG_M386
		inq_data->dev_qual = 0x00;              // Fixed media
#else
		inq_data->dev_qual = 0x80;              // Removable media
#endif
	}

#else
	/*
	 * report the arrays as removable disks.
	 */
	if (scp->cmnd[0] == SS_INQUIR) {
		struct scatterlist	*sg;
		bc_inq_data_t 		*inq_data;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
		if (scsi_sg_count(scp) == 0) {
#else
		if (scp->use_sg == 0) {
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
			inq_data = (bc_inq_data_t *) scsi_sglist(scp);
#else
			inq_data = (bc_inq_data_t *) scp->request_buffer;
#endif
			inq_data->dev_qual = 0x80;             
		} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
			sg = (struct scatterlist *) scsi_sglist(scp);
#else
			sg = (struct scatterlist *)scp->request_buffer;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
			inq_data = kmap_atomic(sg_page(sg), KM_IRQ0) + sg->offset;
#else
			inq_data = kmap_atomic(sg->page, KM_IRQ0) + sg->offset;
#endif
			inq_data->dev_qual = 0x80;            
			kunmap_atomic(inq_data - sg->offset, KM_IRQ0);
		}
	}
	//scp->device->removable = 1;
#endif

	scp->SCp.ptr = NULL;

	if (srb->status == BC_SRB_STATUS_SUCCESS) {
		bc_printk(BC_DEBUG2, "bc_msg_srb_complete: seq_num %d  SUCCESS\n",
			srb->seq_num);
		scp->result = DID_OK << 16 | COMMAND_COMPLETE << 8 | GOOD;

		GET_IO_REQUEST_LOCK_IRQSAVE(irql);
		scp->scsi_done (scp);
		PUT_IO_REQUEST_LOCK_IRQRESTORE(irql);
//		bc_printk(BC_DEBUG2, "bc_msg_srb_complete: seq_num %d  scsi_done DONE\n",
//			srb->seq_num);
		srb->seq_num = -1;
		kfree(srb);
		return;
	} 
	
	/*
	 * no device
	 */
	if (srb->status == BC_SRB_STATUS_NO_DEVICE
		|| srb->status == BC_SRB_STATUS_INVALID_LUN
		|| srb->status == BC_SRB_STATUS_INVALID_TARGET_ID
		|| srb->status == BC_SRB_STATUS_INVALID_PATH_ID) 
	{
		scp->result = DID_BAD_TARGET << 16;
		GET_IO_REQUEST_LOCK_IRQSAVE(irql);
		scp->scsi_done (scp);
		PUT_IO_REQUEST_LOCK_IRQRESTORE(irql);
		bc_printk(BC_DEBUG2, "bc_msg_srb_complete: seq_num %d STATUS_NO_DEVICE\n",
			srb->seq_num);
		srb->seq_num = -1;
		kfree(srb);
		return;
	}
	/*
	 * Something went wrong.  May need to check specific error codes
	 */

	scp->result = DID_OK << 16 | COMMAND_COMPLETE << 8 | CHECK_CONDITION;

	bc_set_sense_data( (char *)scp->sense_buffer,
				SENKEY_HW_ERR,
				SENCODE_INTERNAL_TARGET_FAILURE,
				ASENCODE_INTERNAL_TARGET_FAILURE,
				0, 0, 0, 0);

	/*
	 * we do not implement all the mode pages for the mode sense command.
	 * Suppress the failure warning for those commands
	 */
	if (scp->cmnd[0] == SS_MODESEN) {
		switch(scp->cmnd[2]) {
		    case 0x01:		/* Read-Write Error Recovery mode */
		    case 0x02:		/* Disconnect-Reconnect	mode	  */
		    case 0x08:		/* Caching mode			  */
		    case 0x0a:		/* Control mode			  */
			    warn = 0;
			    break;

		    default:
			    break;
		}
	}

	if (warn) {
		bc_printk (BC_DEBUG, "bc_msg_srb_complete: seq_num %d srb failed, status 0x%x\n", 
			srb->seq_num, srb->status);
		/*
		 * dump the command that failed
		 */
		bc_dump_scp(scp);
	}

	GET_IO_REQUEST_LOCK_IRQSAVE(irql);
	scp->scsi_done (scp);
	PUT_IO_REQUEST_LOCK_IRQRESTORE(irql);

	bc_printk(BC_DEBUG, "bc_msg_srb_complete: seq_num %d  scsi_done DONE\n",
		srb->seq_num);
	/*
	 * free dma mappings JWF
	 */
	srb->seq_num = -1;
	kfree(srb);

}

/* 
 * send a flush cache or shutdown srb to the OSIC.
 * Wait for IO to stop
 */

void
bc_msg_send_srb_function (bc_softstate_t *state, int function)
{
	bc_srb_t	*srb;
	bc_send_arg_t	args;
	int 		size;
	int		queued;
	int		sg_list_size;
	bc_sg_list_t	*bc_sg;

	bc_printk(BC_DEBUG, "bc_msg_send_srb_function\n");

	
	sg_list_size = sizeof (bc_sg_list_t) + (BC_SG_MAX_ELEMENTS-1) * sizeof(bc_sg_elem_t);
	size = sizeof(bc_srb_t) + sg_list_size + state->memsize_per_srb;

	/*
	 * Use GFP_ATOMIC instead of GP_KERNEL because we get called from the srb tasklet
	 */
	if ((srb =  kmalloc( size, GFP_ATOMIC)) == 0) {
		bc_printk(BC_WARN,"bc_msg_send_srb_function: could not alloc %d bytes for srb\n", size);
		return;
	}
	

	srb->function     = function;
	srb->status       = BC_SRB_STATUS_SUCCESS;
	srb->bus          = 0;
	srb->target       = 0;
	srb->lun          = 0;
	srb->flags        = BC_SRB_FLAGS_UNSPECIFIED_DIRECTION;
	srb->data_len     = 0;
	srb->cdb          = 0;
	srb->cdb_len      = 0;
	srb->scsi_context = 0;
	srb->sg_list	  = (bc_sg_list_t *)&srb->private32[0];
	srb->dev_private  = (char *)srb->sg_list + sg_list_size;
	srb->seq_num 	  = bc_srb_seq_num;
	bc_srb_seq_num++;

	bc_sg = srb->sg_list;
	bc_sg->sg_num_elem = 0;
	bc_sg->sg_mem_type = BC_MEM_UNSET;
	bc_sg->sg_elem[0].dma_paddr = 0;
	bc_sg->sg_elem[0].size = 0;

	/*
	 * on a uniprocessor kernel, the spin_locks disappear and we have no
	 * lock to prevent reentering the OSIC.
	 * we can't allow bc_msg_timer to execute while we're in the middle of
	 * this call to the OSIC, as it also enters the OSIC.
	 * So, disable the bottom-half handlers, which will prevent the msg_timer
	 * from running.  After the bh handlers are re-enabled, they'll get executed later
	 */

	local_bh_disable();
	spin_lock(&state->osic_lock);
	check_lock(state);
	state->osic_locked = 1;
	state->osic_lock_holder = "bc_msg_send_srb_function";
	args.call_type = BC_CTS_SEND_SRB;
	args.u.send_srb.srb = srb;
        args.u.send_srb.queued = 0;
	bc_send_msg(&args);
	queued = args.u.send_srb.queued;

	state->osic_locked = 0;
	spin_unlock(&state->osic_lock);
	local_bh_enable();

	/*
	 * expected state is queued == 1
	 */

	if (!queued) {
		bc_printk(BC_WARN, "bc_msg_send_srb_function: seq_num %d  NOT Queued\n",
			 srb->seq_num);
		kfree(srb);
		return;
	}

	/*
	 * wait for the srb_function to complete
	 */
	bc_printk(BC_DEBUG, "bc_msg_send_srb_function: waiting for completion\n");
	down(&state->init_sema);
	bc_printk(BC_DEBUG, "bc_msg_send_srb_function: complete\n");
	return;
}



/*
 * build a scatter/gather list of Virtual addresses to send to OSIC.
 * We only do this when all sg elements are already mapped into the kernel.
 * We've already checked that this is the case.
 */

void __inline__
bc_msg_build_sg_virt( bc_srb_t *srb)
{

	int		i;
	struct scatterlist *sg;
	struct scsi_cmnd	*scp;
	bc_sg_list_t	*bc_sg = 0;

	scp = (struct scsi_cmnd *)srb->scsi_context;
	BC_ASSERT(scp != NULL);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	sg = (struct scatterlist *) scsi_sglist(scp);
#else
	sg = (struct scatterlist *)scp->request_buffer;
#endif

	bc_sg = srb->sg_list; 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	bc_sg->sg_num_elem = scsi_sg_count(scp);
#else
	bc_sg->sg_num_elem = scp->use_sg;
#endif
	bc_sg->sg_mem_type = BC_MEM_VADDR;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	for (i = 0; i < scsi_sg_count(scp); i++) {
#else
	for (i = 0; i < scp->use_sg; i++) {
#endif
		bc_sg->sg_elem[i].size = sg[i].length;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
		bc_sg->sg_elem[i].v_addr = page_address(sg_page(&(sg[i]))) + sg[i].offset;
#else
		bc_sg->sg_elem[i].v_addr = page_address(sg[i].page) + sg[i].offset;
#endif
	}
}

/*
 * build a scatter/gater list of physical address to send to the OSIC.
 */
void __inline__
bc_msg_build_sg_phys( bc_srb_t *srb)
{

	int		i;
	struct scatterlist *sg;
	dma_addr_t	dma_addr;
	struct scsi_cmnd	*scp;
	bc_adapter_t 	*adapter;
	bc_sg_list_t	*bc_sg;

	adapter = bc_dev[0];			

	scp = (struct scsi_cmnd *)srb->scsi_context;
	BC_ASSERT(scp != NULL);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	sg = (struct scatterlist *) scsi_sglist(scp);
#else
	sg =  (struct scatterlist *) scp->request_buffer;
#endif

	bc_sg = srb->sg_list;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	bc_sg->sg_num_elem = scsi_sg_count(scp);
#else
	bc_sg->sg_num_elem = scp->use_sg;
#endif
	bc_sg->sg_mem_type = BC_MEM_PADDR | BC_MEM_USE_EXTERNAL_LIST_COPY;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	for (i = 0; i < scsi_sg_count(scp); i++) {
#else
	for (i = 0; i < scp->use_sg; i++) {
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	    dma_addr = page_to_phys(sg_page(&(sg[i]))) + sg[i].offset;
#else
	    dma_addr = page_to_phys(sg[i].page) + sg[i].offset;
#endif
	    bc_sg->sg_elem[i].dma_paddr = dma_addr;
	    bc_sg->sg_elem[i].size =  sg[i].length;
	}
}


void
bc_msg_build_sg( bc_srb_t *srb)
{
	struct scsi_cmnd	*scp;
	bc_sg_list_t	*bc_sg;
	int		has_highmem;

#if defined(CONFIG_HIGHMEM) || (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,1))
        int             i;
        struct scatterlist *sg;
#endif

	scp = (struct scsi_cmnd *)srb->scsi_context;
	BC_ASSERT(scp != NULL);
	
	bc_sg = srb->sg_list;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	if ( scsi_sg_count(scp) == 0) {					/* has to be a virtual address */
#else
	if (scp->use_sg == 0) {					/* has to be a virtual address */
#endif
		bc_sg->sg_num_elem = 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
		bc_sg->sg_elem[0].size   = scsi_bufflen(scp);
		bc_sg->sg_elem[0].v_addr = scsi_sglist(scp);
#else
		bc_sg->sg_elem[0].size   = scp->request_bufflen;
		bc_sg->sg_elem[0].v_addr = scp->request_buffer;
#endif
		bc_sg->sg_mem_type = BC_MEM_VADDR;

		BC_PRINTK(BC_DEBUG2,"bc_msg_build_sg: scp: 0x%p use_sg 0, buf 0x%8x:%8x, len %d\n",
			  scp,
			  (bc_uint32_t)(bc_sg->sg_elem[0].dma_paddr >> 32),
			  (bc_uint32_t)(bc_sg->sg_elem[0].dma_paddr & 0xffffffff),
			  bc_sg->sg_elem[0].size);
		return;
	} 

	has_highmem = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,1)
    {
	int		has_virt, has_phys;
	int		target;
	dma_addr_t	dma_addr;

	has_virt = 0;
	has_phys = 0;
	target = scp->target;

	/*
         * On some 2.4 versions, the sg driver will send us virtual addresses even
	 * though we've declared ourselves a highmem device
	 * If the IO is to the config device, we can still send virtual addresses.
	 * If the IO is to an array, we may be forced to convert to physical
	 * addresses depending on the underlying hardware.
	 */
	
	sg =  (struct scatterlist *) scp->request_buffer;
	for (i = 0; i < scp->use_sg; i++) {
            if (sg[i].address)
		has_virt = 1;

	    if (sg[i].page)
		has_phys = 1;

	    /*
	     * The following code is paranoid, but the check is inexpensive
	     */
            if (sg[i].address && sg[i].page)
                BUG();

            if (!sg[i].address && !sg[i].page)
                BUG();

	    if (has_virt && has_phys)
		BUG();
	}


	/*
         * if this is the 'config' target, 16, we know we don't do any real IO to it.
	 */

	if (has_virt) {
	    bc_sg = srb->sg_list;
	    bc_sg->sg_num_elem = scp->use_sg;

	    sg =  (struct scatterlist *) scp->request_buffer;

	    if ((target == 16) || (ForcePhysAddr == 0)) {
		for (i = 0; i < scp->use_sg; i++) {
		    bc_sg->sg_elem[i].size  = sg[i].length;
		    bc_sg->sg_elem[i].v_addr = sg[i].address;
		}
		bc_sg->sg_mem_type = BC_MEM_VADDR;
		return;
	    }

	    /*
	     * must convert to physical addresses
	     * Target != 16 and ForcePhysAddr is set
	     */

	    for (i = 0; i < scp->use_sg; i++) {
		    bc_sg->sg_elem[i].size = sg[i].length;
		    dma_addr = virt_to_phys(sg[i].address);
		    bc_sg->sg_elem[i].dma_paddr = dma_addr;
	    }
	    bc_sg->sg_mem_type = BC_MEM_PADDR;
	    return;
	}
    }
#endif

	if (ForcePhysAddr && scp->device->id != 16) {
	    bc_msg_build_sg_phys(srb);
	    return;
	}

	/*
	 * Run through all the elements and see if they are all in low memory
	 */
#ifdef CONFIG_HIGHMEM
	{
		int             num_pages;

		sg =  (struct scatterlist *) scp->request_buffer;
		for (i = 0; i < scp->use_sg; i++) {
			num_pages = (sg[i].offset + sg[i].length + PAGE_SIZE - 1) >> PAGE_SHIFT;
			if ( PageHighMem(sg[i].page + num_pages - 1)) {
                 has_highmem = 1;
                 break;
            }
		}
	}
#endif
	if (has_highmem)
	    bc_msg_build_sg_phys(srb);
	else
	    bc_msg_build_sg_virt(srb);

	return;
}
		
/*
 * We really don't have to do anything on i386 or amd x86_64 machines
 */

void 
bc_msg_release_sg(bc_srb_t *srb)
{

}


/*
 * convert a vmalloc'd virtual address to physical
 */
static __inline__ dma_addr_t
bc_virt_to_phys( void * addr)
{
	struct page *page;
	dma_addr_t paddr, offset, ret; 

	ret = (dma_addr_t)0;

	page = vmalloc_to_page(addr);
	if (!page)
		return(ret);

	paddr = page_to_phys(page);

	offset = (unsigned long)addr & (PAGE_SIZE-1);

	ret = paddr | offset;

	BC_PRINTK(BC_DEBUG3,", addr %p, paddr 0x%llx, offset 0x%llx, ret 0x%llx \n",
		addr, (bc_uint64_t)paddr, (bc_uint64_t)offset, (bc_uint64_t)ret);

	return ret;
}

int
bc_vmap_to_phys( struct pci_dev *hwdev,
		void *vaddr, 
		size_t len,
		size_t *len_mapped,
		dma_addr_t *dma_addr)
{
		unsigned long offset;
		unsigned long mapped;
		dma_addr_t paddr;


		paddr  = bc_virt_to_phys(vaddr);
		if (!paddr) {
			*dma_addr = (dma_addr_t )0;
			*len_mapped = 0;
			return(0);
		}

		offset = (unsigned long)vaddr & (PAGE_SIZE-1);

		mapped = PAGE_SIZE - offset;

		if (len < mapped)
			*len_mapped = len;
		else
			*len_mapped = mapped;

		*dma_addr = paddr;

		return (*dma_addr);

}
		
void 
bc_msg_map_mem(struct map_memory_s *map)
{
	bc_adapter_t 	*adapter;
	void		*vaddr;
	dma_addr_t	dma_addr;
	int		len;
	void		*private_start;		// Adapter private memory
	void		*private_end;
	int		offset;
	size_t		len_mapped;
	bc_uint64_t	paddr;
 
	/* Physical memory type, needs no translation. */
	if ((map->memory_id & MEM_TYPE) == BC_MEM_PADDR) {
		map->physical_address = map->address;
        } 

	/* Virtual addresses are a whole lot of fun. */
	else if ((map->memory_id & MEM_TYPE) == BC_MEM_VADDR) {
		adapter = (bc_adapter_t *) map->dev_handle;
		vaddr = (void *)(bc_uint_ptr_t)map->address;
		len   = (size_t)map->number_bytes;

		private_start = adapter->private_mem.vaddr;
		private_end  =  private_start + adapter->private_mem.size;

	        /*
 	         * check to see if the request is in the device private memory region.
	         */
		if ((vaddr >= private_start) && (vaddr < private_end) ) {
			if (vaddr + len >=  private_end) {
				bc_printk(BC_WARN, "bc_msg_map_mem: invalid address range: %p len %d\n",
					  vaddr, len);
				map->number_bytes = 0;
				return;
			}

                        // bc_printk(BC_DEBUG3, "private space map\n");
			offset = vaddr - private_start;
			paddr = (bc_uint64_t) adapter->private_mem.dma_address + offset;
			map->number_bytes = (bc_uint64_t)len;

		}
		else if (((unsigned long)vaddr >= VMALLOC_START) && ((unsigned long)vaddr < VMALLOC_END)) {
			bc_vmap_to_phys(adapter->pdev, vaddr, len, &len_mapped, &dma_addr);
			paddr = (bc_uint64_t) dma_addr;
			map->number_bytes = (bc_uint64_t)len_mapped;
		}
		else {
                        // bc_printk(BC_DEBUG3,"      lo memory\n");
			dma_addr  = pci_map_single(adapter->pdev, vaddr, len,  PCI_DMA_BIDIRECTIONAL);
			paddr = (bc_uint64_t) dma_addr;
			map->number_bytes = (bc_uint64_t)len;
		}

		map->physical_address  = paddr;
	}


	/* No known memory type is coding error */
	else  {
		bc_printk(BC_ERROR, "bc_msg_map_mem: invalid memory type %x\n", map->memory_id);
		BUG();
	}


	BC_PRINTK(BC_DEBUG3, "bc_msg_map_mem: paddr 0x%x::%x len %d\n",
			(bc_uint32_t)(map->physical_address >> 32),
                        (bc_uint32_t)(map->physical_address & 0xffffffff),
                        (bc_uint32_t)map->number_bytes );
	return;

}


void 
bc_msg_unmap_mem(struct unmap_memory_s *unmap)
{

	bc_adapter_t 	*adapter;
	dma_addr_t	dma_addr;
	int		len;
	dma_addr_t	private_start;		// Adapter private memory
	dma_addr_t	private_end;
	bc_uint64_t	paddr;
	int seq_num;

	adapter = (bc_adapter_t *) unmap->dev_handle;
	paddr = unmap->physical_address;
	len   = (bc_uint32_t)unmap->number_bytes;

	if (unmap->srb)
		seq_num = unmap->srb->seq_num;
	else
		seq_num = -1;

	dma_addr = (dma_addr_t) paddr;
	private_start = adapter->private_mem.dma_address;
	private_end  =  private_start + adapter->private_mem.size;

	BC_PRINTK(BC_DEBUG3,
		  "bc_msg_unmap_mem: srb num %i priv_start: 0x%llx priv_end 0x%llx request 0x%llx len %d\n",
		  seq_num, (bc_uint64_t)private_start, (bc_uint64_t)private_end, (bc_uint64_t)dma_addr, len);
		

	/*
 	 * check to see if the request is in the device private memory region.
	 */
	if ((dma_addr >= private_start) && (dma_addr < private_end) ) {
		if (dma_addr + len >=  private_end) {
			bc_printk(BC_WARN, "bc_msg_unmap_mem: invalid address range: 0x%llx len %d\n",
				(bc_uint64_t)dma_addr, len);
			unmap->number_bytes = 0;
			return;
		}
		unmap->number_bytes  = (bc_uint64_t)len;
		return;
	}

//	pci_unmap_single(adapter->pdev, dma_addr, len, PCI_DMA_BIDIRECTIONAL);
	unmap->number_bytes = (bc_uint64_t)len;
	return;


}
		
	
void
bc_msg_timeout_done(unsigned long data)
{
	bc_softstate_t *state;

	state = (bc_softstate_t *)data;
	init_timer(&state->msg_timeout);
	up(&state->msg_timeout_sema);
}

void
bc_msg_timeout( int to)
{
	bc_softstate_t *state;

	state = &bc_state;
	/*
	 * set up timeout
	 */

	init_timer(&state->msg_timeout);
	state->msg_timeout.expires = jiffies  + to;
	state->msg_timeout.data = (unsigned long)state;
	state->msg_timeout.function = bc_msg_timeout_done;
	add_timer(&state->msg_timeout);
	down(&state->msg_timeout_sema);

}

void
bc_set_sense_data (char	*sense,
		    uint8_t	sense_key,
		    uint8_t	sense_code,
		    uint8_t	add_sense_code,
		    uint8_t	incorrect_len,
		    uint8_t	bit_ptr,
		    uint32_t	field_ptr,
		    uint32_t	residue)
{
	sense[0] = 0xF0;	// Sense data Valid, err code 70h (current error)
	sense[1] = 0;		// Segment number, always zero

	if (incorrect_len) {
		sense[2] = sense_key | 0x20;	// Set the ILI bit | sense key
		sense[3] = BYTE3 (residue);
		sense[4] = BYTE2 (residue);
		sense[5] = BYTE1 (residue);
		sense[6] = BYTE0 (residue);
	} else
		sense[2] = sense_key;	// Sense key

	if (sense_key == SENKEY_ILLEGAL)
		sense[7] = 10;		// Additional sense length
	else
		sense[7] = 6;		// Additional sense length

	sense[12] = sense_code;		// Additional sense code
	sense[13] = add_sense_code;	// Additional sense code qualifier
	if (sense_key == SENKEY_ILLEGAL) {
		sense[15] = 0;

		if (sense_code == SENCODE_INVALID_PARAM_FIELD)
			sense[15] = 0x80;	// Std sense key specific field
		// Illegal parameter is in the parameter block

		if (sense_code == SENCODE_INVALID_CDB_FIELD)
			sense[15] = 0xc0;	// Std sense key specific field
		// Illegal parameter is in the CDB block
		sense[15] |= bit_ptr;
		sense[16] = field_ptr >> 8;	// MSB
		sense[17] = field_ptr;		// LSB
	}
}


// STATS
int
bc_msg_stats(char *buf, int buf_size)
{
	int cnt, len, i;
	bc_softstate_t *state;
	bc_stats_t *sp;
	char *cp;

	state = &bc_state;
	sp = &state->stats;
	
	cnt = 0;
	cp = buf;

	len = snprintf(cp, buf_size, "srb total  %d  pending %d scb total %d pending %d\n",
		sp->srb_total,
		atomic_read(&sp->srb_pending),
		sp->scb_total,
		atomic_read(&sp->scb_pending));

	cp  += len;
	cnt += len;

	len = snprintf(cp, buf_size - cnt, "target  ");
	cp  += len;
	cnt += len;

	for (i = 0; i < MAX_ARRAY; i++) {
		len = snprintf(cp, buf_size - cnt, "%5d ", i);
		cp += len;
		cnt += len;
	}

	len = snprintf(cp, buf_size - cnt, "\nTotal  ");
	cp  += len;
	cnt += len;

	for (i = 0; i < MAX_ARRAY; i++) {
		len = snprintf(cp, buf_size - cnt,  "%5d ", sp->target_total[i]);
		cp += len;
		cnt += len;
	}


	len = snprintf(cp, buf_size - cnt, "\nPending  ");
	cp  += len;
	cnt += len;

	for (i = 0; i < MAX_ARRAY; i++) {
		len = snprintf(cp, buf_size - cnt, "%5d ", atomic_read(&sp->target_pending[i]));
		cp += len;
		cnt += len;
	}

	len = snprintf(cp, buf_size - cnt, "\nOSIC lock %d holder %s intr lock %d intr_pending %d\n", 
		state->osic_locked,
		state->osic_lock_holder,
		spin_is_locked(&state->intr_lock),
		state->intr_pending);
	cp  += len;
	cnt += len;

	len = snprintf(cp, buf_size - cnt, "max SRBs sent %u max intr waiting %u total intr waiting %u max intr delay %u\n", 
		sp->max_srbs_sent,
		sp->max_intr_waiting,
		sp->total_intr_waiting,
		sp->max_intr_delay);
	sp->max_srbs_sent = sp->max_intr_waiting = sp->total_intr_waiting = sp->max_intr_delay = 0;
	cp  += len;
	cnt += len;
	*cp = '\0';

	return(cnt);
}

void bc_sysrq_intr (	int key,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
			struct pt_regs *pt_regs,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			struct kbd_struct * kbd,
#endif
#endif
			struct tty_struct * tty)
{
	bc_softstate_t *state;

	state = &bc_state;
	state->intr_pending += 1;			/* this really could become an atomic counter now */

	if (bc_msg_level >= BC_INFO2)			/* only bump msg level if debug was set when module loaded */
		bc_msg_level += 1;

	tasklet_schedule(&state->srb_q.tasklet);
	bc_wakeup_all_threads();
	bc_printk(BC_ALERT, "scheduling tasklet interrupt\n");
}


void bc_sysrq_state (	int key,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
			struct pt_regs *pt_regs,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			struct kbd_struct * kbd,
#endif
#endif
			struct tty_struct * tty)
{

	bc_msg_stats(bc_stats_buf, sizeof(bc_stats_buf));
	bc_printk(BC_ALERT, bc_stats_buf);
	bc_mop_stats(bc_stats_buf, sizeof(bc_stats_buf));
	bc_printk(BC_ALERT, bc_stats_buf);
	
}

