
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 *
 *   MODULE: bc_msg.h
 *
 *	This module contains the interface structures used to communicate
 * 	with the OSIC.
 *
 *
 *   Created By: Chris Franklin
 *
 */

#ifndef BC_MSG_H
#define BC_MSG_H

#include "bc_types.h"
#include "bc_srb.h"

#define BC_COOKIE_VALUE_LO 0x2A1DC02E
#define BC_COOKIE_VALUE_HI 0x38887900

#define BC_INTERFACE_VERSION 0x100

#define BC_INTERFACE_CHECKSUM (BC_COOKIE_VALUE_LO + BC_COOKIE_VALUE_HI + BC_INTERFACE_VERSION)


typedef void (bc_function_t)( void );

#define BC_NUM_TUNABLE_PARAMS 32
typedef enum {
	bc_DoSorted = 0,
	bc_R1SeekMethod,
	bc_AffinityReadRateLimit,
	bc_HonestQueueLimit,
	bc_DoHonestQueue,
	bc_AccountedSizeOutstandingLimit,
	bc_R1_StreamAccessThreshold,
	bc_FlushDirtyDataDelay,
	bc_StreamActivateThreshold,
	bc_FlushBurstLimit,
	bc_FlushResourceQueueEntriesLimit,
	bc_ActiveRaid5FlushesLimit,
	bc_ActiveFlushesLimit,
	bc_DoStream,
	bc_TwoTripWriteHitRate,
	bc_TwoTripReadHitRate,
	bc_TwoTripReadRate,
	bc_TwoTripWriteRate,
	bc_OneTripWriteHitRate,
	bc_OneTripReadHitRate,
	bc_RandomWriteHitRate,
	bc_RandomReadHitRate,
	bc_OneTripReadRate,
	bc_OneTripWriteRate,
	bc_RAID1_AffinityLockSize,
	bc_WriteBypassThreshold,
	bc_DoSmart,
	bc_SmartPollInterval,
} bc_tunable_param_t;

#define RC_SINGLE_IO_ADDRESS_COUNT 17	/* must match what the OSIC sends. */
#define BC_MAX_MEM_OPERANDS	   16

/*
 * memory IDs
 */
#define BC_MEM_UNSET 	0
#define BC_MEM_VADDR 	0x1000
#define BC_MEM_PADDR 	0x2000
#define BC_MEM_MAPPED	0x4000

#define MEM_TYPE	0xf000
#define MEM_OP_MASK	0x0f00
#define MEM_SUB_TYPE	0x00ff

/*
 * force the OSIC to do operations externally
 */
#define BC_MEM_USE_EXTERNAL_LIST_XOR		0x0100
#define BC_MEM_USE_EXTERNAL_LIST_COPY		0x0200
#define BC_MEM_USE_EXTERNAL_LIST_CMP		0x0400
#define BC_MEM_USE_EXTERNAL_CP_CLR		0x0800

typedef enum {
	BC_MEM_LIST_XOR = 1,
	BC_MEM_LIST_COPY,
	BC_MEM_LIST_CMP,
	BC_MEM_CLEAR,
	BC_MEM_COPY
} bc_mem_opcode_t;

/*
 * this memory operation id is opaque to the driver
 */
typedef union bc_mem_op_id_s {
	void 		  *addr;
	bc_uint64_t	 u64;
} bc_mem_op_id_t;


/*
 * a chained list of memory address lists
 * each list can not exceed RC_SINGLE_IO_ADDRESS_COUNT number of elements
 * addresses can be virtual or physical
 * all addresses in chain must be the same type
 *
 * this is a very messy way to pass the parameters, but we're stuck with it.
 */

typedef struct bc_addr_list_s {
	void			*reserved_priv1;	// Reserved for external OSIC use only
	void			*reserved_priv2;	// Reserved for external OSIC use only
	bc_uint8_t 		mem_elem_count;		// elemnt count for this list
	bc_uint8_t		reserved_priv3;		// Reserved for external OSIC use only
	bc_uint16_t 	        mem_id;			// Memory space indentifier
	struct bc_addr_list_s 	*next;			// If more than SINGLE_IO_ADDRESS_COUNT elements, more lists.
	bc_sg_elem_t 		sg_list[RC_SINGLE_IO_ADDRESS_COUNT]; 
} bc_addr_list_t;


typedef struct bc_mem_list_s {
	bc_addr_list_t *addr;
	bc_uint32_t	start_element;
	bc_uint32_t	start_offset;
} bc_mem_list_t;


/*
 * this is a very messy way to pass the parameters, but we're stuck with it
 */
typedef struct bc_mem_list_op_s {
	struct bc_addr_list_s *dst_addr_list;	// address list for destination 		XOR DMA
	bc_uint32_t dst_element;		// starting dest element			    DMA
	bc_uint32_t dst_offset;			// starting dest offset 			XOR DMA CMP
	union{
	    struct bc_addr_list_s **addr_list_array; 	// Array of address lists for source(s)	XOR     CMP
	    struct bc_addr_list_s *addr_list;	// source address list				    DMA
	} src;
	bc_uint32_t src_element;		// starting src element				    DMA
	bc_uint32_t src_offset;			// startint src offset				    DMA
	bc_uint32_t array_count;		// number of sources in array 			XOR     CMP
	bc_uint32_t sector_count;		// Number of sectors				XOR DMA CMP
	bc_uint32_t flags;			// engine flags					XOR
} bc_mem_list_op_t;

#define BC_MEMOP_FLAG_XOR_MOVE_FIRST	0x01	// Ignore destination as part of XOR operation
#define BC_MEMOP_FLAG_XOR_CHECK_ZERO	0x02	// Return success only if final result is zero

typedef struct bc_mem_op_s {
	struct bc_mem_op_s *next;
        bc_mem_op_id_t	id;		// opaque
        bc_mem_opcode_t	opcode;
	bc_uint32_t	status;		// immediate status returned
	union {
	    bc_mem_list_op_t *list;
	    struct {
		bc_uint64_t	dst;
		bc_uint64_t	src;
		bc_uint32_t	dst_id;
		bc_uint32_t	src_id;
		bc_uint32_t	byte_count;
	    } cp;
	    struct {
		bc_uint64_t	dst;
		bc_uint32_t	dst_id;
		bc_uint32_t	byte_count;
	    } clr;
	} mem;
} bc_mem_op_t;


typedef struct bc_mem_op_resp_s {
	bc_mem_op_id_t	id;		// opaque
	bc_uint32_t	status;
} bc_mem_op_resp_t;



/*
 * Memory Operation Request Block
 */
typedef struct  bc_morb_s {
        bc_mem_op_id_t	        id;		// opaque
        bc_mem_opcode_t	        opcode;         
	bc_uint32_t	        status;		 // immediate status returned
	bc_uint32_t			num_arrays;
	bc_uint32_t			byte_count;
	bc_sg_list_t 		*sg_list[16];	  /* start of lists, variable size */
} bc_morb_t;


//
//  Enumerate <C>all <T>ype for <S>end
//
#define BC_CTS_TEST			1
#define BC_CTS_GET_INFO			2
#define BC_CTS_INIT_CONTROLLER		3
#define BC_CTS_FINAL_INIT		4
#define BC_CTS_TIMER_CALL		5
#define BC_CTS_INTERRUPT_CALL		6
#define BC_CTS_SEND_SRB			7
#define BC_CTS_PAUSE			8
#define BC_CTS_RESUME			9
#define BC_CTS_MEMORY_OP_RESP		11

typedef	struct bc_init_controller_s {
	//
	//	Sent to STH from driver
	//
	void *controller_memory;		// cpu memory per controller
	void *controller_handle;
	void *pci_config_space;
	bc_uint32_t pci_config_space_length;
	void *bar_memory[6];			// Virtually mapped memory space
	bc_uint32_t bar_length[6];		// Length of each space
	bc_uint16_t bar_port[6];	    // Port bars
	bc_uint16_t controller_memory_id;	// memory type
	bc_uint16_t reserved;
	
	//
	//	Input from STH
	//
	bc_uint32_t max_split_virtual_memory_size_needed;
	bc_uint32_t max_split_cache_memory_size_needed;
#define virtual_memory_size_needed min_virtual_memory_size_needed
#define cache_memory_size_needed min_cache_memory_size_needed
	bc_uint32_t min_virtual_memory_size_needed;
	bc_uint32_t min_cache_memory_size_needed;
	
} bc_init_controller_t;

typedef struct bc_final_init_s {
	//
	//	Sent to STH from driver
	//
	void *virtual_memory;
	void *cache_memory;
	bc_uint32_t  size_of_virtual_memory_allocated;
	bc_uint32_t  size_of_cache_memory_allocated;
	bc_uint16_t virtual_memory_id;		// memory type
	bc_uint16_t cache_memory_id;		// memory type
	bc_uint32_t timer_interval;

} bc_final_init_t;


//
//	Structure which contains the send argument information
//
typedef struct bc_send_arg_s {
    bc_uint32_t call_type;
    union {
	struct {
	    //
	    //	Sent to STH from driver
	    //
	    bc_uint32_t max_total_memory;
	    bc_uint32_t max_cache_memory;
	    bc_uint32_t max_virtual_memory;
	    bc_uint32_t controller_count;
	    bc_uint32_t	max_sg_map_elements;	// maximun number of sg elements allocated on each srb by the sender    
	    bc_uint32_t max_print_severity;	// Sets the max severity for getting back print messages

	    //
	    //	Input from the STH
	    //
	    bc_uint32_t memory_size_per_controller;
	    bc_uint32_t memory_size_per_srb;
	    bc_uint32_t timer_interval;
	    bc_uint32_t random_seed;
	    bc_uint64_t	max_lba;		// in blocks, we use LBA
	    bc_uint32_t parameter_mask;
	    bc_uint32_t tunable_parameters[BC_NUM_TUNABLE_PARAMS];
	} get_info;
	struct bc_init_controller_s init_controller;
	struct bc_final_init_s final_init;
	struct {
	    //
	    //	Sent to STH from driver
	    //
	    void *controller_handle;
	} interrupt_call;
	struct {
	    struct bc_srb_s *srb;
	    bc_uint32_t	     queued;	// 1 if queued
	} send_srb;
	bc_mem_op_resp_t mem_op_resp;

    } u;
} bc_send_arg_t;


//
//  Enumerate <C>all <T>ype for <R>eceive
//
#define BC_CTR_TEST				1000
#define BC_CTR_PRINTF				1001
#define BC_CTR_SRB_DONE				1002
#define BC_CTR_INIT_DONE			1003
#define BC_CTR_ASSERTION_FAILURE		1004
#define BC_CTR_EVENT				1005
#define BC_CTR_MAP_MEMORY			1006
#define BC_CTR_UNMAP_MEMORY			1007
#define BC_CTR_WAIT_MICROSECONDS		1008
#define BC_CTR_MEMORY_OP			1009

#define BC_CTR_PFTYPE_32_BIT	1
#define BC_CTR_PFTYPE_64_BIT	2
#define BC_CTR_PFTYPE_POINTER	3

#define BC_CTR_EVENT_CONFIG_CHANGE_DETECTED		101
#define BC_CTR_EVENT_CONFIG_ARRAY_OFFLINE		102
#define BC_CTR_EVENT_CONFIG_DISK_OFFLINE		103

#define BC_CTR_PRINTF_SEV_PANIC			0	// After printing, system should be crashed
#define BC_CTR_PRINTF_SEV_ALERT			1	// Errors that must be printed for the user
#define BC_CTR_PRINTF_SEV_CRITICAL		2	// Errors that must be printed for the user
#define BC_CTR_PRINTF_SEV_ERROR			3	// Warnings that should be printed for the user
#define BC_CTR_PRINTF_SEV_WARNING		4	// Warnings that should be printed for the user
#define BC_CTR_PRINTF_SEV_NOTICE		5	// Warnings that should be printed for the user
#define BC_CTR_PRINTF_SEV_INFO			6	// Informational messages for the user
#define BC_CTR_PRINTF_SEV_UNKNOWN		7	// Uncatagorized output, but by may more more the debug
#define BC_CTR_PRINTF_SEV_DEBUG_MIN		8	// Debug only starting point: Any severity at the level

//
//	Structure which contains the receive argument information
//

typedef struct alloc_dma_map_s {

	bc_uint64_t	physical_address;	// RETURNS: Physical address of translation
	bc_uint64_t	address_handle;		// RETURNS: Handle associated with this address
	bc_uint64_t	number_bytes;		// IN/OUT: Number of contiguous bytes translation is valid for
	void		*dev_handle;		// device structure pointer
	bc_srb_t	*srb;
	bc_uint64_t	address;

	bc_uint16_t	memory_id;			// Memory ID for what is being mapped.
} alloc_dma_map_t, map_memory_t;

#define map_memory_s alloc_dma_map_s
#define unmap_memory_s free_dma_map_s

typedef struct free_dma_map_s {
	void		*dev_handle;		// device structure pointer
	bc_srb_t	*srb;
	bc_uint64_t	physical_address;	// phys addr to unmap
	bc_uint64_t	address_handle;		// Handle for address map
	bc_uint64_t	number_bytes;		// number of bytes to unmap
	bc_uint16_t	memory_id;			// Memory ID for what is being mapped.
} free_dma_map_t, unmap_memory_t;

typedef struct bc_print_args_s {
	struct {
		bc_uint64_t val;
		bc_uint64_t type;
	} params[10];
	bc_uint32_t	severity;
	bc_uint64_t	count;
	const char	*format_string;
} bc_print_args_t;

typedef struct bc_receive_arg_s {
    bc_uint32_t call_type;
    bc_uint32_t padding;
    union {
		struct bc_print_args_s printf;
		struct {
		    struct bc_srb_s *srb;
		} srb_done;
		struct {
			char *file_name;
			bc_uint32_t line_number;
		} assertion_failure;
    		struct {
    			bc_uint32_t bc_notification_type;
			bc_uint8_t bc_bus_changed;
		} event;
		map_memory_t	map_memory;
		unmap_memory_t	unmap_memory;
		bc_mem_op_t 	*mem_op;
		struct {
			bc_uint32_t microseconds;
		} wait_microseconds;
    } u;
} bc_receive_arg_t;




//
//	Base structure that allows communication with hardware interface
//
typedef struct bc_interface_s {
    bc_uint32_t cookie_lo;
    bc_uint32_t cookie_hi;
    bc_uint32_t version;
    bc_uint32_t checksum;

    bc_function_t *send_function;
    struct bc_send_arg_s *send_arg;
    
    bc_function_t *receive_function;
    struct bc_receive_arg_s *receive_arg;

} bc_interface_t;




/*
 * function protos
 */
void bc_msg_mem_op(bc_mem_op_t *mop);

#endif //BC_MSG_H
