
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 *	SCSI Commands 
 */
#define	SS_TEST			0x00	/* Test unit ready */
#define SS_REZERO		0x01	/* Rezero unit */
#define	SS_REQSEN		0x03	/* Request Sense */
#define SS_REASGN		0x07	/* Reassign blocks */
#define	SS_READ			0x08	/* Read 6   */
#define	SS_WRITE		0x0A	/* Write 6  */
#define	SS_INQUIR		0x12	/* inquiry */
#define	SS_ST_SP		0x1B	/* Start/Stop unit */
#define	SS_LOCK			0x1E	/* prevent/allow medium removal */
#define SS_RESERV		0x16	/* Reserve */
#define SS_RELES		0x17	/* Release */
#define SS_MODESEN		0x1A	/* Mode Sense 6 */
#define	SS_RDCAP		0x25	/* Read Capacity 10 */ 
#define	SM_READ			0x28	/* Read 10  */
#define	SM_WRITE		0x2A	/* Write 10 */
#define SS_SEEK			0x2B	/* Seek */

#define	SM_READ_10		0x28	/* Read  10 */
#define	SM_WRITE_10		0x2A	/* Write 10 */
#define SM_READ_16		0x88	/* Read  16 */
#define SM_WRITE_16		0x8a	/* Write 16 */
#define SM_REPORT_LUNS		0xa0	/* 12 byte cmd */


/*
 * SERVICE_ACTION_ is the command in byte 0
 * SA_READ_CAP is the 'action' in byte 1
 * Use SERVICE_ACTION_IN with SA_READ_CAP
 */
#undef  SERVICE_ACTION_IN
#undef  SERVICE_ACTION_OUT
#define SERVICE_ACTION_IN	0x9E
#define SERVICE_ACTION_OUT	0x9F
#define	SA_READ_CAP		0x10	/* Read Capacity 16 */



#define BYTE0( x ) ( unsigned char )( x )
#define BYTE1( x ) ( unsigned char )( x >> 8  )
#define BYTE2( x ) ( unsigned char )( x >> 16 )
#define BYTE3( x ) ( unsigned char )( x >> 24 )
	

/* 
 * group 6 commands
 *
 * the 3 most signficant bits of addr[0] define the LUN
 *  Mask the lun by using "addr[0] & 1f" 
 */


typedef struct bc_cmd6{
        unsigned char opcode;            /* opcode         	*/
        unsigned char addr[3];           /* block address   	*/
        unsigned char len;               /* transfer length     */
        unsigned char control;           /* control field      	*/
}bc_cmd6_t;

/*  
 * group 10 commands.
 */ 

typedef struct bc_cmd10{
        unsigned char opcode;      /* opcode                  */
        unsigned char lun;         /* logical unit number:    */
        unsigned char addr[4];     /* block address    	      */
        unsigned char reserved;    /* reserved field          */
        unsigned char len[2];      /* transfer length         */
        unsigned char control;     /* control byte            */
}bc_cmd10_t;

/*  
 * group 16 commands.
 */ 

typedef struct bc_cmd16{
        unsigned char opcode;      /* opcode                  */
        unsigned char lun;         /* logical unit number:    */
        unsigned char addr[8];     /* block address    	      */
        unsigned char len[4];      /* transfer length         */
        unsigned char group;       /* reserved field          */
        unsigned char control;     /* control byte            */
}bc_cmd16_t;

typedef union bc_scb {
	struct bc_cmd6	scsi6;
	struct bc_cmd10	scsi10;
	struct bc_cmd16	scsi16;
} bc_scb_t;



typedef struct inq_cmd {
	uint8_t	opcode;
	uint8_t	lun;
	uint8_t	page;
	uint8_t  resv;
	uint8_t	len;
	uint8_t  ctrl;
}    inq_cmd_t;


/*
 * SCSI inquiry data
 */
typedef struct bc_inq_data {
	uint8_t             dev_type;		/* Peripheral qualifier | Peripheral Device type  */
	uint8_t             dev_qual;		/* RMB | Device type Qualifier  */
	uint8_t             version;		/* ISO version | ECMA version | ANSI-approved version */
	uint8_t             data_fmt;		/* AENC | TrmIOP | Response data format */
	uint8_t             add_len;		/* Additional length (n-4) */
	uint8_t             pad1[2];		/* reserved - must be zero */
	uint8_t             flags;		/* RelAdr | WBus32 | WBus16 |  Sync  | Linked |reserved| CmdQue | SftRe */
	uint8_t             vendor_id[8];	/* vendor ID */
	uint8_t             product_id[16];	/* Product ID */
	uint8_t             rev_level[4];	/* Product Revision Level */
	uint8_t		    vendor_specific[20];
} bc_inq_data_t;


/*
 * values for dev_type: Peripheral device type in plain English
 */
#define	INQD_PDT_DA		0x00	/* Direct-access (DISK) device */
#define	INQD_PDT_PROC		0x03	/* Processor device */
#define	INQD_PDT_CHNGR		0x08	/* Changer (jukebox, scsi2) */
#define	INQD_PDT_COMM		0x09	/* Communication device (scsi2) */
#define	INQD_PDT_NOLUN2 	0x1f	/* Unknown Device (scsi2) */
#define	INQD_PDT_NOLUN		0x7f	/* Logical Unit Not Present */

#define	INQD_PDT_DMASK		0x1F	/* Peripheral Device type Mask */
#define	INQD_PDT_QMASK		0xE0	/* Peripheral Device Qualifer Mask */

typedef struct sense_data_s {
	uint8_t	error_code;		// 70h (current errors), 71h(deferred errors)
	uint8_t 	valid:1;	// A valid bit of one indicates that the information 
					// field contains valid information as defined in the
					// SCSI-2 Standard.

	uint8_t 	segment_number; // Only used for COPY, COMPARE, or COPY AND VERIFY 
					   // commands

	uint8_t 	sense_key:4;	// Sense Key
	uint8_t 	reserved:1;
	uint8_t 	ili:1;		// Incorrect length Indicator
	uint8_t 	eom:1;		// End Of Medium - reserved for random access devices
	uint8_t 	file_mark:1;	// file_mark - reserved for random access devices

	uint8_t 	information[4];	// for direct-access devices, contains the unsigned 
					// logical block address or Residue associated with 
					// the sense key 
	uint8_t 	add_sense_len;	// number of additional sense bytes to follow this field
	uint8_t 	cmnd_info[4];	// not used
	uint8_t 	asc;		// Additional Sense Code
	uint8_t 	ascq;		// Additional Sense Code Qualifier
	uint8_t 	fruc;		// Field Replaceable Unit Code - not used

	uint8_t 	bit_ptr:3;	// indicates which byte of the CDB or parameter data
					// was in error
	uint8_t 	bpv:1;		// bit pointer valid (BPV): 1- indicates that 
					// the bit_ptr field has valid value
	uint8_t 	reserved2:2;
	uint8_t 	cd:1;		// command data bit: 1- illegal parameter in CDB.
					//                   0- illegal parameter in data.
	uint8_t 	sksv:1;

	uint8_t 	field_ptr[2];	// byte of the CDB or parameter data in error
} sense_data_t;

/*
 * Sense keys
 */
#define SENKEY_NO_SENSE      0x00
#define SENKEY_UNDEFINED     0x01
#define SENKEY_NOT_READY     0x02
#define SENKEY_MEDIUM_ERR    0x03
#define SENKEY_HW_ERR        0x04
#define SENKEY_ILLEGAL       0x05
#define SENKEY_ATTENTION     0x06
#define SENKEY_PROTECTED     0x07
#define SENKEY_BLANK         0x08
#define SENKEY_V_UNIQUE      0x09
#define SENKEY_CPY_ABORT     0x0A
#define SENKEY_ABORT         0x0B
#define SENKEY_EQUAL         0x0C
#define SENKEY_VOL_OVERFLOW  0x0D
#define SENKEY_MISCOMP       0x0E
#define SENKEY_RESERVED      0x0F

/*
 * Sense codes
 */
#define SENCODE_NO_SENSE                        0x00
#define SENCODE_END_OF_DATA                     0x00
#define SENCODE_BECOMING_READY                  0x04
#define SENCODE_INIT_CMD_REQUIRED               0x04
#define SENCODE_PARAM_LIST_LENGTH_ERROR         0x1A
#define SENCODE_INVALID_COMMAND                 0x20
#define SENCODE_LBA_OUT_OF_RANGE                0x21
#define SENCODE_INVALID_CDB_FIELD               0x24
#define SENCODE_LUN_NOT_SUPPORTED               0x25
#define SENCODE_INVALID_PARAM_FIELD             0x26
#define SENCODE_PARAM_NOT_SUPPORTED             0x26
#define SENCODE_PARAM_VALUE_INVALID             0x26
#define SENCODE_RESET_OCCURRED                  0x29
#define SENCODE_LUN_NOT_SELF_CONFIGURED_YET     0x3E
#define SENCODE_INQUIRY_DATA_CHANGED            0x3F
#define SENCODE_SAVING_PARAMS_NOT_SUPPORTED     0x39
#define SENCODE_DIAGNOSTIC_FAILURE              0x40
#define SENCODE_INTERNAL_TARGET_FAILURE         0x44
#define SENCODE_INVALID_MESSAGE_ERROR           0x49
#define SENCODE_LUN_FAILED_SELF_CONFIG          0x4c
#define SENCODE_OVERLAPPED_COMMAND              0x4E

/*
 * Additional sense codes
 */
#define ASENCODE_NO_SENSE                       0x00
#define ASENCODE_END_OF_DATA                    0x05
#define ASENCODE_BECOMING_READY                 0x01
#define ASENCODE_INIT_CMD_REQUIRED              0x02
#define ASENCODE_PARAM_LIST_LENGTH_ERROR        0x00
#define ASENCODE_INVALID_COMMAND                0x00
#define ASENCODE_LBA_OUT_OF_RANGE               0x00
#define ASENCODE_INVALID_CDB_FIELD              0x00
#define ASENCODE_LUN_NOT_SUPPORTED              0x00
#define ASENCODE_INVALID_PARAM_FIELD            0x00
#define ASENCODE_PARAM_NOT_SUPPORTED            0x01
#define ASENCODE_PARAM_VALUE_INVALID            0x02
#define ASENCODE_RESET_OCCURRED                 0x00
#define ASENCODE_LUN_NOT_SELF_CONFIGURED_YET    0x00
#define ASENCODE_INQUIRY_DATA_CHANGED           0x03
#define ASENCODE_SAVING_PARAMS_NOT_SUPPORTED    0x00
#define ASENCODE_DIAGNOSTIC_FAILURE             0x80
#define ASENCODE_INTERNAL_TARGET_FAILURE        0x00
#define ASENCODE_INVALID_MESSAGE_ERROR          0x00
#define ASENCODE_LUN_FAILED_SELF_CONFIG         0x00
#define ASENCODE_OVERLAPPED_COMMAND             0x00
