
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 *
 *   MODULE: bc_srb.h
 *
 *		This module describes the SCSI Request Block.
 *
 *   Created By: Eric Noya
 *
 *
 */

#ifndef BC_SRB_H
#define BC_SRB_H

#include "bc_types.h"

#define BC_SG_MAX_ELEMENTS 33

//
// SRB types
//
#define BC_SRB_EXECUTE_SCSI			0x00
#define BC_SRB_CLAIM_DEVICE			0x01
#define BC_SRB_IO_CONTROL			0x02
#define BC_SRB_RECEIVE_EVENT			0x03
#define BC_SRB_RELEASE_QUEUE			0x04
#define BC_SRB_ATTACH_DEVICE			0x05
#define BC_SRB_RELEASE_DEVICE			0x06
#define BC_SRB_SHUTDOWN				0x07
#define BC_SRB_FLUSH				0x08
#define BC_SRB_ABORT_COMMAND			0x10
#define BC_SRB_RELEASE_RECOVERY			0x11
#define BC_SRB_RESET_BUS			0x12
#define BC_SRB_RESET_DEVICE			0x13
#define BC_SRB_TERMINATE_IO			0x14
#define BC_SRB_FLUSH_QUEUE			0x15
#define BC_SRB_REMOVE_DEVICE			0x16
#define BC_SRB_WMI				0x17
#define BC_SRB_LOCK_QUEUE			0x18
#define BC_SRB_UNLOCK_QUEUE			0x19
#define BC_SRB_RESET_LOGICAL_UNIT		0x20
#define BC_SRB_RESTART				0x21

//
// SRB Status
//

#define BC_SRB_STATUS_PENDING			0x00
#define BC_SRB_STATUS_SUCCESS			0x01
#define BC_SRB_STATUS_ABORTED			0x02
#define BC_SRB_STATUS_ABORT_FAILED		0x03
#define BC_SRB_STATUS_ERROR			0x04
#define BC_SRB_STATUS_BUSY			0x05
#define BC_SRB_STATUS_INVALID_REQUEST		0x06
#define BC_SRB_STATUS_INVALID_PATH_ID		0x07
#define BC_SRB_STATUS_NO_DEVICE			0x08
#define BC_SRB_STATUS_TIMEOUT			0x09
#define BC_SRB_STATUS_SELECTION_TIMEOUT		0x0A
#define BC_SRB_STATUS_COMMAND_TIMEOUT		0x0B
#define BC_SRB_STATUS_MESSAGE_REJECTED		0x0D
#define BC_SRB_STATUS_BUS_RESET			0x0E
#define BC_SRB_STATUS_PARITY_ERROR		0x0F
#define BC_SRB_STATUS_REQUEST_SENSE_FAILED	0x10
#define BC_SRB_STATUS_NO_HBA			0x11
#define BC_SRB_STATUS_DATA_OVERRUN		0x12
#define BC_SRB_STATUS_UNEXPECTED_BUS_FREE	0x13
#define BC_SRB_STATUS_PHASE_SEQUENCE_FAILURE	0x14
#define BC_SRB_STATUS_BAD_SRB_BLOCK_LENGTH	0x15
#define BC_SRB_STATUS_REQUEST_FLUSHED		0x16
#define BC_SRB_STATUS_INVALID_LUN		0x20
#define BC_SRB_STATUS_INVALID_TARGET_ID		0x21
#define BC_SRB_STATUS_BAD_FUNCTION		0x22
#define BC_SRB_STATUS_ERROR_RECOVERY		0x23
#define BC_SRB_STATUS_NOT_POWERED		0x24
#define BC_SRB_STATUS_SGLE_ERROR		0x25

//
// SRB Flags
//
#define BC_SRB_FLAGS_QUEUE_ACTION_ENABLE	0x00000002
#define BC_SRB_FLAGS_DISABLE_DISCONNECT		0x00000004
#define BC_SRB_FLAGS_DISABLE_SYNCH_TRANSFER	0x00000008

#define BC_SRB_FLAGS_BYPASS_FROZEN_QUEUE	0x00000010
#define BC_SRB_FLAGS_DISABLE_AUTOSENSE		0x00000020
#define BC_SRB_FLAGS_DATA_IN			0x00000040
#define BC_SRB_FLAGS_DATA_OUT			0x00000080
#define BC_SRB_FLAGS_NO_DATA_TRANSFER		0x00000000
#define BC_SRB_FLAGS_UNSPECIFIED_DIRECTION      (BC_SRB_FLAGS_DATA_IN | BC_SRB_FLAGS_DATA_OUT)

#define BC_SRB_FLAGS_NO_QUEUE_FREEZE		0x00000100
#define BC_SRB_FLAGS_ADAPTER_CACHE_ENABLE	0x00000200
#define BC_SRB_FLAGS_FREE_SENSE_BUFFER		0x00000400

#define BC_SRB_FLAGS_IS_ACTIVE			0x00010000
#define BC_SRB_FLAGS_ALLOCATED_FROM_ZONE	0x00020000
#define BC_SRB_FLAGS_SGLIST_FROM_POOL		0x00040000
#define BC_SRB_FLAGS_BYPASS_LOCKED_QUEUE	0x00080000

#define BC_SRB_FLAGS_NO_KEEP_AWAKE		0x00100000
#define BC_SRB_FLAGS_PORT_DRIVER_ALLOCSENSE	0x00200000
#define BC_SRB_FLAGS_PORT_DRIVER_SENSEHASPORT	0x00400000
#define BC_SRB_FLAGS_DONT_START_NEXT_PACKET	0x00800000

#define BC_SRB_FLAGS_PORT_DRIVER_RESERVED	0x0F000000
#define BC_SRB_FLAGS_CLASS_DRIVER_RESERVED	0xF0000000

#define BC_SRB_GLOBAL_UPDATE  			 	 0
#define BC_SRB_LOCAL_UPDATE_ONLY  		0x01000000

/*
 * SRB layout
 *
 * The entire SRB is allocated in one chunk.
 * The sg_map and the device private space are allocated out of the 'private' space.
 *
 *  --------------------------- 
 * |                           |
 * |        SRB                |
 * |                           |
 *  --------------------------- 
 * |                           |
 * |        sg_map             |
 * |                           |
 *  --------------------------- 
 * |                           |
 * |   device private space    |
 * |                           |
 *  --------------------------- 
 *
 */



typedef union {
	void 		*virt_addr;
	bc_uint64_t	phys_addr;
} addr_elem_t;

#define dma_paddr	addr.phys_addr
#define v_addr		addr.virt_addr

#ifndef BC_PACKED
#define BC_PACKED	
#endif

typedef struct BC_PACKED bc_sg_elem_s {
	addr_elem_t		addr;
	bc_uint32_t		size;	 /* set to machine natural size */
} bc_sg_elem_t;

typedef struct bc_sg_list_s {
	bc_uint16_t	sg_mem_type;	/* Memory ID to pass to routines */
	bc_uint16_t     reserved;
	bc_uint32_t	sg_num_elem;
	bc_sg_elem_t	sg_elem[1];	/* variable size array */
} bc_sg_list_t;

typedef struct bc_srb_s {
	struct bc_srb_s			*next;
	bc_uint32_t			function;
	bc_uint32_t			status;
	bc_uint32_t			bus;
	bc_uint32_t			target;
	bc_uint32_t			lun;
	bc_uint32_t			flags;
	bc_uint32_t			seq_num;
	bc_uint32_t			data_len;
	bc_uint32_t			cdb_len;
	void				*cdb;
	bc_sg_list_t			*sg_list;
	void				*scsi_context;
	void				*dev_private;
	void				(* callback) (struct bc_srb_s *);
	bc_uint32_t			private32[1];
} bc_srb_t;




#endif /*BC_SRB_H*/
